/*
*************************************************************************************************
*     __  __  ____    _           ____     ___    _         ____     ___     ___     ___        *
*     \ \/ / |  _ \  ( )  ___    / ___|   / _ \  | |       |___ \   / _ \   / _ \   / _ \       *
*      \  /  | |_) | |/  / __|   \___ \  | | | | | |         __) | | | | | | | | | | | | |      *
*      /  \  |  __/      \__ \    ___) | | |_| | | |___     / __/  | |_| | | |_| | | |_| |      *
*     /_/\_\ |_|         |___/   |____/   \__\_\ |_____|   |_____|  \___/   \___/   \___/       *
*                                                                                               *
*                                                                                               *
*    Module : XP_PP.H                                                                           *
*                                                                                               *
*    Purpose: Implementation for a wrapper class framework for SQL Server Extended Stored       *
*             Procedures (provided as part of the book "Super SQL Server Systems")              *
*                                                                                               *
*    History: None                                                                              *
*                                                                                               *
*    Copyright (c) 2004 - 2006 by Joseph Gama and PJ Naughter.                                  *
*                                                                                               *
*************************************************************************************************
*/



#ifndef __XP_PP_H__
#define __XP_PP_H__



///////////////////// Includes ////////////////////////////////////////////////

#ifndef _ODS_SRV_H_
#pragma message("To avoid this message please put srv.h into your PCH (normally stdafx.h)")
#include <srv.h> 
#endif

#ifndef _INC_CRTDBG
#pragma message("To avoid this message please put crtdbg.h into your PCH (normally stdafx.h)")
#include <crtdbg.h>
#endif



//////////////////// Forward Macros / Defines / classes ///////////////////////

enum XP_VARIANT_PARAMETER_TYPE
{
  XP_PT_EMPTY,
  XP_PT_BIT,
  XP_PT_ASCIICHAR,
  XP_PT_ASCIITEXT,
  XP_PT_UNICODECHAR,
  XP_PT_UNICODETEXT,
  XP_PT_TCHAR,
  XP_PT_TTEXT,
  XP_PT_INT1,
  XP_PT_INT2,
  XP_PT_INT4,
  XP_PT_INT8,
  XP_PT_FLT4,
  XP_PT_FLT8,
  XP_PT_DECIMAL,
  XP_PT_MONEY4,
  XP_PT_MONEY,
  XP_PT_DATETIM4,
  XP_PT_DATETIME,
  XP_PT_NUMERIC,
  XP_PT_BINARY,
  XP_PT_VARBINARY,
  XP_PT_GUID,
  XP_PT_IMAGE,
};


#ifndef DEFAULTPRECISION
#define DEFAULTPRECISION 18
#endif

#ifndef DEFAULTSCALE
#define DEFAULTSCALE 0
#endif


//Forward declaration
class CExtendedStoredProcedure;
class XP_PARAMETER_DATA;


//struct which represents a definition for an individual parameter of an XP
struct XP_PARAMETER_DEFINITION
{
//Member variables which should be setup initially
  enum ParameterType
  {
    IN_PARAMETER = 0,
    OUT_PARAMETER = 1,
    IN_OR_OUT_PARAMETER = 2,
  }                         m_ParameterType;           //Whether this parameter can be [IN], [OUT] or either
  int                       m_nColumn;                 //The column number of this parameter
  DBCHAR*                   m_szName;                  //The name of this parameter
  unsigned __int64          m_AllowableTypes;          //The allowable data types for this parameter
  BOOL                      m_bAllowNull;              //Should the data for this parameter be allowed to be DB null
  BOOL                      m_bOptional;               //TRUE implies this parameter is optional, FALSE means required
  XP_VARIANT_PARAMETER_TYPE m_RuntimeType;             //The data type which should be constructed for this parameter at runtime
	BYTE                      m_precision;               //Used for setting up DECIMAL and NUMERIC data types
	BYTE                      m_scale;
};


//Struct which encapsulates a SRVIMAGE or SRVVARBINARY data type in a XP_VARIANT_PARAMETER
struct XP_DBIMAGE
{
  BYTE* pbyData;
  int   cbSize;
};

#define XP_DBVARBINARY XP_DBIMAGE


//A "VARIANT" like class to hold the data of a XP parameter at runtime
class XP_VARIANT_PARAMETER 
{
public:
//Constructors / Destructors
  XP_VARIANT_PARAMETER();
  ~XP_VARIANT_PARAMETER();

//Methods
  BOOL Construct(CExtendedStoredProcedure& XP, XP_VARIANT_PARAMETER_TYPE ToType, BYTE CurrentType, BYTE* pCurrentData, ULONG cbCurrentSize, BYTE Precision, BYTE Scale);
  void Clear();

//Member variables
  XP_VARIANT_PARAMETER_TYPE pt;
  union 
  {
    DBBOOL          bVal;          //XP_PT_BIT
    char            cCharVal;      //XP_PT_ASCIICHAR
    char*           pszText;       //XP_PT_ASCIITEXT
    wchar_t         wCharVal;      //XP_PT_UNICODECHAR
    wchar_t*        pszwText;      //XP_PT_UNICODETEXT
    TCHAR           tCharVal;      //XP_PT_TCHAR
    TCHAR*          ptszText;      //XP_PT_TCHARTEXT
    DBTINYINT       byVal;         //XP_PT_INT1
    DBSMALLINT      iVal;          //XP_PT_INT2
    DBINT           lVal;          //XP_PT_INT4
    __int64         llVal;         //XP_PT_INT8
    DBFLT4          fltVal;        //XP_PT_FLT4
    DBFLT8          dblVal;        //XP_PT_FLT8
    DBDECIMAL       decVal;        //XP_PT_DECIMAL 
    DBMONEY4        mny4Val;       //XP_PT_MONEY4
    DBMONEY         mnyVal;        //XP_PT_MONEY
    DBDATETIM4      dt4Val;        //XP_PT_DATETIM4
    DBDATETIME      dtVal;         //XP_PT_DATETIME
    DBNUMERIC       numericVal;    //XP_PT_NUMERIC
    DBBINARY        byBinaryVal;   //XP_PT_BINARY
    XP_DBVARBINARY  varBinaryVal;  //XP_PT_VARBINARY
    GUID            guidVal;       //XP_PT_GUID
    XP_DBIMAGE      ImageVal;      //XP_PT_IMAGE
  } __XP_VARIANT_PARAMETER_NAME_1;
};


#define XP_PT(X)           ((X).pt)
#define XP_UNION(X, Y)     ((X).__XP_VARIANT_PARAMETER_NAME_1.Y)
#define XP_BIT(X)          XP_UNION(X, bVal)
#define XP_ASCIICHAR(X)    XP_UNION(X, cCharVal)
#define XP_ASCIITEXT(X)    XP_UNION(X, pszText)
#define XP_UNICODECHAR(X)  XP_UNION(X, wCharVal)
#define XP_UNICODETEXT(X)  XP_UNION(X, pszwText)
#define XP_TCHAR(X)        XP_UNION(X, tCharVal)
#define XP_TTEXT(X)        XP_UNION(X, ptszText)
#define XP_INT1(X)         XP_UNION(X, byVal)
#define XP_INT2(X)         XP_UNION(X, iVal)
#define XP_INT4(X)         XP_UNION(X, lVal)
#define XP_INT8(X)         XP_UNION(X, llVal)
#define XP_FLT4(X)         XP_UNION(X, fltval)
#define XP_FLT8(X)         XP_UNION(X, dblVal)
#define XP_DECIMAL(X)      XP_UNION(X, decVal)
#define XP_MONEY4(X)       XP_UNION(X, mny4Val)
#define XP_MONEY(X)        XP_UNION(X, mnyVal)
#define XP_DATETIM4(X)     XP_UNION(X, dt4Val)
#define XP_DATETIME(X)     XP_UNION(X, dtVal)
#define XP_NUMERIC(X)      XP_UNION(X, numericVal)
#define XP_BINARY(X)       XP_UNION(X, byBinaryVal)
#define XP_VARBINARY(X)    XP_UNION(X, varBinaryVal)
#define XP_GUID(X)         XP_UNION(X, guidValVal)
#define XP_IMAGE(X)        XP_UNION(X, ImageVal)


//class which represents a fleshed out parameter of an XP at runtime
class XP_PARAMETER_DATA
{
public:
//Constructors / Destructors
  XP_PARAMETER_DATA();
  ~XP_PARAMETER_DATA();

//member variables
  BYTE                 m_bType;                   //The data type of this parameter 
  ULONG                m_cbMaxLen;                //The maximum length in bytes of this parameter
  ULONG                m_cbActualLen;             //The actual length in bytes of this parameter 
  BOOL                 m_fNull;                   //TRUE if this parameter is null, otherwise FALSE 
  BYTE*                m_pData;                   //The actual data for this parameter, 
  BOOL                 m_bActualOutputParameter;  //TRUE if the parameter IS an output parameter, FALSE implies an input parameter
  XP_VARIANT_PARAMETER m_Data;                    //The actual parameter data in a VARIANT like structure
};


//Class which encapsulates one XP
class CExtendedStoredProcedure
{
public:
//Constructors / Destructors
  CExtendedStoredProcedure();
  virtual ~CExtendedStoredProcedure();

//Virtual functions
  virtual SRVRETCODE main(SRV_PROC* pSrvProc, DBCHAR* szFunctionName); //The main entry point for XPs

//Other methods
#ifdef _DEBUG
  static void CheckForHeapLeaks(_CrtMemState* pBeginMemState);         //Checks and reports on Heap memory leaks
#endif

//Wrapper Methods for the ODS API
  int         Convert(int srctype, void* src, DBINT srclen, int desttype, void* dest, DBINT destlen) const;
  static BOOL WillConvert(int srctype, int desttype);
  int         Describe(int colnumber, DBCHAR* column_name, int namelen, DBINT desttype, DBINT destlen, DBINT srctype, DBINT srclen, void* srcdata) const;
  int         GetBindToken(char* bindtoken) const;
  int         MessageHandler(int errornum, BYTE severity, BYTE state, int oserrnum, char* errtext, int errtextlen, char* oserrtext, int oserrtextlen) const;
  int         ParamInfo(int n, BYTE* pbType, ULONG* pcbMaxLen, ULONG* pcbActualLen, BYTE* pbData, BOOL* pfNull) const;
  DBCHAR*     ParamName(int n, int* len = NULL) const;
  int         ParamNumber(DBCHAR* name, int namelen = SRV_NULLTERM) const;
  int         ParamSetOutput(int n, BYTE* pbData, ULONG cbLen, BOOL fNull) const;
  int         ParamStatus(int n) const;
  BOOL        IsReturnParam(int n) const;
  DBCHAR*     PField(int field, int* len = NULL) const;
  void*       PFieldEx(int field, int* len = NULL) const;
  ULONG       SessionMessageLCID() const;
  WCHAR*      InstanceName() const;
  int         ParamCount() const;
  int         SendDone(DBUSMALLINT status, DBUSMALLINT info, DBINT count) const; 
  int         SendErrorMsg(DBCHAR* message, int msglen = SRV_NULLTERM, DBINT msgnum = 0, DBTINYINT Class = 0, DBTINYINT state = 1, DBUSMALLINT linenum = 0, DBCHAR* rpcname = NULL, int rpcnamelen = 0) const;
  int         SendInfoMsg(DBCHAR* message, int msglen = SRV_NULLTERM, DBINT msgnum = 0, DBTINYINT Class = 0, DBTINYINT state = 1, DBUSMALLINT linenum = 0, DBCHAR* rpcname = NULL, int rpcnamelen = 0) const;
  int         SendUnicodeMsg(WCHAR* message, int msglen = SRV_NULLTERM, int msgnum = 0, BYTE severity = 0) const;
  int         SendRow() const;
  int         SetColumnData(int column, void* data) const;
  int         SetColumnLength(int column, int len) const;
  int         SetUserDefinedType(int column, DBINT user_type) const;
  BOOL        GotAttention() const;
  int         GetDTCXact(void** ppv) const;
  int         ImpersonateClient() const;
  int         RevertToSelf() const;
    
protected:
//Member variables
  SRV_PROC*                m_pSrvProc;                                                                       //The client connection handle  
  DBCHAR*                  m_szFunctionName;                                                                 //The name of this XP
  XP_PARAMETER_DEFINITION* m_pParameterDefinitions;                                                          //The array of parameter definitions for this XP
  int                      m_nParameters;                                                                    //the number of parameters for this XP (corresponds to the size of the 2 array member variables)
  XP_PARAMETER_DATA*       m_pParameterData;                                                                 //The array of fleshed out runtime parameters

//Other Methods
  virtual SRVRETCODE Run() = 0;                                                                                    //The function derived classes should override to implement their specific functionality
  virtual SRVRETCODE SetupExpectedParameters() { return 0; };                                                      //Derived classes should use this function to setup m_pExpectedParameters, return 0 to indicate success, anything else to indicate failure
  virtual SRVRETCODE VerifyAndInitializeParameters();                                                              //The function which does verification of parameters as setup in m_pExpectedParameters, return 0 to indicate success, anything else to indicate failure
  virtual SRVRETCODE AutoConvertParameters();                                                                      //The function which does auto conversion of parameters, return 0 to indicate success, anything else to indicate failure
  virtual void       DisplayUsage();                                                                               //Function which gets called to didplay usage of the XP
  virtual SRVRETCODE SetExpectedParameters(const XP_PARAMETER_DEFINITION* pParameterDefinitions, int nParameters); //Allows you to specify the parameters for this XP. Used by END_IMPLEMENT_XP_PARAMETERS, return 0 to indicate success, anything else to indicate failure
};



///////////////////// Other Macros / Defines ////////////////////////////////////////

#define IMPLEMENT_VERSION_XP() \
  extern "C" __declspec(dllexport) ULONG __GetXpVersion() \
	{ \
		return ODS_VERSION; \
	}

#ifdef _DEBUG
  #define XP_MEM_STATE(memState) \
    _CrtMemState memState; \
    _CrtMemCheckpoint(&memState);

  #define XP_MEM_DIFF(memState) \
    CExtendedStoredProcedure::CheckForHeapLeaks(&memState);

#else
  #define XP_MEM_STATE(memState)
  #define XP_MEM_DIFF(memState)

#endif

#define IMPLEMENT_XP(xpName, class) \
  extern "C" SRVRETCODE __declspec(dllexport) xpName(SRV_PROC* srvproc) \
  { \
    XP_MEM_STATE(beginMemState) \
    \
    SRVRETCODE retCode = 0; \
    { \
      class _xp; \
      retCode = _xp.main(srvproc, #xpName); \
    } \
    \
    XP_MEM_DIFF(beginMemState) \
    \
    return retCode; \
  }

#define IMPLEMENT_XP_NO_MEMCHECK(xpName, class) \
  extern "C" SRVRETCODE __declspec(dllexport) xpName(SRV_PROC* srvproc) \
  { \
    class _xp; \
    return _xp.main(srvproc, #xpName); \
  }

#define DECLARE_XP_PARAMETERS() \
  static XP_PARAMETER_DEFINITION sm_ExpectedParams[]; \
  virtual SRVRETCODE SetupExpectedParameters();

#define BEGIN_IMPLEMENT_XP_PARAMETERS(class) \
  XP_PARAMETER_DEFINITION class::sm_ExpectedParams[] = \
  { 

#define END_IMPLEMENT_XP_PARAMETERS(class) \
  }; \
  SRVRETCODE class::SetupExpectedParameters() \
  { \
    return SetExpectedParameters(sm_ExpectedParams, sizeof(sm_ExpectedParams) / sizeof(XP_PARAMETER_DEFINITION)); \
  }

#define XP_INPUT_PARAMETER(ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType) \
  { XP_PARAMETER_DEFINITION::IN_PARAMETER, ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, DEFAULTPRECISION, DEFAULTSCALE },

#define XP_INPUT_PARAMETER_PRECISION_SCALE(ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, Precision, Scale) \
  { XP_PARAMETER_DEFINITION::IN_PARAMETER, ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, Precision, Scale },

#define XP_OUTPUT_PARAMETER(ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional) \
  { XP_PARAMETER_DEFINITION::OUT_PARAMETER, ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, XP_PT_EMPTY, DEFAULTPRECISION, DEFAULTSCALE },

#define XP_INPUT_OR_OUTPUT_PARAMETER(ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType) \
  { XP_PARAMETER_DEFINITION::IN_OUT_PARAMETER, ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, DEFAULTPRECISION, DEFAULTSCALE },

#define XP_INPUT_OR_OUTPUT_PARAMETER_PRECISION_SCALE(ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, Precision, Scale) \
  { XP_PARAMETER_DEFINITION::IN_OUT_PARAMETER, ColumnNumber, ColumnName, AllowableTypes, AllowNull, Optional, RuntimeType, Precision, Scale },

#ifdef _AFX
  #define XP_ASSERT(expr) ASSERT(expr)
#elif defined(__ATLBASE_H__)
  #define XP_ASSERT(expr) ATLASSERT(expr)
#else
  #include <assert.h>
  #define XP_ASSERT(expr) assert(expr)
#endif


const unsigned __int64 XP_ALLOW_TEXT         = 0x0000000000000001; 
const unsigned __int64 XP_ALLOW_GUID         = 0x0000000000000002; 
const unsigned __int64 XP_ALLOW_NCHAR	       = 0x0000000000000004;
const unsigned __int64 XP_ALLOW_VARBINARY    = 0x0000000000000008; 
const unsigned __int64 XP_ALLOW_VARCHAR      = 0x0000000000000010; 
const unsigned __int64 XP_ALLOW_BINARY       = 0x0000000000000020; 
const unsigned __int64 XP_ALLOW_IMAGE        = 0x0000000000000040; 
const unsigned __int64 XP_ALLOW_CHAR         = 0x0000000000000080; 
const unsigned __int64 XP_ALLOW_INT1         = 0x0000000000000100; 
const unsigned __int64 XP_ALLOW_BIT          = 0x0000000000000200; 
const unsigned __int64 XP_ALLOW_INT2         = 0x0000000000000400;  
const unsigned __int64 XP_ALLOW_DECIMAL	     = 0x0000000000000800;  
const unsigned __int64 XP_ALLOW_INT4         = 0x0000000000001000;  
const unsigned __int64 XP_ALLOW_DATETIM4     = 0x0000000000002000; 
const unsigned __int64 XP_ALLOW_FLT4         = 0x0000000000004000; 
const unsigned __int64 XP_ALLOW_MONEY        = 0x0000000000008000;  
const unsigned __int64 XP_ALLOW_DATETIME     = 0x0000000000010000;  
const unsigned __int64 XP_ALLOW_FLT8         = 0x0000000000020000;  
const unsigned __int64 XP_ALLOW_NUMERIC	     = 0x0000000000040000;  
const unsigned __int64 XP_ALLOW_SSVARIANT    = 0x0000000000080000;
const unsigned __int64 XP_ALLOW_NTEXT	       = 0x0000000000100000;
const unsigned __int64 XP_ALLOW_MONEY4       = 0x0000000000200000;
const unsigned __int64 XP_ALLOW_INT8         = 0x0000000000400000;
const unsigned __int64 XP_ALLOW_BIGVARBINARY = 0x0000000000800000;
const unsigned __int64 XP_ALLOW_BIGVARCHAR   = 0x0000000001000000;
const unsigned __int64 XP_ALLOW_BIGBINARY    = 0x0000000002000000;
const unsigned __int64 XP_ALLOW_BIGCHAR      = 0x0000000004000000;
const unsigned __int64 XP_ALLOW_NVARCHAR	   = 0x0000000008000000;
const unsigned __int64 XP_ALLOW_ANY      	   = 0xFFFFFFFFFFFFFFFF;


//Defintion of some XP++ specific error codes which can be returned from your XP
const int XP_ERROR_FAIL_VERIFY_AND_INITIALIZE_PARAMETERS = 1;
const int XP_ERROR_FAIL_AUTO_CONVERT_PARAMETERS          = 2;
const int XP_ERROR_FAIL_SET_EXPECTED_PARAMETERS          = 3;
const int XP_ERROR_LAST_ERROR                            = 3; //Client code can use this define to define their own custom error numbers which do not colide with XP++


#endif //__XP_PP_H__
