#include "stdafx.h"
#include "md5.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	8000-16
#define LEN_HASH	16 //hash length - 16 bytes

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) XP_MD5(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC XP_MD5 data, hash OUTPUT", srvproc);//send usage
}

RETCODE __declspec(dllexport) XP_MD5(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);

  //error if !=2 parameters entered
  if (iParamCount != 2)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, 2) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterData[MAX_LEN_INPUT];
  memset(inputParameterData,0,MAX_LEN_INPUT); 
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }

  //read parameter 2's attributes
  BYTE  bType2;       //data type of parameter
  ULONG cbMaxLen2;    //maximum length of parameter
  ULONG cbActualLen2; //actual length of parameter
  BOOL  fNull2;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 2
  if (!ValidInputDataTypes(bType2))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 
  
  //output variable must hold at least 16 bytes
  if (cbActualLen2 > LEN_HASH)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 must be 16 bytes.", srvproc);
    return XP_ERROR;
  }

  //main routine
  md5_context ctx;
  unsigned char output[16];
  md5_starts(&ctx);
  md5_update(&ctx, (uint8 *)inputParameterData, cbActualLen);
  md5_finish(&ctx, output);

  //return data in output parameter
  srv_paramsetoutput(srvproc, 2, (unsigned char *)output, 16, 0);
 
  return XP_NOERROR; //We got this far so return success
}