#include "stdafx.h"
#include "rijndael.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	8000-16
#define MAX_LEN_KEY	32 //maximum length key - 32 bytes

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) XP_RIJNDAEL(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC XP_RIJNDAEL data, key, dataEncrypted OUTPUT", srvproc);//send usage
}

RETCODE __declspec(dllexport) XP_RIJNDAEL(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);

  //error if !=3 parameters entered
  if (iParamCount != 3)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, 2) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, 3) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 3 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterData[MAX_LEN_INPUT];
  memset(inputParameterData,0,MAX_LEN_INPUT); 
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }


  //read parameter 2's attributes
  BYTE  bType2;       //data type of parameter
  ULONG cbMaxLen2;    //maximum length of parameter
  ULONG cbActualLen2; //actual length of parameter
  BOOL  fNull2;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 2
  if (!ValidInputDataTypes(bType2))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 2 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen2 > MAX_LEN_KEY)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 2 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull2)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterKey[MAX_LEN_KEY];
  if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, (BYTE*) inputParameterKey, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.", srvproc);
    return XP_ERROR;
  }


  //read parameter 3's attributes
  BYTE  bType3;       //data type of parameter
  ULONG cbMaxLen3;    //maximum length of parameter
  ULONG cbActualLen3; //actual length of parameter
  BOOL  fNull3;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 3, &bType3, &cbMaxLen3, &cbActualLen3, NULL, &fNull3) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 3.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 3
  if (!ValidInputDataTypes(bType3))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 3 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //main routine

  Rijndael Rdata;
  int outputLen=cbActualLen + 16;
  unsigned char output[8000];
  Rdata.init(Rijndael::CBC,Rijndael::Encrypt,(const unsigned char *)inputParameterKey,Rijndael::Key32Bytes);
  int len = Rdata.padEncrypt((const unsigned char *)inputParameterData,outputLen*8,output);
  if(len < 0)
  {
    ShowErrorAndUsageMsg("Error! Could not initialize crypto engine.", srvproc);
    return XP_ERROR;
  }

  //return data in output parameter
  srv_paramsetoutput(srvproc, 3, (unsigned char *)output, len/8, 0);
 
  return XP_NOERROR; //We got this far so return success
}