#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	20 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_cap1b(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC xp_cap1 data, dataCapitalized OUTPUT. dataCapitalized will have the first character in upper case and the rest in lower case", srvproc);//send usage
}

RETCODE __declspec(dllexport) xp_cap1b(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);
  
  //error if !=2 parameters entered
  if (iParamCount != 2)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }
  
  //Get the parameter position of the named parameters
  int intParam1 = srv_paramnumber(srvproc, "@data", SRV_NULLTERM);
  int intParam2 = srv_paramnumber(srvproc, "@dataCapitalized", SRV_NULLTERM);

  //verify parameters' names
  if (intParam1 == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 'data' not found.", srvproc);
    return XP_ERROR;
  }
  if (intParam2 == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 'dataCapitalized' not found.", srvproc);
    return XP_ERROR;
  }
  
  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, intParam1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error! Parameter 'data' should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, intParam2) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error! Parameter 'dataCapitalized' should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 'data''s attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, intParam1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data'.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 'data'
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 'data' must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 'data' (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 'data' is too big.", srvproc);
    return XP_ERROR;
  }  

  //Reject parameter 'data' if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 'data' cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterData[MAX_LEN_INPUT];
  if (srv_paraminfo(srvproc, intParam1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data'.", srvproc);
    return XP_ERROR;
  }

  //read parameter 'dataCapitalized''s attributes
  BYTE  bType2;       //data type of parameter
  ULONG cbMaxLen2;    //maximum length of parameter
  ULONG cbActualLen2; //actual length of parameter
  BOOL  fNull2;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, intParam2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'dataCapitalized'.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 'dataCapitalized'
  if (!ValidInputDataTypes(bType2))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 'dataCapitalized' must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //truncate string if the output is shorter than the input
  if (cbActualLen > cbMaxLen2)
    cbActualLen = cbMaxLen2; //set input length equal to output

  //main routine
  CharLowerBuffA(inputParameterData, cbActualLen); //change to lower case
  inputParameterData[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData[0]); //turn 1st character into upper case

  //return data in output parameter
  srv_paramsetoutput(srvproc, intParam2, (unsigned char *)inputParameterData, cbActualLen, 0);
  
  return XP_NOERROR;//We got this far so return success
}

