#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	20 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_cap2(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC xp_cap2 data OUTPUT. data will have the first character in upper case and the rest in lower case", srvproc); //send usage
}

RETCODE __declspec(dllexport) xp_cap2(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);
  
  //error if !=1 parameters entered
  if (iParamCount != 1)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterData[MAX_LEN_INPUT];
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }

  //main routine
  CharLowerBuffA(inputParameterData, cbActualLen); //change to lower case
  inputParameterData[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData[0]); //turn 1st character into upper case

  //return data in output parameter
  srv_paramsetoutput(srvproc, 1, (unsigned char *)inputParameterData, cbActualLen, 0);
 
  return XP_NOERROR; //We got this far so return success
}