#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	20 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_cap3(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc); //send error message
  XPprint("Usage: EXEC xp_cap3 data OUTPUT, [MoreData]. data is concatenated with MoreData and both will have the first character in upper case and the rest in lower case", srvproc); //send usage
}

RETCODE __declspec(dllexport) xp_cap3(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);
  
  //error if we do not have 1 or 2 parameters
  if ((iParamCount != 1) && (iParamCount != 2))
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }  

  //check for default second parameter
  bool param2Default = false;
  //Parameter 2 is default if there is only 1 parameter or the DEFAULT keyword is used
  if ((iParamCount == 1) || ((srv_paramstatus(srvproc, 2) & SRV_PARAMDEFAULT) == SRV_PARAMDEFAULT))
    param2Default = true;

  //validate first parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read first parameter's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.", srvproc);
    return XP_ERROR;
  }

  //Reject the first parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  char inputParameterData[MAX_LEN_INPUT + 1];
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }
  inputParameterData[cbActualLen] = '\0'; //NULL terminate the data now that we have received its data

  //the second parameter is not read if it is default
  DBCHAR inputParameterData2[MAX_LEN_INPUT + 1];
  ULONG cbActualLen2; //actual length of parameter
  if (param2Default)
  { 
    char* pszDefaultParameter2 = "Zoe";
    strcpy(inputParameterData2, pszDefaultParameter2); //the default value is 'Zoe'
    cbActualLen2 = strlen(pszDefaultParameter2);
  }
  else
  {
    //read second parameter's attributes
    BYTE  bType2;       //data type of parameter
    ULONG cbMaxLen2;    //maximum length of parameter
    BOOL  fNull2;       //true if parameter is NULL
    if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  	{
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.",srvproc);
      return XP_ERROR;
	  }

    //validate second parameter's I/O type
    if ((srv_paramstatus(srvproc, 2) & SRV_PARAMRETURN) == 1)
	  {
      ShowErrorAndUsageMsg("Error! Parameter 2 should not be an output parameter.",srvproc);
      return XP_ERROR;
	  }

    //accept only char data for second parameter
    if (!ValidInputDataTypes(bType2))
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 2 must be ASCII data.", srvproc);
      return XP_ERROR;
	  }

    //Reject second parameter if it is NULL
    if (fNull)
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 2 cannot be null.", srvproc);
      return XP_ERROR;
	  }

    //avoid buffer overflow in parameter 2 (because this sample uses simple static sized stack arrays to hold the parameter data)
    if (cbActualLen2 > MAX_LEN_INPUT)
    {
      ShowErrorAndUsageMsg("Error!, The size of parameter 2 is too big.", srvproc);
      return XP_ERROR;
    }

    //Get the second parameter's data
    if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, (BYTE*) inputParameterData2, &fNull2) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.", srvproc);
      return XP_ERROR;
	  }
    inputParameterData2[cbActualLen2] = '\0'; //NULL terminate the data now that we have received its data
  }

  //main routine
  char OutputParameterData[(MAX_LEN_INPUT*2) + 1]; //Allow double the input size for the local variable
                                                   //even though eventually it may get truncated. This means
                                                   //we avoid having to write complicated logic before we 
                                                   //concatenate the strings together

  //concatenate the parameters' data
  strcpy(OutputParameterData, inputParameterData);
  _strlwr(OutputParameterData); //change to lower case
  OutputParameterData[0] = (char) CharUpperA((LPSTR) (WORD) OutputParameterData[0]); //turn 1st character into upper case
  if (cbActualLen2 > 0)
  {
    _strlwr(inputParameterData2); //change to lower case
    inputParameterData2[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData2[0]); //turn 1st character into upper case

    strcat(OutputParameterData, " ");
    strcat(OutputParameterData, inputParameterData2);
  }  

  //truncate string if the length is bigger than the stipulated maximum
  if (strlen(OutputParameterData) > MAX_LEN_INPUT)
	  OutputParameterData[MAX_LEN_INPUT] = '\0';

  //return data in output parameter
  srv_paramsetoutput(srvproc, 1, (unsigned char *)OutputParameterData, strlen(OutputParameterData), 0);

  return XP_NOERROR; //We got this far so return success
}

