#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	20 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif

RETCODE __declspec(dllexport) xp_cap5(SRV_PROC* srvproc);

#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc); //send error message
  XPprint("Usage: EXEC xp_cap5 [data1 OUTPUT], [data2 OUTPUT], [data3 OUTPUT]. Each optional parameter will have the first character in upper case and the rest in lower case", srvproc); //send usage
}

RETCODE __declspec(dllexport) xp_cap5(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);

  //error if 1>n>3 parameters entered
  if ((iParamCount<1) || (iParamCount>3))
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.", srvproc);
    return XP_ERROR;
  }
 
  //get the parameter position of the named parameters
  int intParam1 = srv_paramnumber(srvproc, "@data1", SRV_NULLTERM);
  int intParam2 = srv_paramnumber(srvproc, "@data2", SRV_NULLTERM);
  int intParam3 = srv_paramnumber(srvproc, "@data3", SRV_NULLTERM);  

  //read parameter data1's attributes
  char inputParameterData[MAX_LEN_INPUT];
  ULONG cbActualLen = 0; //actual length of parameter
  if (intParam1 != 0)
  {
    //validate parameters' IO type
    if ((srv_paramstatus(srvproc, intParam1) & SRV_PARAMRETURN) == 0)
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data1' should not be an input parameter.", srvproc);
      return XP_ERROR;
	  }

    //read parameter 'data1''s attributes
    BYTE  bType;       //data type of parameter
    ULONG cbMaxLen;    //maximum length of parameter
    BOOL  fNull;       //true if parameter is NULL
    if (srv_paraminfo(srvproc, intParam1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data1'.", srvproc);
      return XP_ERROR;
	  }

    //accept only char data for parameter 'data1'
    if (!ValidInputDataTypes(bType))
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data1' must be ASCII data.", srvproc);
      return XP_ERROR;
	  } 

    //avoid buffer overflow in parameter 'data1' (because this sample uses simple static sized stack arrays to hold the parameter data)
    if (cbActualLen > MAX_LEN_INPUT)
    {
      ShowErrorAndUsageMsg("Error!, The size of parameter 'data1' is too big.", srvproc);
      return XP_ERROR;
    }  

    //Reject parameter 'data1' if it is NULL
    if (fNull)
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data1' cannot be null.", srvproc);
      return XP_ERROR;
	  }

    //Get the input parameter data
    if (srv_paraminfo(srvproc, intParam1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data1'.", srvproc);
      return XP_ERROR;
	  }
  }

  //read parameter data2 if entered
  char inputParameterData2[MAX_LEN_INPUT];
  ULONG cbActualLen2 = 0;  //actual length of parameter
  if (intParam2 != 0)
  {
    //validate parameters' IO type
    if ((srv_paramstatus(srvproc, intParam2) & SRV_PARAMRETURN) == 0)
  	{
      ShowErrorAndUsageMsg("Error! Parameter 'data2' should not be an input parameter.", srvproc);
      return XP_ERROR;
	  }

    //read parameter 'data2''s attributes
    BYTE  bType2;        //data type of parameter
    ULONG cbMaxLen2;     //maximum length of parameter
    BOOL  fNull2;        //true if parameter is NULL
    if (srv_paraminfo(srvproc, intParam2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data2'.", srvproc);
      return XP_ERROR;
	  }

    //accept only char data for parameter 'data2'
    if (!ValidInputDataTypes(bType2))
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data2' must be ASCII data.", srvproc);
      return XP_ERROR;
	  } 

    //avoid buffer overflow in parameter 'data2' (because this sample uses simple static sized stack arrays to hold the parameter data)
    if (cbActualLen2 > MAX_LEN_INPUT)
    {
      ShowErrorAndUsageMsg("Error!, The size of parameter 'data2' is too big.", srvproc);
      return XP_ERROR;
    }  

    //Reject parameter 'data2' if it is NULL
    if (fNull2)
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data2' cannot be null.", srvproc);
      return XP_ERROR;
	  }

    //Get the input parameter data
    if (srv_paraminfo(srvproc, intParam2, &bType2, &cbMaxLen2, &cbActualLen2, (BYTE*) inputParameterData2, &fNull2) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data2'.", srvproc);
      return XP_ERROR;
	  }
  }

  //read parameter 3
  char inputParameterData3[MAX_LEN_INPUT];
  ULONG cbActualLen3 = 0; //actual length of parameter
  if (intParam3 != 0)
  {
    //validate parameters' IO type
    if ((srv_paramstatus(srvproc, intParam3) & SRV_PARAMRETURN) == 0)
  	{
      ShowErrorAndUsageMsg("Error! Parameter 'data3' should not be an input parameter.",srvproc);
      return XP_ERROR;
	  }

    //read parameter 'data3''s attributes
    BYTE  bType3;       //data type of parameter
    ULONG cbMaxLen3;    //maximum length of parameter
    BOOL  fNull3;       //true if parameter is NULL
    if (srv_paraminfo(srvproc, intParam3, &bType3, &cbMaxLen3, &cbActualLen3, NULL, &fNull3) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data3'.",srvproc);
      return XP_ERROR;
	  }

    //accept only char data for parameter 'data3'
    if (!ValidInputDataTypes(bType3))
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data3' must be ASCII data.", srvproc);
      return XP_ERROR;
	  } 

    //avoid buffer overflow in parameter 'data1' (because this sample uses simple static sized stack arrays to hold the parameter data)
    if (cbActualLen3 > MAX_LEN_INPUT)
    {
      ShowErrorAndUsageMsg("Error!, The size of parameter 'data3' is too big.", srvproc);
      return XP_ERROR;
    }  

    //Reject parameter 'data3' if it is NULL
    if (fNull3)
	  {
      ShowErrorAndUsageMsg("Error!, Parameter 'data3' cannot be null.", srvproc);
      return XP_ERROR;
	  }

    //Get the input parameter data
    if (srv_paraminfo(srvproc, intParam3, &bType3, &cbMaxLen3, &cbActualLen3, (BYTE*) inputParameterData3, &fNull3) == FAIL)
	  {
      ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 'data3'.", srvproc);
      return XP_ERROR;
	  }
  }

  //main routine
  //Parameter 1
  if (intParam1 != 0)
  {
    CharLowerBuffA(inputParameterData, cbActualLen); //change to lower case
    inputParameterData[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData[0]); //turn 1st character into upper case

    //return data in output parameter
    srv_paramsetoutput(srvproc, 1, (unsigned char *)inputParameterData, cbActualLen, 0);
  }

  //Parameter 2
  if (intParam2 != 0)
  {
    CharLowerBuffA(inputParameterData2, cbActualLen2); //change to lower case
    inputParameterData2[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData2[0]); //turn 1st character into upper case

    //return data in output parameter
    srv_paramsetoutput(srvproc, 2, (unsigned char *)inputParameterData2, cbActualLen2, 0);
  }

  //Parameter 3
  if (intParam3 != 0)
  {
    CharLowerBuffA(inputParameterData3, cbActualLen3); //change to lower case
    inputParameterData3[0] = (char) CharUpperA((LPSTR) (WORD) inputParameterData3[0]); //turn 1st character into upper case

    //return data in output parameter
    srv_paramsetoutput(srvproc, 3, (unsigned char *)inputParameterData3, cbActualLen3, 0);
  }

  return XP_NOERROR; //We got this far so return success
}

