#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	40 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_cap7(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept char and nchar data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNVARCHAR || bType == SRVNCHAR || bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//validate nchar data
inline BOOL ValidUnicode(BYTE bType)
{
  return (bType == SRVNVARCHAR || bType == SRVNCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC xp_cap7 data, dataCapitalized OUTPUT. dataCapitalized will have the first character in upper case and the rest in lower case", srvproc);//send usage
}

RETCODE __declspec(dllexport) xp_cap7(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);
    
  //error if !=2 parameters entered
  if (iParamCount != 2)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, 2) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be char data.", srvproc);
    return XP_ERROR;
  } 

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  BYTE inputParameterData[MAX_LEN_INPUT]; //We allocate exactly just the bytes we need since all the subsequent functions do not assume that the data is null terminated
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }

  //read parameter 2's attributes
  BYTE  bType2;       //data type of parameter
  ULONG cbMaxLen2;    //maximum length of parameter
  ULONG cbActualLen2; //actual length of parameter
  BOOL  fNull2;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 2
  if (!ValidInputDataTypes(bType2))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 must be char data.", srvproc);
    return XP_ERROR;
  } 

  //main routine
  wchar_t ConvertedData[4000];
  int nConverted = srv_convert(srvproc, bType, inputParameterData, cbActualLen, SRVNCHAR, ConvertedData, -1);
  if (nConverted == -1)
  {
    ShowErrorAndUsageMsg("Error!, Failed to convert Parameter 1's data to UNICODE.", srvproc);
    return XP_ERROR;
  }
  CharLowerBuffW(ConvertedData, nConverted);          //change to lower case
  ConvertedData[0] = (wchar_t) CharUpperW((LPWSTR) ConvertedData[0]); //turn 1st character into upper case

  char OutputData[8000];
  nConverted = srv_convert(srvproc, SRVNCHAR, ConvertedData, nConverted, bType2, OutputData, -1);
  if (nConverted == -1)
  {
    ShowErrorAndUsageMsg("Error!, Failed to convert the output data.", srvproc);
    return XP_ERROR;
  }

  //truncate string if the output is shorter than the input
  if (((ULONG) nConverted) > cbMaxLen2)
    nConverted = cbMaxLen2; //set input length equal to output

  //return data in output parameter
  srv_paramsetoutput(srvproc, 2, (unsigned char *)OutputData, nConverted, 0);

  return XP_NOERROR; //We got this far so return success
}