#include "stdafx.h"

#define XP_NOERROR	 0 //return code - no error
#define XP_ERROR	   1 //return code - error
#define MAXCOLDATA	20 //maximum length of column data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_hello2(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc); //send error message
  XPprint("Usage: EXEC xp_hello2 name. name must be ASCII data, and a maximum of 20 characters in size", srvproc); //send usage
}

RETCODE __declspec(dllexport) xp_hello2(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);

  //error if !=1 parameters entered
  if (iParamCount != 1)
  {
    ShowErrorAndUsageMsg("Error! Wrong number of parameters.",srvproc);
    return XP_ERROR;
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error! Parameter 1 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.", srvproc);
    return XP_ERROR;
  }  

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAXCOLDATA)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  DBCHAR inputParameterData[MAXCOLDATA + 1];
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }
  inputParameterData[cbActualLen] = '\0'; //NULL terminate the data now that we have received its data

  //Return the output data as a message
  DBCHAR OutputData[MAXCOLDATA + 8];//corresponds to +6 characters for "Hello ", an extra 1 for the trailing "!" character and finally 1 more for the trailing NULL
  wsprintfA(OutputData, "Hello %s!", inputParameterData);

  //send message
  XPprint(OutputData, srvproc);
    
  //We got this far so return success
  return XP_NOERROR;
}