#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	8000 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_palindrome(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only boolean data, even if NULL
inline BOOL ValidOutputDataTypes(BYTE bType)
{
  return (bType == SRVBIT || bType == SRVBITN);
}

//validate nchar data
inline BOOL ValidUnicode(BYTE bType)
{
  return (bType == SRVNVARCHAR || bType == SRVNCHAR);
}

//validate all numeric data
inline BOOL ValidNumber(BYTE bType)
{
  return (bType == SRVDECIMALN || bType == SRVNUMERICN || bType == SRVFLTN || bType == SRVMONEYN || bType == SRVINTN || bType == SRVINT1 || bType == SRVINT2 || bType == SRVINT4 || bType == SRVINT8 || bType == SRVDECIMAL || bType == SRVNUMERIC || bType == SRVFLT4 || bType == SRVFLT8 || bType == SRVMONEY || bType == SRVMONEY4);
}

//validate char data
inline BOOL ValidASCII(BYTE bType)
{
  return (bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC xp_palindrome data, result OUTPUT. result will be 1 if data is a palindrome", srvproc);//send usage
}

//returns 1 if the input is a palindrome
BYTE palindrome(const char* InputString, int lenInput)
{
	for (int i=0; i<lenInput/2; i++)
  {
		if (InputString[i] != InputString[lenInput-i-1])
			return 0;
	}
	return 1;
}

//returns 1 if the input is a palindrome (Unicode)
BYTE palindromeN(const wchar_t* InputString, int lenInput)
{
	for (int i=0; i<lenInput/2; i++)
  {
		if (InputString[i] != InputString[lenInput-i-1])
			return 0;
	}
	return 1;
}

RETCODE __declspec(dllexport) xp_palindrome(SRV_PROC* srvproc)
{
  //get number of parameters
  int iParamCount = srv_rpcparams(srvproc);
    
  //error if !=2 parameters entered
  if (iParamCount != 2)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return (XP_ERROR);
  }

  //validate parameter's I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an output parameter.",srvproc);
    return XP_ERROR;
  }
  if ((srv_paramstatus(srvproc, 2) & SRV_PARAMRETURN) == 0)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 should not be an input parameter.",srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }

  //avoid buffer overflow in parameter 1 (because this sample uses simple static sized stack arrays to hold the parameter data)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.", srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  DBCHAR inputParameterData[MAX_LEN_INPUT + 1];
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (BYTE*) inputParameterData, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.", srvproc);
    return XP_ERROR;
  }
  inputParameterData[cbActualLen] = '\0'; //NULL terminate the data now that we have received its data

  //read parameter 2's attributes
  BYTE  bType2;       //data type of parameter
  ULONG cbMaxLen2;    //maximum length of parameter
  ULONG cbActualLen2; //actual length of parameter
  BOOL  fNull2;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 2, &bType2, &cbMaxLen2, &cbActualLen2, NULL, &fNull2) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 2.",srvproc);
    return XP_ERROR;
  }

  //accept only bit data for parameter 2
  if (!ValidOutputDataTypes(bType2))
  {
    ShowErrorAndUsageMsg("Error!, Parameter 2 must be BIT.", srvproc);
    return XP_ERROR;
  } 

  //main routine
  BYTE byResult = 0;
  if (ValidASCII(bType))
    byResult = palindrome((const char*)inputParameterData, cbActualLen);
  else if (ValidUnicode(bType))
    byResult = palindromeN((const wchar_t*)inputParameterData, cbActualLen/2);
  else
  {
    //Check to see if we can convert the input parameter to a vchar
    if (srv_willconvert(bType, SRVCHAR))
	  {
	    char ConvertedData[8000];
		  memset(ConvertedData, 0, 8000);
      srv_convert(srvproc, bType, inputParameterData, cbActualLen, SRVCHAR, ConvertedData, -1);
		  if (ValidNumber(bType)) //numbers have variable length
        cbActualLen = strlen(ConvertedData);
		  byResult = palindrome((const char*)ConvertedData, cbActualLen);
	  }
  }

  //return data in output parameter
  srv_paramsetoutput(srvproc, 2, &byResult, 1, 0);
 
  return XP_NOERROR; //We got this far so return success
}