/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPReadFile.h"


/////////////////////////////// Implementation ////////////////////////////////

#define FILENAME_PARAMETER_INDEX   0
#define DATABUFFER_PARAMETER_INDEX 1
#define OFFSET_PARAMETER_INDEX     2

BEGIN_IMPLEMENT_XP_PARAMETERS(CReadFileStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_ANY,                                                                                                   TRUE, FALSE)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE, XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CReadFileStoredProcedure)

void CReadFileStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SAVEFILE @FileToReadFrom @DataToRead [@OffsetToReadFrom]");
}

BOOL CReadFileStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get the name of the file to read from (already setup for us in CExtendedStoredProcedure::VerifyAndInitializeParameters)
  TCHAR* pszFilename = XP_TTEXT(m_pParameterData[FILENAME_PARAMETER_INDEX].m_Data);
  XP_ASSERT(pszFilename != NULL);

  //Open the specified file
  HANDLE hFile = CreateFile(pszFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile != INVALID_HANDLE_VALUE)
  {
    //Seek to the specified offest if present
    if (m_pParameterData[OFFSET_PARAMETER_INDEX].m_bType != SRVNULL)
    {
      DBINT SeekPosition = XP_INT4(m_pParameterData[OFFSET_PARAMETER_INDEX].m_Data);
      if (SetFilePointer(hFile, SeekPosition, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
      {
        //Close the file before we return
        CloseHandle(hFile);

        sprintf(szErrorMsg, "%s: Failed to seek to the specified file position %d in %s, Error:%d", m_szFunctionName, SeekPosition, pszFilename, GetLastError());
        SendErrorMsg(szErrorMsg);
        return FALSE;
      }
    }

    //And read the data from disk
    DWORD dwBytesRead = 0;
    ReadFile(hFile, m_pParameterData[DATABUFFER_PARAMETER_INDEX].m_pData, m_pParameterData[DATABUFFER_PARAMETER_INDEX].m_cbMaxLen, &dwBytesRead, NULL);

    //Don't forget to close the file before we return
    CloseHandle(hFile);

    //Return the data in the output parameter
	  ParamSetOutput(2, m_pParameterData[DATABUFFER_PARAMETER_INDEX].m_pData, dwBytesRead, FALSE);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Could not open the file %s, Error:%d", m_szFunctionName, pszFilename, GetLastError());
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }

  //We got here, so the running of the XP was successful
  return TRUE;
}
