/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPSaveFile.h"


/////////////////////////////// Implementation ////////////////////////////////

#define FILENAME_PARAMETER_INDEX   0
#define DATABUFFER_PARAMETER_INDEX 1
#define OFFSET_PARAMETER_INDEX     2

BEGIN_IMPLEMENT_XP_PARAMETERS(CSaveFileStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_ANY,                                                                                                   FALSE, FALSE, XP_PT_VARBINARY)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE,  XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CSaveFileStoredProcedure)

void CSaveFileStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SAVEFILE @FileToSaveTo @DataToSave [@OffsetToWriteTo]");
}

BOOL CSaveFileStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get the name of the file to save to (already setup for us in CExtendedStoredProcedure::VerifyAndInitializeParameters)
  TCHAR* pszFilename = XP_TTEXT(m_pParameterData[FILENAME_PARAMETER_INDEX].m_Data);
  XP_ASSERT(pszFilename);

  //Open the specified file
  HANDLE hFile = CreateFile(pszFilename, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile != INVALID_HANDLE_VALUE)
  {
    //Seek to the specified offest if present
    if (m_pParameterData[OFFSET_PARAMETER_INDEX].m_bType != SRVNULL)
    {
      DBINT SeekPosition = XP_INT4(m_pParameterData[OFFSET_PARAMETER_INDEX].m_Data);
      if (SetFilePointer(hFile, SeekPosition, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
      {
        //Close the file before we return
        CloseHandle(hFile);

        sprintf(szErrorMsg, "%s: Failed to seek to the specified file position %d in %s, Error:%d", m_szFunctionName, SeekPosition, pszFilename, GetLastError());
        SendErrorMsg(szErrorMsg);
        return FALSE;
      }
    }

    //And write the data to disk
    DWORD dwBytesWritten = 0;
    WriteFile(hFile, XP_VARBINARY(m_pParameterData[DATABUFFER_PARAMETER_INDEX].m_Data).pbyData, XP_VARBINARY(m_pParameterData[DATABUFFER_PARAMETER_INDEX].m_Data).cbSize, &dwBytesWritten, NULL);

    //Don't forget to close the file before we return
    CloseHandle(hFile);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Could not open the file %s, Error:%d", m_szFunctionName, pszFilename, GetLastError());
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }

  //We got here, so the running of the XP was successful
  return TRUE;
}
