/*
*************************************************************************************************
*     __  __  ____    _           ____     ___    _         ____     ___     ___     ___        *
*     \ \/ / |  _ \  ( )  ___    / ___|   / _ \  | |       |___ \   / _ \   / _ \   / _ \       *
*      \  /  | |_) | |/  / __|   \___ \  | | | | | |         __) | | | | | | | | | | | | |      *
*      /  \  |  __/      \__ \    ___) | | |_| | | |___     / __/  | |_| | | |_| | | |_| |      *
*     /_/\_\ |_|         |___/   |____/   \__\_\ |_____|   |_____|  \___/   \___/   \___/       *
*                                                                                               *
*                                                                                               *
*    Module : XP_PP.CPP                                                                         *
*                                                                                               *
*    Purpose: Implementation for a wrapper class framework for SQL Server Extended Stored       *
*             Procedures                                                                        *
*                                                                                               *
*    History: None                                                                              *
*                                                                                               *
*    Copyright (c) 2004 by Joseph Gama and PJ Naughter.                                         *
*                                                                                               *
*************************************************************************************************
*/



//////////////// Includes /////////////////////////////////////////////////////

#include "stdafx.h"
#ifndef _NEW_
#pragma message("To avoid this message please put new into your PCH (normally stdafx.h)")
#include <new>
#endif
#include "xp_pp.h"



//////////////// Macros / Defines /////////////////////////////////////////////

//Automatically link to the ODS library
#pragma comment(lib, "Opends60.lib")



//////////////// Implementation ///////////////////////////////////////////////

XP_VARIANT_PARAMETER::XP_VARIANT_PARAMETER() : pt(XP_PT_EMPTY)
{
}

XP_VARIANT_PARAMETER::~XP_VARIANT_PARAMETER()
{
  Clear();
}

BOOL XP_VARIANT_PARAMETER::Construct(CExtendedStoredProcedure& XP, XP_VARIANT_PARAMETER_TYPE ToType, BYTE CurrentType, BYTE* pCurrentData, ULONG cbCurrentSize, BYTE Precision, BYTE Scale)
{
  Clear(); //Clear up anything which is currently in this object

  BOOL bSuccess = TRUE; //Assume the best

  //Set the type
  pt = ToType;

  switch (ToType)
  {
    case XP_PT_BIT:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVBIT, &__XP_VARIANT_PARAMETER_NAME_1.bVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.bVal)) != -1);
      break;
    }
#ifndef _UNICODE
    case XP_PT_TCHAR: //Deliberate fallthrough
#endif
    case XP_PT_ASCIICHAR:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVCHAR, &__XP_VARIANT_PARAMETER_NAME_1.cCharVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.cCharVal)) != -1);
      break;
    }
#ifndef _UNICODE
    case XP_PT_TTEXT: //Deliberate fallthrough
#endif
    case XP_PT_ASCIITEXT:
    {
      int nNewSizeInLogicalCharacters = max(100, cbCurrentSize + 1);

      try
      {
        __XP_VARIANT_PARAMETER_NAME_1.pszText = new char[nNewSizeInLogicalCharacters];
        if (__XP_VARIANT_PARAMETER_NAME_1.pszText)
        {
          int nConverted = XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVTEXT, __XP_VARIANT_PARAMETER_NAME_1.pszText, nNewSizeInLogicalCharacters-1);
          if (nConverted != -1)
            __XP_VARIANT_PARAMETER_NAME_1.pszText[nConverted] = '\0'; //Null terminate the data
          bSuccess = (nConverted != -1);
        }
        else
          bSuccess = FALSE;
      }
#ifdef _AFX
      catch(CMemoryException* pEx)
      {
        pEx->Delete();
        bSuccess = FALSE;
      }
#endif
      catch(std::bad_alloc&)
      {
        bSuccess = FALSE;
      }
      break;
    }
#ifdef _UNICODE
    case XP_PT_TCHAR: //Deliberate fallthrough
#endif
    case XP_PT_UNICODECHAR:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVNCHAR, &__XP_VARIANT_PARAMETER_NAME_1.wCharVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.wCharVal)) != -1);
      break;
    }
#ifdef _UNICODE
    case XP_PT_TTEXT: //Deliberate fallthrough
#endif
    case XP_PT_UNICODETEXT:
    {
      int nNewSizeInLogicalCharacters = max(100, cbCurrentSize + 1);
      try
      {
        __XP_VARIANT_PARAMETER_NAME_1.pszwText = new wchar_t[nNewSizeInLogicalCharacters];
        if (__XP_VARIANT_PARAMETER_NAME_1.pszwText)
        {
          int nConverted = XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVNTEXT, __XP_VARIANT_PARAMETER_NAME_1.pszwText, nNewSizeInLogicalCharacters*sizeof(wchar_t) - sizeof(wchar_t));
          if (nConverted != -1)
            __XP_VARIANT_PARAMETER_NAME_1.pszwText[nConverted/2] = '\0'; //Null terminate the data
          bSuccess = (nConverted != -1);
        }
        else
          bSuccess = FALSE;
      }
#ifdef _AFX
      catch(CMemoryException* pEx)
      {
        pEx->Delete();
        bSuccess = FALSE;
      }
#endif
      catch(std::bad_alloc&)
      {
        bSuccess = FALSE;
      }
      break;
    }
    case XP_PT_INT1:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVINT1, &__XP_VARIANT_PARAMETER_NAME_1.byVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.byVal)) != -1);
      break;
    }
    case XP_PT_INT2:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVINT2, &__XP_VARIANT_PARAMETER_NAME_1.iVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.iVal)) != -1);
      break;
    }
    case XP_PT_INT4:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVINT4, &__XP_VARIANT_PARAMETER_NAME_1.lVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.lVal)) != -1);
      break;
    }
    case XP_PT_INT8:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVINT8, &__XP_VARIANT_PARAMETER_NAME_1.llVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.llVal)) != -1);
      break;
    }
    case XP_PT_FLT4:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVFLT4, &__XP_VARIANT_PARAMETER_NAME_1.fltVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.fltVal)) != -1);
      break;
    }
    case XP_PT_FLT8:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVFLT8, &__XP_VARIANT_PARAMETER_NAME_1.dblVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.dblVal)) != -1);
      break;
    }
    case XP_PT_DECIMAL:
    {
      __XP_VARIANT_PARAMETER_NAME_1.decVal.precision = Precision; 
      __XP_VARIANT_PARAMETER_NAME_1.decVal.scale = Scale;
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVDECIMAL, &__XP_VARIANT_PARAMETER_NAME_1.decVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.decVal)) != -1);
      break;
    }
    case XP_PT_MONEY4:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVMONEY4, &__XP_VARIANT_PARAMETER_NAME_1.mny4Val, sizeof(__XP_VARIANT_PARAMETER_NAME_1.mny4Val)) != -1);
      break;
    }
    case XP_PT_MONEY:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVMONEY, &__XP_VARIANT_PARAMETER_NAME_1.mnyVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.mnyVal)) != -1);
      break;
    }
    case XP_PT_DATETIM4:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVDATETIM4, &__XP_VARIANT_PARAMETER_NAME_1.dt4Val, sizeof(__XP_VARIANT_PARAMETER_NAME_1.dt4Val)) != -1);
      break;
    }
    case XP_PT_DATETIME:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVDATETIME, &__XP_VARIANT_PARAMETER_NAME_1.dtVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.dtVal)) != -1);
      break;
    }
    case XP_PT_NUMERIC:
    {
      __XP_VARIANT_PARAMETER_NAME_1.numericVal.precision = Precision;
      __XP_VARIANT_PARAMETER_NAME_1.numericVal.scale = Scale;
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVNUMERIC, &__XP_VARIANT_PARAMETER_NAME_1.numericVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.numericVal)) != -1);
      break;
    }
    case XP_PT_BINARY:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVBINARY, &__XP_VARIANT_PARAMETER_NAME_1.byBinaryVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.byBinaryVal)) != -1);
      break;
    }
    case XP_PT_VARBINARY:
    {
      __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.cbSize = cbCurrentSize;
      try
      {
        __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData = new BYTE[__XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.cbSize];
        if (__XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData)
          bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVBINARY, __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData, __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.cbSize) != -1);
        else
          bSuccess = FALSE;
      }
#ifdef _AFX
      catch(CMemoryException* pEx)
      {
        pEx->Delete();
        bSuccess = FALSE;
      }
#endif
      catch(std::bad_alloc&)
      {
        bSuccess = FALSE;
      }
      break;
    }
    case XP_PT_GUID:
    {
      bSuccess = (XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVGUID, &__XP_VARIANT_PARAMETER_NAME_1.guidVal, sizeof(__XP_VARIANT_PARAMETER_NAME_1.guidVal)) != -1);
      break;
    }
    case XP_PT_IMAGE:
    {
      __XP_VARIANT_PARAMETER_NAME_1.ImageVal.cbSize = max(100, cbCurrentSize);
      try
      {
        __XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData = new BYTE[__XP_VARIANT_PARAMETER_NAME_1.ImageVal.cbSize];
        if (__XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData)
        {
          int nConverted = XP.Convert(CurrentType, pCurrentData, cbCurrentSize, SRVIMAGE, __XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData, __XP_VARIANT_PARAMETER_NAME_1.ImageVal.cbSize);
          if (nConverted != -1)
            __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.cbSize = nConverted;
          bSuccess = (nConverted != -1);
        }
        else
          bSuccess = FALSE;
      }
#ifdef _AFX
      catch(CMemoryException* pEx)
      {
        pEx->Delete();
        bSuccess = FALSE;
      }
#endif
      catch(std::bad_alloc&)
      {
        bSuccess = FALSE;
      }
      break;
    }
    case XP_PT_EMPTY: //Deliberate fallthrough
    default:
    {
      break;
    }
  }

  //Clear if not successful
  if (!bSuccess)
    Clear();

  return bSuccess;
}

void XP_VARIANT_PARAMETER::Clear()
{
  switch (pt)
  {
    case XP_PT_ASCIITEXT:
    {
      if (__XP_VARIANT_PARAMETER_NAME_1.pszText)
        delete [] __XP_VARIANT_PARAMETER_NAME_1.pszText;
      __XP_VARIANT_PARAMETER_NAME_1.pszText = NULL;
      break;
    }
    case XP_PT_UNICODETEXT:
    {
      if (__XP_VARIANT_PARAMETER_NAME_1.pszwText)
        delete [] __XP_VARIANT_PARAMETER_NAME_1.pszwText;
      __XP_VARIANT_PARAMETER_NAME_1.pszwText = NULL;
      break;
    }
    case XP_PT_TTEXT:
    {
      if (__XP_VARIANT_PARAMETER_NAME_1.ptszText)
        delete [] __XP_VARIANT_PARAMETER_NAME_1.ptszText;
      __XP_VARIANT_PARAMETER_NAME_1.ptszText = NULL;
      break;
    }
    case XP_PT_VARBINARY:
    {
      if (__XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData)
        delete [] __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData;
      __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.pbyData = NULL;
      __XP_VARIANT_PARAMETER_NAME_1.varBinaryVal.cbSize = 0;
      break;
    }
    case XP_PT_IMAGE:
    {
      if (__XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData)
        delete [] __XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData;
      __XP_VARIANT_PARAMETER_NAME_1.ImageVal.pbyData = NULL;
      __XP_VARIANT_PARAMETER_NAME_1.ImageVal.cbSize = 0;
      break;
    }
    default:
    {
      break;
    }
  }

  pt = XP_PT_EMPTY;
}


XP_PARAMETER_DATA::XP_PARAMETER_DATA() : m_bType(SRVNULL), m_cbMaxLen(0), m_cbActualLen(0), m_fNull(FALSE), m_pData(NULL), m_bActualOutputParameter(FALSE)
{
}

XP_PARAMETER_DATA::~XP_PARAMETER_DATA()
{
  if (m_pData)
  {
    delete [] m_pData;
    m_pData = NULL;
  }
}


CExtendedStoredProcedure::CExtendedStoredProcedure() : m_pSrvProc(NULL), m_szFunctionName(NULL), m_pParameterDefinitions(NULL), m_nParameters(0), m_pParameterData(NULL)
{
}

CExtendedStoredProcedure::~CExtendedStoredProcedure()
{
  if (m_pParameterData)
  {
    delete [] m_pParameterData;
    m_pParameterData = NULL;
  }
  if (m_pParameterDefinitions)
  {
    delete [] m_pParameterDefinitions;
    m_pParameterDefinitions = NULL;
  }
}

void CExtendedStoredProcedure::DisplayUsage()
{
  //Display a simple help message, Derived classes should implement this to provide more meaningful help
  SendInfoMsg("The developer of this Extended Stored Procedure needs to implement their own version of the DisplayUsage method");
}

BOOL CExtendedStoredProcedure::VerifyAndInitializeParameters()
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);
  XP_ASSERT(m_pParameterDefinitions);
  XP_ASSERT(m_pParameterData);
  XP_ASSERT(m_szFunctionName);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[1024];

  //Verify and initialize each parameter
  for (int i=0; i<m_nParameters; i++)
  {
    if (m_pParameterDefinitions[i].m_nColumn == 0)
    {
      //If the parameter does not have a column number, then it should be named
      XP_ASSERT(m_pParameterDefinitions[i].m_szName);
      XP_ASSERT(strlen(m_pParameterDefinitions[i].m_szName));

      //Get the column number for the named parameter
      m_pParameterDefinitions[i].m_nColumn = ParamNumber(m_pParameterDefinitions[i].m_szName);
    }

    //If the specified parameter is present, then get its details
    if (m_pParameterDefinitions[i].m_nColumn)
    {
      //Get the info for the parameter
      if ((ParamInfo(m_pParameterDefinitions[i].m_nColumn, &m_pParameterData[i].m_bType, &m_pParameterData[i].m_cbMaxLen, &m_pParameterData[i].m_cbActualLen, NULL, &m_pParameterData[i].m_fNull) == FAIL)) 
      {
        if (!m_pParameterDefinitions[i].m_bOptional)
        {
          if (i == 0)
          {
            //If we failed on the first parameter, then just display the usage
            DisplayUsage();
          }
          else
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Failed in call to srv_paraminfo for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Failed in call to srv_paraminfo for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
          }
          return FALSE;
        }
      }
      else
      {
        //Flesh out each parameter

        //Allocate some heap memory for the parameters data
        XP_ASSERT(m_pParameterData[i].m_pData == NULL);
        try
        {
          m_pParameterData[i].m_pData = new BYTE[max(m_pParameterData[i].m_cbActualLen, m_pParameterData[i].m_cbMaxLen)];
          if (m_pParameterData[i].m_pData == NULL)
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }
        }
  #ifdef _AFX
        catch(CMemoryException* pEx)
        {
          pEx->Delete();

          if (m_pParameterDefinitions[i].m_szName)
            sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
          else
            sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
          SendErrorMsg(szErrorMsg);
          return FALSE;
        }
  #endif
        catch(std::bad_alloc&)
        {
          if (m_pParameterDefinitions[i].m_szName)
            sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
          else
            sprintf(szErrorMsg, "%s: Failed to allocate memory for the data for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
          SendErrorMsg(szErrorMsg);
          return FALSE;
        }

        //Get the actual parameter data if it is not currently NULL
        if (!m_pParameterData[i].m_fNull) 
        {
          if (ParamInfo(m_pParameterDefinitions[i].m_nColumn, &m_pParameterData[i].m_bType, &m_pParameterData[i].m_cbMaxLen, &m_pParameterData[i].m_cbActualLen, m_pParameterData[i].m_pData, &m_pParameterData[i].m_fNull) == FAIL)
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Failed in call to srv_paraminfo when getting the data for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Failed in call to srv_paraminfo when getting the data for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }
        }

        if (m_pParameterData[i].m_bType != SRVNULL)
        {
          //Check if the parameter should be [IN] or [OUT]
          m_pParameterData[i].m_bActualOutputParameter = IsReturnParam(m_pParameterDefinitions[i].m_nColumn);
          if (m_pParameterDefinitions[i].m_ParameterType == XP_PARAMETER_DEFINITION::OUT_PARAMETER && !m_pParameterData[i].m_bActualOutputParameter)
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Parameter %s[%d] should be an output parameter", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Parameter %d should be an output parameter", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }
          else if (m_pParameterDefinitions[i].m_ParameterType == XP_PARAMETER_DEFINITION::IN_PARAMETER && m_pParameterData[i].m_bActualOutputParameter)
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Parameter %s[%d] should be an input parameter", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Parameter %d should be an input parameter", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }

          //Check if null data is allowed for this parameter
          if (m_pParameterData[i].m_fNull && (m_pParameterDefinitions[i].m_bAllowNull == FALSE))
          {
            if (m_pParameterDefinitions[i].m_szName)
              sprintf(szErrorMsg, "%s: Parameter %s[%d] does not accept null data", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
            else
              sprintf(szErrorMsg, "%s: Parameter %d does not accept null data", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }

          //Check the allowable types for this parameter
          switch (m_pParameterData[i].m_bType)
          {
            case SRVTEXT:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_TEXT) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVTEXT", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVTEXT", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVGUID:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_GUID) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVGUID", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVGUID", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVVARBINARY:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_VARBINARY) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVVARBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVVARBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            } 
            case SRVINTN:
            {
              switch (m_pParameterData[i].m_cbMaxLen)
              {
                case sizeof(DBTINYINT):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT1) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT1", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT1", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                case sizeof(DBSMALLINT):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT2) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT2", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT2", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                case sizeof(DBINT):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT4) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                default:
                {
                  break;
                }
              }
              break;
            }
            case SRVVARCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_VARCHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVVARCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVVARCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBINARY:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BINARY) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVIMAGE:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_IMAGE) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVIMAGE", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVIMAGE", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_CHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVINT1:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT1) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT1", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT1", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBIT:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIT) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBIT", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBIT", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVINT2:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT2) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT2", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT2", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVDECIMAL:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DECIMAL) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDECIMAL", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDECIMAL", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVINT4:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT4) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVDATETIM4:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DATETIM4) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDATETIM4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDATETIM4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVFLT4:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_FLT4) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVFLT4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVFLT4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVMONEY:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_MONEY) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVMONEY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVMONEY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVDATETIME:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DATETIME) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDATETIME", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDATETIME", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVFLT8:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_FLT8) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVFLT8", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVFLT8", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVNUMERIC:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_NUMERIC) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVNUMERIC", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVNUMERIC", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVSSVARIANT:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_SSVARIANT) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVVSSVARIANT", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVVSSVARIANT", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVNTEXT:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_NTEXT) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVNTEXT", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVNTEXT", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBITN:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIT) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBITN", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBITN", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVDECIMALN:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DECIMAL) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDECIMALN", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDECIMALN", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVNUMERICN:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_NUMERIC) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVNUMERICN", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVNUMERICN", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVFLTN:
            {
              switch (m_pParameterData[i].m_cbMaxLen)
              {
                case sizeof(DBFLT4):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_FLT4) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVFLT4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVFLT4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                case sizeof(DBFLT8):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_FLT8) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVFLT8", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVFLT8", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                default:
                {
                  break;
                }
              }
              break;
            }
            case SRVMONEYN:
            {
              switch (m_pParameterData[i].m_cbMaxLen)
              {
                case sizeof(DBMONEY4):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_MONEY4) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVMONEY4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVMONEY4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                case sizeof(DBMONEY):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_MONEY) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVMONEY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVMONEY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                default:
                {
                  break;
                }
              }
              break;
            }
            case SRVDATETIMN:
            {
              switch (m_pParameterData[i].m_cbMaxLen)
              {
                case sizeof(DBDATETIM4):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DATETIM4) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDATETIM4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDATETIM4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                case sizeof(DBDATETIME):
                {
                  if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_DATETIME) == 0)
                  {
                    if (m_pParameterDefinitions[i].m_szName)
                      sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVDATETIME", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                    else
                      sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVDATETIME", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                    SendErrorMsg(szErrorMsg);
                    return FALSE;
                  }
                  break;
                }
                default:
                {
                  break;
                }
              }
              break;
            }
            case SRVMONEY4:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_MONEY4) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVMONEY4", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVMONEY4", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVINT8:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_INT8) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVINT8", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVINT8", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBIGVARBINARY:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIGVARBINARY) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBIGVARBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBIGVARBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBIGVARCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIGVARCHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBIGVARCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBIGVARCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBIGBINARY:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIGBINARY) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBIGBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBIGBINARY", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVBIGCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_BIGCHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVBIGCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVBIGCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVNVARCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_NVARCHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVNVARCHARL", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVNVARCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            case SRVNCHAR:
            {
              if ((m_pParameterDefinitions[i].m_AllowableTypes & XP_ALLOW_NCHAR) == 0)
              {
                if (m_pParameterDefinitions[i].m_szName)
                  sprintf(szErrorMsg, "%s: Parameter %s[%d] is not allowed to be SRVNCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
                else
                  sprintf(szErrorMsg, "%s: Parameter %d is not allowed to be SRVNCHAR", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
                SendErrorMsg(szErrorMsg);
                return FALSE;
              }
              break;
            }
            default:
            {
              if (m_pParameterDefinitions[i].m_szName)
                sprintf(szErrorMsg, "%s: Parameter %s[%d] is an unknown data type: Type: %d", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn, m_pParameterData[i].m_bType);
              else
                sprintf(szErrorMsg, "%s: Parameter %d is an unknown data type: Type: %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn, m_pParameterData[i].m_bType);
              SendErrorMsg(szErrorMsg);
              return FALSE;

              break;
            }
          }
        }
      }
    }
    else
    {
      //Specified parameter has not been provided, check to see if it is required
      if (m_pParameterDefinitions[i].m_bOptional == FALSE)
      {
        //Validate the parameters
        XP_ASSERT(m_pParameterDefinitions[i].m_szName);
        XP_ASSERT(strlen(m_pParameterDefinitions[i].m_szName));

        sprintf(szErrorMsg, "%s: Parameter %s is required but was not provided", m_szFunctionName, m_pParameterDefinitions[i].m_szName);
        SendErrorMsg(szErrorMsg);

        //Also allow the display usage function to be called since there was a gross error in the parameters sent to us
        DisplayUsage();

        return FALSE;
      }
    }
  }

  return TRUE;
}

BOOL CExtendedStoredProcedure::AutoConvertParameters()
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);
  XP_ASSERT(m_pParameterDefinitions);
  XP_ASSERT(m_pParameterData);
  XP_ASSERT(m_szFunctionName);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[1024];

  //initialize each parameter
  for (int i=0; i<m_nParameters; i++)
  {
    //do the auto conversion if required
    if (m_pParameterDefinitions[i].m_nColumn && (m_pParameterData[i].m_bType != SRVNULL) && (!m_pParameterData[i].m_fNull) && (m_pParameterDefinitions[i].m_RuntimeType != XP_PT_EMPTY))
    {
      if (!m_pParameterData[i].m_Data.Construct(*this, m_pParameterDefinitions[i].m_RuntimeType, m_pParameterData[i].m_bType, m_pParameterData[i].m_pData, m_pParameterData[i].m_cbActualLen, m_pParameterDefinitions[i].m_precision, m_pParameterDefinitions[i].m_scale))
      {
        if (m_pParameterDefinitions[i].m_szName)
          sprintf(szErrorMsg, "%s: Failed when converting the data for parameter %s[%d]", m_szFunctionName, m_pParameterDefinitions[i].m_szName, m_pParameterDefinitions[i].m_nColumn);
        else
          sprintf(szErrorMsg, "%s: Failed when converting the data for parameter %d", m_szFunctionName, m_pParameterDefinitions[i].m_nColumn);
        SendErrorMsg(szErrorMsg);
        return FALSE;
      }
    }
  }

  return TRUE;
}

BOOL CExtendedStoredProcedure::SetExpectedParameters(const XP_PARAMETER_DEFINITION* pParameterDefinitions, int nParameters) 
{ 
  //Validate the parameters
  XP_ASSERT(pParameterDefinitions);
  XP_ASSERT(nParameters);
  XP_ASSERT(m_pParameterDefinitions == NULL);
  XP_ASSERT(m_pParameterData == NULL);

  //We need to take a copy of the parameters sent to us because by default 'DECLARE_XP_PARAMETERS' sets up 
  //a static member variable to contain the parameter array. If we were to use this directly, this would 
  //cause multithreading problems when two instances of the XP are run from different threads by SQL Server
  //and then both try to modify values in their parameters array which in fact is shared between them.

  m_nParameters = nParameters;

  try
  {
    m_pParameterDefinitions = new XP_PARAMETER_DEFINITION[nParameters];
    if (m_pParameterDefinitions == NULL)
      return FALSE;
  }
#ifdef _AFX
  catch(CMemoryException* pEx)
  {
    pEx->Delete();
    return FALSE;
  }
#endif
  catch(std::bad_alloc&)
  {
    return FALSE;
  }

  try
  {
    m_pParameterData = new XP_PARAMETER_DATA[nParameters];
    if (m_pParameterData == NULL)
    {
      delete [] m_pParameterDefinitions;
      m_pParameterDefinitions = NULL;
      return FALSE;
    }
  }
#ifdef _AFX
  catch(CMemoryException* pEx)
  {
    pEx->Delete();
    delete [] m_pParameterDefinitions;
    m_pParameterDefinitions = NULL;
    return FALSE;
  }
#endif
  catch(std::bad_alloc&)
  {
    delete [] m_pParameterDefinitions;
    m_pParameterDefinitions = NULL;
    return FALSE;
  }

  memcpy(m_pParameterDefinitions, pParameterDefinitions, sizeof(XP_PARAMETER_DEFINITION) * nParameters);

  return TRUE;
}

#ifdef _DEBUG
void CExtendedStoredProcedure::CheckForHeapLeaks(_CrtMemState* pBeginMemState)
{
  //Verify the integrity of the heap as it stands currently
  XP_ASSERT(_CrtCheckMemory());

  //Get the current memory state
  _CrtMemState endMemState;
  _CrtMemCheckpoint(&endMemState);

  //Check to see if there is a difference
  _CrtMemState diffMemState;
  int nMemDiff = _CrtMemDifference(&diffMemState, pBeginMemState, &endMemState);
  if (nMemDiff)
  {
    _CrtDumpMemoryLeaks();

    XP_ASSERT(FALSE);  //If this ASSERT occurs, then you should check you heap allocations for memory leaks.
                       //A good way to find where these leaks occur would be to use _CrtSetBreakAlloc in 
                       //combination with the details as reported to the debug output window by 
                       //_CrtDumpMemoryLeaks above. Please see the MSDN for further information on using 
                       //the MS Debug C Runtime
  }
}
#endif

SRVRETCODE CExtendedStoredProcedure::main(SRV_PROC* pSrvProc, DBCHAR* szFunctionName)
{
  //Validate the parameters
  XP_ASSERT(pSrvProc);
  XP_ASSERT(szFunctionName);

  //Save away the parameters as member variables
  m_pSrvProc = pSrvProc;
  m_szFunctionName = szFunctionName;

  //call the virtual function to setup the expected parameters
  SetupExpectedParameters();

  //Verify the parameters if we have some
  if (m_nParameters)
  {
    if ((m_pParameterDefinitions == NULL) || (m_pParameterData == NULL))
      return 1;  

    if (!VerifyAndInitializeParameters())
      return 1;

    if (!AutoConvertParameters())
      return 1;
  }

  //Call the virtual function which does the real work of the XP
  return Run() ? 0 : 1;  
}

int CExtendedStoredProcedure::Convert(int srctype, void* src, DBINT srclen, int desttype, void* dest, DBINT destlen) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_convert(m_pSrvProc, srctype, src, srclen, desttype, dest, destlen);
}

BOOL CExtendedStoredProcedure::WillConvert(int srctype, int desttype)
{
  //Call the SDK version
  return srv_willconvert(srctype, desttype);
}

int CExtendedStoredProcedure::Describe(int colnumber, DBCHAR* column_name, int namelen, DBINT desttype, DBINT destlen, DBINT srctype, DBINT srclen, void* srcdata) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_describe(m_pSrvProc, colnumber, column_name, namelen, desttype, destlen, srctype, srclen, srcdata);
}

int CExtendedStoredProcedure::GetBindToken(char* bindtoken) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_getbindtoken(m_pSrvProc, bindtoken);
}

int CExtendedStoredProcedure::MessageHandler(int errornum, BYTE severity, BYTE state, int oserrnum, char* errtext, int errtextlen, char* oserrtext, int oserrtextlen) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_message_handler(m_pSrvProc, errornum, severity, state, oserrnum, errtext, errtextlen, oserrtext, oserrtextlen);
}

int CExtendedStoredProcedure::ParamInfo(int n, BYTE* pbType, ULONG* pcbMaxLen, ULONG* pcbActualLen, BYTE* pbData, BOOL* pfNull) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_paraminfo(m_pSrvProc, n, pbType, pcbMaxLen, pcbActualLen, pbData, pfNull);
}

DBCHAR* CExtendedStoredProcedure::ParamName(int n, int* len) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_paramname(m_pSrvProc, n, len);
}

int CExtendedStoredProcedure::ParamNumber(DBCHAR* name, int namelen) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_paramnumber(m_pSrvProc, name, namelen);
}

int CExtendedStoredProcedure::ParamSetOutput(int n, BYTE* pbData, ULONG cbLen, BOOL fNull) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_paramsetoutput(m_pSrvProc, n, pbData, cbLen, fNull);
}

int CExtendedStoredProcedure::ParamStatus(int n) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_paramstatus(m_pSrvProc, n);
}

BOOL CExtendedStoredProcedure::IsReturnParam(int n) const
{
  //Call the ParamStatus helper function
  return (ParamStatus(n) & 0x1) == 1;
}

DBCHAR* CExtendedStoredProcedure::PField(int field, int* len) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_pfield(m_pSrvProc, field, len);
}

void* CExtendedStoredProcedure::PFieldEx(int field, int* len) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_pfieldex(m_pSrvProc, field, len);
}

ULONG CExtendedStoredProcedure::SessionMessageLCID() const
{
  //Call the helper function PFieldEx
  ULONG nReturn = 0;
  void* pFieldEx = PFieldEx(SRV_MSGLCID);
  if (pFieldEx)
    nReturn = *((ULONG*) pFieldEx);

  return nReturn;
}

WCHAR* CExtendedStoredProcedure::InstanceName() const
{
  return ((WCHAR*) PFieldEx(SRV_INSTANCENAME));
}

int CExtendedStoredProcedure::ParamCount() const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_rpcparams(m_pSrvProc);
}

int CExtendedStoredProcedure::SendDone(DBUSMALLINT status, DBUSMALLINT info, DBINT count) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_senddone(m_pSrvProc, status, info, count);
}

int CExtendedStoredProcedure::SendErrorMsg(DBCHAR* message, int msglen, DBINT msgnum, DBTINYINT Class, DBTINYINT state, DBUSMALLINT linenum, DBCHAR* rpcname, int rpcnamelen) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_sendmsg(m_pSrvProc, SRV_MSG_ERROR, msgnum, Class, state, rpcname, rpcnamelen, linenum, message, msglen);
}

int CExtendedStoredProcedure::SendInfoMsg(DBCHAR* message, int msglen, DBINT msgnum, DBTINYINT Class, DBTINYINT state, DBUSMALLINT linenum, DBCHAR* rpcname, int rpcnamelen) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_sendmsg(m_pSrvProc, SRV_MSG_INFO, msgnum, Class, state, rpcname, rpcnamelen, linenum, message, msglen);
}

int CExtendedStoredProcedure::SendRow() const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_sendrow(m_pSrvProc);
}

int CExtendedStoredProcedure::SetColumnData(int column, void* data) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_setcoldata(m_pSrvProc, column, data);
}

int CExtendedStoredProcedure::SetColumnLength(int column, int len) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_setcollen(m_pSrvProc, column, len);
}

int CExtendedStoredProcedure::SetUserDefinedType(int column, DBINT user_type) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_setutype(m_pSrvProc, column, user_type);
}

int CExtendedStoredProcedure::SendUnicodeMsg(WCHAR* message, int msglen, int msgnum, BYTE severity) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_wsendmsg(m_pSrvProc, msgnum, severity, message, msglen);
}

BOOL CExtendedStoredProcedure::GotAttention() const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_got_attention(m_pSrvProc);
}

int CExtendedStoredProcedure::GetDTCXact(void** ppv) const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_getdtcxact(m_pSrvProc, ppv);
}

int CExtendedStoredProcedure::ImpersonateClient() const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_impersonate_client(m_pSrvProc);
}

int CExtendedStoredProcedure::RevertToSelf() const
{
  //Validate our variables
  XP_ASSERT(m_pSrvProc);

  //Call the SDK version
  return srv_revert_to_self(m_pSrvProc);
}
