/*
Module : AACOORDINATETRANSFORMATION.H
Purpose: Implementation for the algorithms which convert between the various celestial coordinate systems
Created: PJN / 29-12-2003
History: None


Copyright (c) 2003 - 2004 by PJ Naughter.  (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


#ifndef __AACOORDINATETRANSFORMATION_H__
#define __AACOORDINATETRANSFORMATION_H__


/////////////////////////////// Includes //////////////////////////////////////

#include "AA2DCoordinate.h"


/////////////////////////////// Classes ///////////////////////////////////////

class CAACoordinateTransformation
{
public:
//Conversion functions
  static CAA2DCoordinate Equatorial2Ecliptic(double Alpha, double Delta, double Epsilon);
	static CAA2DCoordinate Ecliptic2Equatorial(double Lambda, double Beta, double Epsilon);
	static CAA2DCoordinate Equatorial2Horizontal(double LocalHourAngle, double Delta, double Latitude);
	static CAA2DCoordinate Horizontal2Equatorial(double A, double h, double Latitude);
	static CAA2DCoordinate Equatorial2Galactic(double Alpha, double Delta);
	static CAA2DCoordinate Galactic2Equatorial(double l, double b);

  static double PI();
  static double DegreesToRadians(double Degrees);
  static double RadiansToDegrees(double Radians);
  static double RadiansToHours(double Radians);
  static double HoursToRadians(double Hours);
  static double HoursToDegrees(double Hours);
  static double DegreesToHours(double Degrees);
  static double DMSToDegrees(double Degrees, double Minutes, double Seconds, BOOL bPositive = TRUE);
  static double MapTo0To360Range(double Degrees);
  static double MapTo0To24Range(double HourAngle);
};

#endif //__AACOORDINATETRANSFORMATION_H__
