/*
Module : AADATE.H
Purpose: Implementation for the algorithms which convert between the Gregorian and Julian calendars and the Julian Day
Created: PJN / 29-12-2003
History: None


Copyright (c) 2003 - 2004 by PJ Naughter.  (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


#ifndef __AADATE_H__
#define __AADATE_H__


//////////////////////////// Macros / Defines /////////////////////////////////

struct CAACalendarDate
{
  long Year;
  long Month;
  long Day;
};


/////////////////////////// Classes ///////////////////////////////////////////

class CAADate
{
public:

//Enums
  enum DAY_OF_WEEK
  {	
    SUNDAY	  = 0,
	  MONDAY	  = 1,
	  TUESDAY	  = 2,
	  WEDNESDAY	= 3,
	  THURSDAY	= 4,
	  FRIDAY	  = 5,
	  SATURDAY	= 6
  };

//Constructors / Destructors
	CAADate();
  CAADate(long Year, long Month, double Day, BOOL bGregorianCalendar = TRUE);
  CAADate(long Year, long Month, double Day, double Hour = 0, double Minute = 0, double Second = 0, BOOL bGregorianCalendar = TRUE);
  CAADate(double JD, BOOL bGregorianCalendar = TRUE);

//Static Methods
  static double          Set(long Year, long Month, double Day, BOOL bGregorianCalendar);
  static BOOL            IsLeap(long Year, BOOL bGregorianCalendar);
  static void            DayOfYearToDayAndMonth(long DayOfYear, BOOL bLeap, long& DayOfMonth, long& Month);
  static CAACalendarDate JulianToGregorian(long Year, long Month, long Day);
  static CAACalendarDate GregorianToJulian(long Year, long Month, long Day);
  static long            INT(double value);

//Non Static methods
	double      Julian() const { return m_dblJulian; };
  BOOL        GregorianCalendar() const { return m_bGregorianCalendar; };
  operator    double() const { return m_dblJulian; };
	long        Day() const;
	long        Month() const;
	long        Year() const;
	long        Hour() const;
	long        Minute() const;
	double      Second() const;
  void        Set(long Year, long Month, double Day, double Hour = 0, double Minute = 0, double Second = 0, BOOL bGregorianCalendar = TRUE);
	void        Set(double JD, BOOL bGregorianCalendar = TRUE);
  void        Get(long& Year, long& Month, long& Day, long& Hour, long& Minute, double& Second) const;
	DAY_OF_WEEK DayOfWeek() const;
  double      DayOfYear() const;
	long        DaysInMonth() const;
 	long        DaysInYear() const;
  BOOL        Leap() const;
  double      FractionalYear() const;

protected:
//Member variables
  double m_dblJulian;          //Julian Day number for this date
  BOOL   m_bGregorianCalendar; //Is this date in the Gregorian Calendar as opposed to the Julian Calendar
};

#endif //__AADATE_H__
