/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_ASTROProcs.h"
#include "AA+\AADynamicalTime.h"
#include "AA+\AAEquinoxesAndSolstices.h"
#include "AA+\AAMoonPhases.h"



/////////////////////////////// Macros / Defines //////////////////////////////

const double JULIAN_1900                 = 2415020.5;
const double ONE3HUNDREDSECONDS_IN_A_DAY = 25920000.0;
const double MINUTES_IN_A_DAY            = 1440.0;



/////////////////////////////// Implementation ////////////////////////////////

#define DATE2JULIAN_DATE   0
#define DATE2JULIAN_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CDate2JulianExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_DATETIME | XP_ALLOW_DATETIM4, FALSE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CDate2JulianExtendedStoredProcedure)

BOOL CDate2JulianExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //What will be the return value
  BOOL bSuccess = FALSE;

  if (m_pParameterData[DATE2JULIAN_DATE].m_cbMaxLen == sizeof(DBDATETIME))
  {
    //Get the datetime value
    DBDATETIME* pDT = (DBDATETIME*) m_pParameterData[DATE2JULIAN_DATE].m_pData;
    XP_ASSERT(pDT);

    //Convert it to a Julian Day
    double dblJulian = JULIAN_1900 + pDT->dtdays + (pDT->dttime / ONE3HUNDREDSECONDS_IN_A_DAY);

	  bSuccess = (ParamSetOutput(DATE2JULIAN_JULIAN+1, (BYTE*) &dblJulian, sizeof(dblJulian), FALSE) != 0);
  }
  else
  {
    XP_ASSERT(m_pParameterData[DATE2JULIAN_DATE].m_cbMaxLen == sizeof(DBDATETIM4));

    //Get the datetim4 value
    DBDATETIM4* pDT = (DBDATETIM4*) m_pParameterData[DATE2JULIAN_DATE].m_pData;
    XP_ASSERT(pDT);

    //Convert it to a Julian Day
    double dblJulian = JULIAN_1900 + pDT->numdays + (pDT->nummins / MINUTES_IN_A_DAY);

	  bSuccess = (ParamSetOutput(DATE2JULIAN_JULIAN+1, (BYTE*) &dblJulian, sizeof(dblJulian), FALSE) != 0);
  }

  return bSuccess;
}

void CDate2JulianExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_DATE2JULIAN @Date, @Julian OUTPUT");
}


#define JULIAN2DATE_JULIAN 0
#define JULIAN2DATE_DATE   1

BEGIN_IMPLEMENT_XP_PARAMETERS(CJulian2DateExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_DATETIME | XP_ALLOW_DATETIM4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CJulian2DateExtendedStoredProcedure)

BOOL CJulian2DateExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //What will be the return value
  BOOL bSuccess = FALSE;

  //Get the input parameter
  double* pdblJulian = (double*) m_pParameterData[JULIAN2DATE_JULIAN].m_pData;
  XP_ASSERT(pdblJulian);

  if (m_pParameterData[JULIAN2DATE_DATE].m_cbMaxLen == sizeof(DBDATETIME))
  {
    //Form the output parameter
    DBDATETIME DT;
    DT.dtdays = (DBINT) (*pdblJulian - JULIAN_1900);
    double intDay;
    double fractionalDay = modf(*pdblJulian + 0.5, &intDay);
    DT.dttime = (ULONG) (fractionalDay * ONE3HUNDREDSECONDS_IN_A_DAY);

	  bSuccess = (ParamSetOutput(JULIAN2DATE_DATE+1, (BYTE*) &DT, sizeof(DT), FALSE) != 0);
  }
  else
  {
    XP_ASSERT(m_pParameterData[JULIAN2DATE_DATE].m_cbMaxLen == sizeof(DBDATETIM4));

    //Form the output parameter
    DBDATETIM4 DT;
    DT.numdays = (USHORT) (*pdblJulian - JULIAN_1900);
    double intDay;
    double fractionalDay = modf(*pdblJulian + 0.5, &intDay);
    DT.nummins = (USHORT) (fractionalDay * MINUTES_IN_A_DAY);

	  bSuccess = (ParamSetOutput(JULIAN2DATE_DATE+1, (BYTE*) &DT, sizeof(DT), FALSE) != 0);
  }

  return bSuccess;
}

void CJulian2DateExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_JULIAN2DATE @Julian, @Date OUTPUT");
}


#define DELTAT_JULIAN 0
#define DELTAT_DELTAT 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CDeltaTExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CDeltaTExtendedStoredProcedure)

BOOL CDeltaTExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  double* pdblJulian = (double*) m_pParameterData[JULIAN2DATE_JULIAN].m_pData;

  //Call the AA+ function to do the leg work
  double dblDeltaT = CAADynamicalTime::DeltaT(*pdblJulian);

  //Return the output parameter
  return (ParamSetOutput(DELTAT_DELTAT+1, (BYTE*) &dblDeltaT, sizeof(dblDeltaT), FALSE) != 0);
}

void CDeltaTExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_DELTAT @Julian, @DeltaT OUTPUT");
}


#define SPRINGEQUINOX_YEAR   0
#define SPRINGEQUINOX_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CSpringEquinoxExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CSpringEquinoxExtendedStoredProcedure)

BOOL CSpringEquinoxExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  DBINT lYear = XP_INT4(m_pParameterData[SPRINGEQUINOX_YEAR].m_Data);

  //Call the AA+ function to do the leg work
  double dblDate = CAAEquinoxesAndSolstices::SpringEquinox(lYear);

  //Return the output parameter
  return (ParamSetOutput(SPRINGEQUINOX_JULIAN+1, (BYTE*) &dblDate, sizeof(dblDate), FALSE) != 0);
}

void CSpringEquinoxExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_SPRING_EQUINOX @Year, @Date OUTPUT");
}



#define SUMMERSOLSTICE_YEAR   0
#define SUMEMRSOLSTICE_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CSummerSolsticeExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CSummerSolsticeExtendedStoredProcedure)

BOOL CSummerSolsticeExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  DBINT lYear = XP_INT4(m_pParameterData[SPRINGEQUINOX_YEAR].m_Data);

  //Call the AA+ function to do the leg work
  double dblDate = CAAEquinoxesAndSolstices::SummerSolstice(lYear);

  //Return the output parameter
  return (ParamSetOutput(SUMEMRSOLSTICE_JULIAN+1, (BYTE*) &dblDate, sizeof(dblDate), FALSE) != 0);
}

void CSummerSolsticeExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_SUMMER_SOLSTICE @Year, @Date OUTPUT");
}



#define AUTUMNEQUINOX_YEAR   0
#define AUTUMNEQUINOX_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CAutumnEquinoxExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CAutumnEquinoxExtendedStoredProcedure)

BOOL CAutumnEquinoxExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  DBINT lYear = XP_INT4(m_pParameterData[SPRINGEQUINOX_YEAR].m_Data);

  //Call the AA+ function to do the leg work
  double dblDate = CAAEquinoxesAndSolstices::AutumnEquinox(lYear);

  //Return the output parameter
  return (ParamSetOutput(AUTUMNEQUINOX_JULIAN+1, (BYTE*) &dblDate, sizeof(dblDate), FALSE) != 0);
}

void CAutumnEquinoxExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_AUTUMN_EQUINOX @Year, @Date OUTPUT");
}



#define WINTERSOLSTICE_YEAR   0
#define WINTERSOLSTICE_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CWinterSolsticeExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CWinterSolsticeExtendedStoredProcedure)

BOOL CWinterSolsticeExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  DBINT lYear = XP_INT4(m_pParameterData[SPRINGEQUINOX_YEAR].m_Data);

  //Call the AA+ function to do the leg work
  double dblDate = CAAEquinoxesAndSolstices::WinterSolstice(lYear);

  //Return the output parameter
  return (ParamSetOutput(WINTERSOLSTICE_JULIAN+1, (BYTE*) &dblDate, sizeof(dblDate), FALSE) != 0);
}

void CWinterSolsticeExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_WINTER_SOLSTICE @Year, @Date OUTPUT");
}



#define MOONPHASEK_YEAR 0
#define MOONPHASEK_K    1

BEGIN_IMPLEMENT_XP_PARAMETERS(CMoonPhaseKExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CMoonPhaseKExtendedStoredProcedure)

BOOL CMoonPhaseKExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  double* pdblYear = (double*) m_pParameterData[MOONPHASEK_YEAR].m_pData;

  //Call the AA+ function to do the leg work
  double K = CAAMoonPhases::K(*pdblYear);

  //Return the output parameter
  return (ParamSetOutput(MOONPHASEK_K+1, (BYTE*) &K, sizeof(K), FALSE) != 0);
}

void CMoonPhaseKExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_MOONPHASEK @Year, @K OUTPUT");
}



#define MOONPHASE_K      0
#define MOONPHASE_JULIAN 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CMoonPhaseExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CMoonPhaseExtendedStoredProcedure)

BOOL CMoonPhaseExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  double* pdblK = (double*) m_pParameterData[MOONPHASE_K].m_pData;

  //Call the AA+ function to do the leg work
  double JulianDate = CAAMoonPhases::TruePhase(*pdblK);

  //Return the output parameter
  return (ParamSetOutput(MOONPHASE_JULIAN+1, (BYTE*) &JulianDate, sizeof(JulianDate), FALSE) != 0);
}

void CMoonPhaseExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ASTRO_MOONPHASE @K, @Date OUTPUT");
}

