/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_CPUUSAGEProc.h"



/////////////////////////////// Macros / Defines //////////////////////////////

typedef NTSTATUS (WINAPI NTQUERYSYSTEMINFORMATION)(SYSTEM_INFORMATION_CLASS, PVOID, ULONG, PULONG);
typedef NTQUERYSYSTEMINFORMATION* LPNTQUERYSYSTEMINFORMATION;



/////////////////////////////// Implementation ////////////////////////////////

// The implementation of the actual XP_CPUUSAGE XP itself. Note that in an ideal world we
// could just use the Win32 SDK function GetSystemTimes to get the CPU Usage but this 
// function is only available on Windows Server 2003 or Windows XP SP1. We could use
// the performance counters in the registry or a PDH wrapper for it, but as an example
// and also to reduce the amount of code we require, lets use the semi-undocumented function
// "NtQuerySystemInformation".

#define CPUUSAGE_TIME_INTERVAL_INDEX 0
#define CPUUSAGE_USAGE_INDEX         1

BEGIN_IMPLEMENT_XP_PARAMETERS(CCPUUsageExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, TRUE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCPUUsageExtendedStoredProcedure)

void CCPUUsageExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_CPUUSAGE @TimeIntervalMS, @CPUUsage OUTPUT");
}

BOOL CCPUUsageExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get the function pointer for "NtQuerySystemInformation"
  LPNTQUERYSYSTEMINFORMATION pNtQuerySystemInformation = (LPNTQUERYSYSTEMINFORMATION) GetProcAddress(GetModuleHandle(_T("NTDLL")), "NtQuerySystemInformation");
  if (pNtQuerySystemInformation == NULL)
  {
    sprintf(szErrorMsg, "%s: Unable to find the function NtQuerySystemInformation, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }
  
  //Also get the idle times on start
  SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION StartSPI;
  NTSTATUS status = pNtQuerySystemInformation(SystemProcessorPerformanceInformation, &StartSPI, sizeof(StartSPI), NULL);
  if (status != NO_ERROR)
  {
    sprintf(szErrorMsg, "%s: An error occured calling NtQuerySystemInformation, Error:%x", m_szFunctionName, status);
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }

  //Sleep for the specified interval. Remember that CPU Usage
  //is a rate and as such means nothing if not measured across a reasonable time interval. A reasonable value
  //to pick would be 500 or 1000 milliseconds
  DBINT lVal = XP_INT4(m_pParameterData[CPUUSAGE_TIME_INTERVAL_INDEX].m_Data);
  Sleep(lVal);

  //Also get the idle times on end
  SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION EndSPI;
  status = pNtQuerySystemInformation(SystemProcessorPerformanceInformation, &EndSPI, sizeof(EndSPI), NULL);
  if (status != NO_ERROR)
  {
    sprintf(szErrorMsg, "%s: An error occured calling NtQuerySystemInformation, Error:%x", m_szFunctionName, status);
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }

  //Now do the calculation 
  __int64 TotalIdleAmount = ((__int64) (EndSPI.IdleTime.QuadPart) - (__int64) (StartSPI.IdleTime.QuadPart));
  __int64 TotalKernelAmount = ((__int64) (EndSPI.KernelTime.QuadPart) - (__int64) (StartSPI.KernelTime.QuadPart));
  __int64 TotalUserAmount = ((__int64) (EndSPI.UserTime.QuadPart) - (__int64) (StartSPI.UserTime.QuadPart));
  int nCPUUsage = (int) (100 - ((TotalIdleAmount * 100) / (TotalKernelAmount + TotalUserAmount)));

  //Return the data in the output parameter    
  ParamSetOutput(CPUUSAGE_USAGE_INDEX+1, (BYTE*) &nCPUUsage, sizeof(nCPUUsage), FALSE);

  return TRUE;
}

