/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPAsymmetricEncryption.h"
#include "XPSignatures.h"


/////////////////////////////// Implementation ////////////////////////////////

#define ENCRYPT_DATA_TO_ENCRYPT_INDEX  0
#define ENCRYPT_ENCRYPTED_DATA_INDEX   1

BEGIN_IMPLEMENT_XP_PARAMETERS(CRSAAsymmetricEncryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CRSAAsymmetricEncryptionExtendedStoredProcedure)

BOOL CRSAAsymmetricEncryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CCryptoAPISignExtendedStoredProcedure::CryptAcquireContext(&hProv, _T("XP_CRYPTOAPI"), NULL, PROV_RSA_FULL, CRYPT_MACHINE_KEYSET | CRYPT_SILENT)) 
  {
    //Get the key exchange key
    HCRYPTKEY hKey;
    if (CryptGetUserKey(hProv, AT_KEYEXCHANGE, &hKey))
    {
      //First check to see what size the encrypted data will be
      DWORD dwPlainTextSize = m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen;
      DWORD dwEncryptedSize = dwPlainTextSize;
      if (CryptEncrypt(hKey, NULL, TRUE, 0, NULL, &dwEncryptedSize, dwPlainTextSize))
      {
        //Allocate memory to contain the encrypted data
        BYTE* pbyEncryptedData = NULL;
        try
        {
          pbyEncryptedData = new BYTE[dwEncryptedSize];
        }
  #ifdef _AFX
        catch(CMemoryException* pEx)
        {
          pEx->Delete();
        }
  #endif
        catch(std::bad_alloc&)
        {
        }

        if (pbyEncryptedData)
        {
          //Transfer the data to be encrypted into the decrypted array because MS Crypto does encryption in place
          memcpy(pbyEncryptedData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_pData, dwPlainTextSize);

          //Ok, Now recall with the buffer to encrypt into
          if (CryptEncrypt(hKey, NULL, TRUE, 0, pbyEncryptedData, &dwPlainTextSize, dwEncryptedSize))
          {
            //Return the data in the output parameter directy
          	bSuccess = (ParamSetOutput(ENCRYPT_ENCRYPTED_DATA_INDEX+1, pbyEncryptedData, dwEncryptedSize, FALSE) == SUCCEED);
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to encrypt the data, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }

          //Finally tidy up the heap memory we have been using
          delete [] pbyEncryptedData; 
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to allocate memory to store the encrypted data in", m_szFunctionName);
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to get the encrypted data size, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }          

      //Destroy the key now that we are finished with it
      CryptDestroyKey(hKey);          
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to acquire the key exchange key, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CRSAAsymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_ENCRYPT @DataToEncrypt @EncryptedData");
}



#define DECRYPT_DATA_TO_DECRYPT_INDEX 0
#define DECRYPT_PUBLICKEY_INDEX       1
#define DECRYPT_DECRYPTED_DATA_INDEX  2

BEGIN_IMPLEMENT_XP_PARAMETERS(CRSAAsymmetricDecryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CRSAAsymmetricDecryptionExtendedStoredProcedure)

BOOL CRSAAsymmetricDecryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CCryptoAPISignExtendedStoredProcedure::CryptAcquireContext(&hProv, _T("XP_CRYPTOAPI"), NULL, PROV_RSA_FULL, CRYPT_MACHINE_KEYSET | CRYPT_SILENT)) 
  {
    //Import the public key into the CSP
    HCRYPTKEY hKey;
    if (CryptImportKey(hProv, m_pParameterData[DECRYPT_PUBLICKEY_INDEX].m_pData, m_pParameterData[DECRYPT_PUBLICKEY_INDEX].m_cbActualLen, 0, 0, &hKey))
    {
      //Ok, Decrypt in place
      DWORD dwPlainTextSize = m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen;
      if (CryptDecrypt(hKey, NULL, TRUE, 0, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, &dwPlainTextSize))
      {
        //Return the data in the output parameter directy
        bSuccess = (ParamSetOutput(DECRYPT_DECRYPTED_DATA_INDEX+1, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, dwPlainTextSize, FALSE) == SUCCEED);
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to decrypt the data, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Destroy the key now that we are finished with it
      CryptDestroyKey(hKey);          
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to import the public key, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CRSAAsymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_DECRYPT @DataToDecrypt @PublicKeyData @DecryptedData");
}



