/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPHMAC.h"


/////////////////////////////// Implementation ////////////////////////////////

#define HMAC_DATA_TO_HMAC_INDEX          0
#define HMAC_PASSWORD_HMAC_INDEX         1
#define HMAC_SESSION_KEY_ALGORITHM_INDEX 2
#define HMAC_HMAC_INDEX                  3

BEGIN_IMPLEMENT_XP_PARAMETERS(CHMACExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY,                                  TRUE,  FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY,                                  TRUE,  FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR | XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CHMACExtendedStoredProcedure)

BOOL CHMACExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, m_dwProviderType, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use SHA1 to generate the hash, as it is longest bit length of any hash algorithm generally available to us)
    HCRYPTHASH hKeyHash;
    if (CryptCreateHash(hProv, CALG_SHA1, 0, 0, &hKeyHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hKeyHash, m_pParameterData[HMAC_PASSWORD_HMAC_INDEX].m_pData, m_pParameterData[HMAC_PASSWORD_HMAC_INDEX].m_cbActualLen, 0))
      {
        //Now derive a session key from the hash
        HCRYPTKEY hKey;  
        ALG_ID SessionAlgID = (ALG_ID) XP_INT4(m_pParameterData[HMAC_SESSION_KEY_ALGORITHM_INDEX].m_Data);
        if (CryptDeriveKey(hProv, SessionAlgID, hKeyHash, 0, &hKey))
        {
          //Create the HMAC object
          HCRYPTHASH hHMAC;
          if (CryptCreateHash(hProv, CALG_HMAC, hKey, 0, &hHMAC))
          {
            //Specify the type of hash to use for generation of the HMAC
            HMAC_INFO hi;
            hi.HashAlgid = m_HashAlgID;
            hi.cbInnerString = 0;
            hi.pbInnerString = NULL;
            hi.cbOuterString = 0;
            hi.pbOuterString = NULL;
            if (CryptSetHashParam(hHMAC, HP_HMAC_INFO, (BYTE*) &hi, 0))
            {
              //Next thing to do is hash the actual data
              if (CryptHashData(hHMAC, m_pParameterData[HMAC_DATA_TO_HMAC_INDEX].m_pData, m_pParameterData[HMAC_DATA_TO_HMAC_INDEX].m_cbActualLen, 0))
              {
                //Get the size of the generated hash
                DWORD dwHashSize = 0;
                DWORD dwSize = sizeof(dwHashSize);
                if (CryptGetHashParam(hHMAC, HP_HASHSIZE, (BYTE*) &dwHashSize, &dwSize, 0))
                {
                  //Allocate some memory (on the stack) and call the function to get the hash itself
                  BYTE* pHashData = (BYTE*) _alloca(dwHashSize);
                  if (CryptGetHashParam(hHMAC, HP_HASHVAL, pHashData, &dwHashSize, 0))
                  {
                    if ((m_pParameterData[HMAC_HMAC_INDEX].m_bType == SRVBIGVARBINARY) || (m_pParameterData[HMAC_HMAC_INDEX].m_bType == SRVVARBINARY) || (m_pParameterData[HMAC_HMAC_INDEX].m_bType == SRVBIGBINARY))
                    {
                      //Return the data in the output parameter directy
          	          bSuccess = (ParamSetOutput(HMAC_HMAC_INDEX+1, pHashData, dwHashSize, FALSE) == SUCCEED);
                    }
                    else
                    {
                      //Convert the hash to a BCD style string first
                      DWORD dwBCDHashSize = dwHashSize*2;
                      BYTE* pBCDHashData = (BYTE*) _alloca(dwBCDHashSize);
                      for (DWORD i=0; i<dwHashSize; i++)
                      {
                        int nChar = (pHashData[i] & 0xF0) >> 4;
                        if (nChar <= 9)
                          pBCDHashData[i*2] = (BYTE) (nChar + '0');
                        else
                          pBCDHashData[i*2] = (BYTE) (nChar - 10 + 'A');

                        nChar = pHashData[i] & 0x0F;
                        if (nChar <= 9)
                          pBCDHashData[i*2 + 1] = (BYTE) (nChar + '0');
                        else
                          pBCDHashData[i*2 + 1] = (BYTE) (nChar - 10 + 'A');
                      }

                      //convert the hash into the output parameter data type
                      int nConvertedDataSize = Convert(SRVVARCHAR, pBCDHashData, dwBCDHashSize, m_pParameterData[HMAC_HMAC_INDEX].m_bType, m_pParameterData[HMAC_HMAC_INDEX].m_pData, m_pParameterData[HMAC_HMAC_INDEX].m_cbMaxLen);
                      if ((nConvertedDataSize > 0))
                      {
                        //Return the data in the output parameter
          	            bSuccess = (ParamSetOutput(HMAC_HMAC_INDEX+1, m_pParameterData[HMAC_HMAC_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
                      }
                      else
                      {
                        sprintf(szErrorMsg, "%s: Failed to convert the hash to the required output type, Error:%d", m_szFunctionName, GetLastError());
                        SendErrorMsg(szErrorMsg);
                      }
                    }
                  }
                  else
                  {
                    sprintf(szErrorMsg, "%s: Failed to get the hash data, Error:%d", m_szFunctionName, GetLastError());
                    SendErrorMsg(szErrorMsg);
                  }
                }
                else
                {
                  sprintf(szErrorMsg, "%s: Failed to get the hash data size, Error:%d", m_szFunctionName, GetLastError());
                  SendErrorMsg(szErrorMsg);
                }
              }
              else
              {
                sprintf(szErrorMsg, "%s: Failed to hash data, Error:%d", m_szFunctionName, GetLastError());
                SendErrorMsg(szErrorMsg);
              }
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed to set the hash type, Error:%d", m_szFunctionName, GetLastError());
              SendErrorMsg(szErrorMsg);
            }

            //Release the HMAC object
            CryptDestroyHash(hHMAC);
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to create HMAC, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }

          //Release the session key object
          CryptDestroyKey(hKey);
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to create session encryption key from hash, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the password hash
      CryptDestroyHash(hKeyHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash for password, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}


CMD2HMACExtendedStoredProcedure::CMD2HMACExtendedStoredProcedure()
{
  m_HashAlgID = CALG_MD2;
  m_dwProviderType = PROV_RSA_FULL;
}

void CMD2HMACExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_MD2_HMAC @DataToHash @Password @SessionKeyAlgorithm @HMAC");
}


CMD4HMACExtendedStoredProcedure::CMD4HMACExtendedStoredProcedure()
{
  m_HashAlgID = CALG_MD4;
  m_dwProviderType = PROV_RSA_FULL;
}

void CMD4HMACExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_MD4_HMAC @DataToHash @Password @SessionKeyAlgorithm @HMAC");
}


CMD5HMACExtendedStoredProcedure::CMD5HMACExtendedStoredProcedure()
{
  m_HashAlgID = CALG_MD5;
  m_dwProviderType = PROV_RSA_FULL;
}

void CMD5HMACExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_MD5_HMAC @DataToHash @Password @SessionKeyAlgorithm @HMAC");
}


CSHA1HMACExtendedStoredProcedure::CSHA1HMACExtendedStoredProcedure()
{
  m_HashAlgID = CALG_SHA1;
  m_dwProviderType = PROV_RSA_FULL;
}

void CSHA1HMACExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SHA1_HMAC @DataToHash @Password @SessionKeyAlgorithm @HMAC");
}




