/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPSymmetricEncryption.h"
#include "Helix.h"
#include "Blowfish.h"


/////////////////////////////// Implementation ////////////////////////////////

#define ENCRYPT_DATA_TO_ENCRYPT_INDEX  0
#define ENCRYPT_PASSWORD_INDEX         1
#define ENCRYPT_ENCRYPTED_DATA_INDEX   2

BEGIN_IMPLEMENT_XP_PARAMETERS(CCryptoAPISymmetricEncryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCryptoAPISymmetricEncryptionExtendedStoredProcedure)

BOOL CCryptoAPISymmetricEncryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, m_dwProviderType, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use SHA1 to generate the hash, as it is longest bit length of any hash algorithm generally available to us)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_SHA1, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[ENCRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[ENCRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Now derive a session key from the hash
        HCRYPTKEY hKey;  
        if (CryptDeriveKey(hProv, m_AlgID, hHash, 0, &hKey))
        {
          //First check to see what size the encrypted data will be
          DWORD dwPlainTextSize = m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen;
          DWORD dwEncryptedSize = dwPlainTextSize;
          if (CryptEncrypt(hKey, NULL, TRUE, 0, NULL, &dwEncryptedSize, dwPlainTextSize))
          {
            //Allocate memory to contain the encrypted data
            BYTE* pbyEncryptedData = NULL;
            try
            {
              pbyEncryptedData = new BYTE[dwEncryptedSize];
            }
      #ifdef _AFX
            catch(CMemoryException* pEx)
            {
              pEx->Delete();
            }
      #endif
            catch(std::bad_alloc&)
            {
            }

            if (pbyEncryptedData)
            {
              //Transfer the data to be encrypted into the decrypted array because MS Crypto does encryption in place
              memcpy(pbyEncryptedData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_pData, dwPlainTextSize);

              //Ok, Now recall with the buffer to encrypt into
              if (CryptEncrypt(hKey, NULL, TRUE, 0, pbyEncryptedData, &dwPlainTextSize, dwEncryptedSize))
              {
                //Return the data in the output parameter directy
          	    bSuccess = (ParamSetOutput(ENCRYPT_ENCRYPTED_DATA_INDEX+1, pbyEncryptedData, dwEncryptedSize, FALSE) == SUCCEED);
              }
              else
              {
                sprintf(szErrorMsg, "%s: Failed to encrypt the data, Error:%d", m_szFunctionName, GetLastError());
                SendErrorMsg(szErrorMsg);
              }

              //Finally tidy up the heap memory we have been using
              delete [] pbyEncryptedData; 
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed to allocate memory to store the encrypted data in", m_szFunctionName);
              SendErrorMsg(szErrorMsg);
            }
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to get the encrypted data size, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }          

          //Destroy the key now that we are finished with it
          CryptDestroyKey(hKey);          
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to create session encryption key from hash, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



CRC2SymmetricEncryptionExtendedStoredProcedure::CRC2SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC2;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC2SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC2_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CRC4SymmetricEncryptionExtendedStoredProcedure::CRC4SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC4;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC4SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC4_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CRC5SymmetricEncryptionExtendedStoredProcedure::CRC5SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC5;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC5SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC5_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CDESSymmetricEncryptionExtendedStoredProcedure::CDESSymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_DES;
  m_dwProviderType = PROV_RSA_FULL;
}

void CDESSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DES_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



C3DES112SymmetricEncryptionExtendedStoredProcedure::C3DES112SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_3DES_112;
  m_dwProviderType = PROV_RSA_FULL;
}

void C3DES112SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_3DES112_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



C3DESSymmetricEncryptionExtendedStoredProcedure::C3DESSymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_3DES;
  m_dwProviderType = PROV_RSA_FULL;
}

void C3DESSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_3DES_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}


CDESXSymmetricEncryptionExtendedStoredProcedure::CDESXSymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_DESX;
  m_dwProviderType = PROV_RSA_FULL;
}

void CDESXSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DESX_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}


CAES128SymmetricEncryptionExtendedStoredProcedure::CAES128SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_128;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES128SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES128_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CAES192SymmetricEncryptionExtendedStoredProcedure::CAES192SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_192;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES192SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES192_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CAES256SymmetricEncryptionExtendedStoredProcedure::CAES256SymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_256;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES256SymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES256_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CSkipjackSymmetricEncryptionExtendedStoredProcedure::CSkipjackSymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_SKIPJACK;
  m_dwProviderType = PROV_FORTEZZA; //We use the Fortezza Provider type as it supports Skipjack
}

void CSkipjackSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SKIPJACK_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



CCYLINK_MEKSymmetricEncryptionExtendedStoredProcedure::CCYLINK_MEKSymmetricEncryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_CYLINK_MEK;
  m_dwProviderType = PROV_DSS_DH; //We use the Diffie-Hellman Provider type as it supports Cylink MEK
}

void CCYLINK_MEKSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_CYLINK_MEK_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}



#define DECRYPT_DATA_TO_DECRYPT_INDEX  0
#define DECRYPT_PASSWORD_INDEX         1
#define DECRYPT_DECRYPTED_DATA_INDEX   2

BEGIN_IMPLEMENT_XP_PARAMETERS(CCryptoAPISymmetricDecryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCryptoAPISymmetricDecryptionExtendedStoredProcedure)

BOOL CCryptoAPISymmetricDecryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, m_dwProviderType, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use SHA1 to generate the hash, as it is longest bit length of any hash algorithm generally available to us)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_SHA1, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[DECRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[DECRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Now derive a session key from the hash
        HCRYPTKEY hKey;  
        if (CryptDeriveKey(hProv, m_AlgID, hHash, 0, &hKey))
        {
          //Ok, Decrypt in place
          DWORD dwPlainTextSize = m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen;
          if (CryptDecrypt(hKey, NULL, TRUE, 0, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, &dwPlainTextSize))
          {
            //Return the data in the output parameter directy
          	bSuccess = (ParamSetOutput(DECRYPT_DECRYPTED_DATA_INDEX+1, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, dwPlainTextSize, FALSE) == SUCCEED);
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to decrypt the data, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }

          //Destroy the key now that we are finished with it
          CryptDestroyKey(hKey);          
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to create session decryption key from hash, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



CRC2SymmetricDecryptionExtendedStoredProcedure::CRC2SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC2;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC2SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC2_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CRC4SymmetricDecryptionExtendedStoredProcedure::CRC4SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC4;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC4SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC4_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CRC5SymmetricDecryptionExtendedStoredProcedure::CRC5SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_RC5;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRC5SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RC5_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CDESSymmetricDecryptionExtendedStoredProcedure::CDESSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_DES;
  m_dwProviderType = PROV_RSA_FULL;
}

void CDESSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DES_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



C3DES112SymmetricDecryptionExtendedStoredProcedure::C3DES112SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_3DES_112;
  m_dwProviderType = PROV_RSA_FULL;
}

void C3DES112SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_3DES112_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



C3DESSymmetricDecryptionExtendedStoredProcedure::C3DESSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_3DES;
  m_dwProviderType = PROV_RSA_FULL;
}

void C3DESSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_3DES_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CDESXSymmetricDecryptionExtendedStoredProcedure::CDESXSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_DESX;
  m_dwProviderType = PROV_RSA_FULL;
}

void CDESXSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DESX_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CAES128SymmetricDecryptionExtendedStoredProcedure::CAES128SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_128;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES128SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES128_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CAES192SymmetricDecryptionExtendedStoredProcedure::CAES192SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_192;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES192SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES192_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CAES256SymmetricDecryptionExtendedStoredProcedure::CAES256SymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_AES_256;
  m_dwProviderType = PROV_RSA_AES; //We require the AES Provider type
}

void CAES256SymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_AES256_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CSkipjackSymmetricDecryptionExtendedStoredProcedure::CSkipjackSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_SKIPJACK;
  m_dwProviderType = PROV_FORTEZZA; //We use the Fortezza Provider type as it supports Skipjack
}

void CSkipjackSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SKIPJACK_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CTEKSymmetricDecryptionExtendedStoredProcedure::CTEKSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_TEK;
  m_dwProviderType = PROV_DH_SCHANNEL; //We use the Fortezza Provider type as it supports TEK
}

void CTEKSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_TEK_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



CCYLINK_MEKSymmetricDecryptionExtendedStoredProcedure::CCYLINK_MEKSymmetricDecryptionExtendedStoredProcedure()
{
  m_AlgID = CALG_CYLINK_MEK;
  m_dwProviderType = PROV_DSS_DH; //We use the Diffie-Hellman Provider type as it supports Cylink MEK
}

void CCYLINK_MEKSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_CYLINK_MEK_DECRYPT @DataToDecrypt @Password @DecryptedData");
}



#define HELIX_ENCRYPT_DATA_TO_ENCRYPT_INDEX  0
#define HELIX_ENCRYPT_PASSWORD_INDEX         1
#define HELIX_ENCRYPT_ENCRYPTED_DATA_INDEX   2
#define HELIX_ENCRYPT_HASH_INDEX             3

void CHelixSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_HELIX_ENCRYPT @DataToEncrypt @Password @EncryptedData @Hash");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CHelixSymmetricEncryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CHelixSymmetricEncryptionExtendedStoredProcedure)

BOOL CHelixSymmetricEncryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP, so that we can hash the password
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use MD5 to generate the hash)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_MD5, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[HELIX_ENCRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[HELIX_ENCRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Get the size of the generated hash
        DWORD dwHashSize = 0;
        DWORD dwSize = sizeof(dwHashSize);
        if (CryptGetHashParam(hHash, HP_HASHSIZE, (BYTE*) &dwHashSize, &dwSize, 0))
        {
          //Allocate some memory (on the stack) and call the function to get the hash itself
          BYTE* pHashData = (BYTE*) _alloca(dwHashSize);
          if (CryptGetHashParam(hHash, HP_HASHVAL, pHashData, &dwHashSize, 0))
          {
            //Now create the class which does the helix encryption
            CHelix helix;
            helix.SetKey(pHashData, dwHashSize);
            CHelixNonce nonce;
            CHelixMAC mac;

            //Allocate memory to contain the encrypted data
            BYTE* pbyEncryptedData = NULL;
            try
            {
              pbyEncryptedData = new BYTE[m_pParameterData[HELIX_ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen];
            }
      #ifdef _AFX
            catch(CMemoryException* pEx)
            {
              pEx->Delete();
            }
      #endif
            catch(std::bad_alloc&)
            {
            }

            if (pbyEncryptedData)
            {
              if (helix.Encrypt(m_pParameterData[HELIX_ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_pData, m_pParameterData[HELIX_ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen, nonce, pbyEncryptedData, mac))
              {
                //Return the data in the 2 output parameter directy
          	    bSuccess = (ParamSetOutput(HELIX_ENCRYPT_ENCRYPTED_DATA_INDEX+1, pbyEncryptedData, m_pParameterData[HELIX_ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen, FALSE) == SUCCEED);
                if (bSuccess)
                  bSuccess = (ParamSetOutput(HELIX_ENCRYPT_HASH_INDEX+1, mac.m_Data, sizeof(mac.m_Data), FALSE) == SUCCEED);
              }
              else
              {
                sprintf(szErrorMsg, "%s: Failed to encrypt the data, Error:%d", m_szFunctionName, GetLastError());
                SendErrorMsg(szErrorMsg);
              }

              //Finally tidy up the heap memory we have been using
              delete [] pbyEncryptedData; 
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed to allocate memory to store the encrypted data in", m_szFunctionName);
              SendErrorMsg(szErrorMsg);
            }
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to get the hash data, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }          
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get the hash data size, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



#define HELIX_DECRYPT_DATA_TO_DECRYPT_INDEX  0
#define HELIX_DECRYPT_PASSWORD_INDEX         1
#define HELIX_DECRYPT_HASH_INDEX             2
#define HELIX_DECRYPT_DECRYPTED_DATA_INDEX   3

void CHelixSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_HELIX_DECRYPT @DataToDecrypt @Password @Hash @DecryptedData");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CHelixSymmetricDecryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CHelixSymmetricDecryptionExtendedStoredProcedure)

BOOL CHelixSymmetricDecryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP, so that we can hash the password
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use MD5 to generate the hash)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_MD5, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[HELIX_DECRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[HELIX_DECRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Get the size of the generated hash
        DWORD dwHashSize = 0;
        DWORD dwSize = sizeof(dwHashSize);
        if (CryptGetHashParam(hHash, HP_HASHSIZE, (BYTE*) &dwHashSize, &dwSize, 0))
        {
          //Allocate some memory (on the stack) and call the function to get the hash itself
          BYTE* pHashData = (BYTE*) _alloca(dwHashSize);
          if (CryptGetHashParam(hHash, HP_HASHVAL, pHashData, &dwHashSize, 0))
          {
            //Now create the class which does the helix decryption
            CHelix helix;
            helix.SetKey(pHashData, dwHashSize);
            CHelixNonce nonce;

            //Setup the correct hash
            CHelixMAC mac;
            memcpy(mac.m_Data, m_pParameterData[HELIX_DECRYPT_HASH_INDEX].m_pData, min(sizeof(mac.m_Data), m_pParameterData[HELIX_DECRYPT_HASH_INDEX].m_cbActualLen));

            //Allocate memory to contain the decrypted data
            BYTE* pbyPlainTextData = NULL;
            try
            {
              pbyPlainTextData = new BYTE[m_pParameterData[HELIX_DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen];
            }
      #ifdef _AFX
            catch(CMemoryException* pEx)
            {
              pEx->Delete();
            }
      #endif
            catch(std::bad_alloc&)
            {
            }

            if (pbyPlainTextData)
            {
              if (helix.Decrypt(m_pParameterData[HELIX_DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, m_pParameterData[HELIX_DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen, nonce, mac, pbyPlainTextData))
              {
                //Return the data in the output parameter directy
          	    bSuccess = (ParamSetOutput(HELIX_DECRYPT_DECRYPTED_DATA_INDEX+1, pbyPlainTextData, m_pParameterData[HELIX_DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen, FALSE) == SUCCEED);
              }
              else
              {
                sprintf(szErrorMsg, "%s: Failed to decrypt the data, Error:%d", m_szFunctionName, GetLastError());
                SendErrorMsg(szErrorMsg);
              }

              //Finally tidy up the heap memory we have been using
              delete [] pbyPlainTextData; 
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed to allocate memory to store the decrypted data in", m_szFunctionName);
              SendErrorMsg(szErrorMsg);
            }
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to get the hash data, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }          
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get hash data size, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



void CBlowfishSymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_BLOWFISH_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CBlowfishSymmetricEncryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CBlowfishSymmetricEncryptionExtendedStoredProcedure)

BOOL CBlowfishSymmetricEncryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //Allocate memory to contain the encrypted data
  BYTE* pbyEncryptedData = NULL;
  DWORD dwEncryptedSize = ((m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen+7)/8)*8;
  DWORD dwPaddedBytes = dwEncryptedSize - m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen;
  if (dwPaddedBytes == 0)
  {
    dwPaddedBytes = 8;
    dwEncryptedSize += 8;
  }
  try
  {
    pbyEncryptedData = new BYTE[dwEncryptedSize];
  }
#ifdef _AFX
  catch(CMemoryException* pEx)
  {
    pEx->Delete();
  }
#endif
  catch(std::bad_alloc&)
  {
  }

  if (pbyEncryptedData)
  {
    //Setup the padding
    memcpy(pbyEncryptedData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_pData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen);
    memset(pbyEncryptedData + dwEncryptedSize - dwPaddedBytes, 0, dwPaddedBytes - 1);
    pbyEncryptedData[dwEncryptedSize - 1] = (BYTE) dwPaddedBytes;

    //Class which does the blowfish decryption
    CBlowFish bf;      
    bf.Initialize(m_pParameterData[ENCRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[ENCRYPT_PASSWORD_INDEX].m_cbActualLen);
    bf.Encode(pbyEncryptedData, pbyEncryptedData, dwEncryptedSize);

    //Return the data in the output parameter directy
    bSuccess = (ParamSetOutput(ENCRYPT_ENCRYPTED_DATA_INDEX+1, pbyEncryptedData, dwEncryptedSize, FALSE) == SUCCEED);

    //Tidy up the heap memory we have used
    delete [] pbyEncryptedData;
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to allocate memory to store the encrypted data in", m_szFunctionName);
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}


void CBlowfishSymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_BLOWFISH_DECRYPT @DataToDecrypt @Password @DecryptedData");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CBlowfishSymmetricDecryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CBlowfishSymmetricDecryptionExtendedStoredProcedure)

BOOL CBlowfishSymmetricDecryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //Allocate memory to contain the decrypted data
  BYTE* pbyPlainTextData = NULL;
  try
  {
    pbyPlainTextData = new BYTE[m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen];
  }
#ifdef _AFX
  catch(CMemoryException* pEx)
  {
    pEx->Delete();
  }
#endif
  catch(std::bad_alloc&)
  {
  }

  if (pbyPlainTextData)
  {
    //Class which does the blowfish decryption
    CBlowFish bf;      
    bf.Initialize(m_pParameterData[DECRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[DECRYPT_PASSWORD_INDEX].m_cbActualLen);
    bf.Decode(m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, pbyPlainTextData, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen);

    //Determine how much padding we have
    DWORD dwDecryptedSize = m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen - pbyPlainTextData[m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen - 1];

    //Return the data in the output parameter directy
    bSuccess = (ParamSetOutput(DECRYPT_DECRYPTED_DATA_INDEX+1, pbyPlainTextData, dwDecryptedSize, FALSE) == SUCCEED);

    //Tidy up the heap memory we have used
    delete [] pbyPlainTextData;
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to allocate memory to store the decrypted data in", m_szFunctionName);
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



void CTEASymmetricEncryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_TEA_ENCRYPT @DataToEncrypt @Password @EncryptedData");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CTEASymmetricEncryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CTEASymmetricEncryptionExtendedStoredProcedure)

void CTEASymmetricEncryptionExtendedStoredProcedure::TEAEncode(long* v, long* k)  
{              
  unsigned long y = v[0];
  unsigned long z = v[1];
  unsigned long sum = 0;
  unsigned long delta = 0x9e3779b9;
  unsigned long n = 32;
  while (n-- > 0) 
  {
    sum += delta;
    y += (z << 4) + k[0] ^ z+sum ^ (z >> 5) + k[1];
    z += (y << 4) + k[2] ^ y+sum ^ (y >> 5) + k[3];
  } 

  v[0] = y; 
  v[1] = z; 
}

BOOL CTEASymmetricEncryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP, so that we can hash the password
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use MD5 to generate the hash)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_MD5, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[HELIX_ENCRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[HELIX_ENCRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Call the function to get the hash itself
        BYTE HashData[16];
        DWORD dwHashSize = 16;
        if (CryptGetHashParam(hHash, HP_HASHVAL, HashData, &dwHashSize, 0))
        {
          BYTE* pbyEncryptedData = NULL;
          DWORD dwEncryptedSize = ((m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen+7)/8)*8;
          DWORD dwPaddedBytes = dwEncryptedSize - m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen;
          if (dwPaddedBytes == 0)
          {
            dwPaddedBytes = 8;
            dwEncryptedSize += 8;
          }

          try
          {
            pbyEncryptedData = new BYTE[dwEncryptedSize];
          }
        #ifdef _AFX
          catch(CMemoryException* pEx)
          {
            pEx->Delete();
          }
        #endif
          catch(std::bad_alloc&)
          {
          }

          if (pbyEncryptedData)
          {
            //Setup the padding
            memcpy(pbyEncryptedData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_pData, m_pParameterData[ENCRYPT_DATA_TO_ENCRYPT_INDEX].m_cbActualLen);
            memset(pbyEncryptedData + dwEncryptedSize - dwPaddedBytes, 0, dwPaddedBytes - 1);
            pbyEncryptedData[dwEncryptedSize - 1] = (BYTE) dwPaddedBytes;

            //Do the actual encryption
            DWORD dwLoops = dwEncryptedSize / 8;
            long* pdwEncryptedData = (long*) pbyEncryptedData;
            long* pdwKey = (long*) &HashData;
            for (DWORD i=0; i<dwLoops; i++)
            {
              TEAEncode(pdwEncryptedData, pdwKey);
              pdwEncryptedData += 2;  
            }

            //Return the data in the output parameter directy
            bSuccess = (ParamSetOutput(ENCRYPT_ENCRYPTED_DATA_INDEX+1, pbyEncryptedData, dwEncryptedSize, FALSE) == SUCCEED);
 
            //Finally tidy up the heap memory we have been using
            delete [] pbyEncryptedData; 
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to allocate memory to store the encrypted data in", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
          }
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get the hash data, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }          
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}


void CTEASymmetricDecryptionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_TEA_DECRYPT @DataToDecrypt @Password @DecryptedData");
}

BEGIN_IMPLEMENT_XP_PARAMETERS(CTEASymmetricDecryptionExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CTEASymmetricDecryptionExtendedStoredProcedure)

void CTEASymmetricDecryptionExtendedStoredProcedure::TEADecode(long* v, long* k)  
{
  unsigned long n = 32;
  unsigned long y = v[0];
  unsigned long z = v[1];
  unsigned long delta = 0x9e3779b9;
  unsigned long sum = delta << 5;

  while (n-- > 0) 
  {
    z -= (y << 4) + k[2] ^ y + sum ^ (y >> 5) + k[3]; 
    y -= (z << 4) + k[0] ^ z + sum ^ (z >> 5) + k[1];
    sum -= delta;  
  }

  v[0] = y; 
  v[1] = z;  
}

BOOL CTEASymmetricDecryptionExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP, so that we can hash the password
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Create the hash object (we will always use MD5 to generate the hash)
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, CALG_MD5, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the password text
      if (CryptHashData(hHash, m_pParameterData[HELIX_DECRYPT_PASSWORD_INDEX].m_pData, m_pParameterData[HELIX_DECRYPT_PASSWORD_INDEX].m_cbActualLen, 0))
      {
        //Call the function to get the hash itself
        BYTE HashData[16];
        DWORD dwHashSize = 16;
        if (CryptGetHashParam(hHash, HP_HASHVAL, HashData, &dwHashSize, 0))
        {
          //Allocate memory to contain the decrypted data
          BYTE* pbyPlainTextData = NULL;
          try
          {
            pbyPlainTextData = new BYTE[m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen];
          }
        #ifdef _AFX
          catch(CMemoryException* pEx)
          {
            pEx->Delete();
          }
        #endif
          catch(std::bad_alloc&)
          {
          }

          if (pbyPlainTextData)
          {
            //Do the actual decryption (in place)
            DWORD dwLoops = m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen / 8;
            memcpy(pbyPlainTextData, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_pData, m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen);
            long* pdwPlainTextData = (long*) pbyPlainTextData;
            long* pdwKey = (long*) &HashData;
            for (DWORD i=0; i<dwLoops; i++)
            {
              TEADecode(pdwPlainTextData, pdwKey);
              pdwPlainTextData += 2;  
            }

            //Determine how much padding we have
            DWORD dwDecryptedSize = m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen - pbyPlainTextData[m_pParameterData[DECRYPT_DATA_TO_DECRYPT_INDEX].m_cbActualLen - 1];

            //Return the data in the output parameter directy
            bSuccess = (ParamSetOutput(DECRYPT_DECRYPTED_DATA_INDEX+1, pbyPlainTextData, dwDecryptedSize, FALSE) == SUCCEED);

            //Tidy up the heap memory we have used
            delete [] pbyPlainTextData;
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to allocate memory to store the decrypted data in", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
          }
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get the hash data, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }          
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash password, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}
