/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_GISProcs.h"




/////////////////////////////// Macros / Defines //////////////////////////////

const double PI = 3.1415926535897932384626433832795;
const double g_deg2rad = PI / 180.0;
const double g_rad2deg = 180.0 / PI;

struct CReferenceEllipsoid
{
  char*  Name;             //The name of the datum
  double EquatorialRadius; //The equatorial radius in meters (a)
  double OneOverF;         //1 / Flattening (1/f)
};

enum ReferenceEllipsoid //These enums correspond to the index of the corresponding values in g_ReferenceEllipsoids
{                       //You can use this list as a reference when calling GIS XPs which require a ReferenceEllipsoidID
  AIRY_1830             = 0,
  MODIFIED_AIRY         = 1,
  AUSTRALIAN_NATIONAL   = 2,
  BESSEL_1841_NAMIBIA   = 3,
  BESSEL_1841           = 4,
  CLARKE_1866           = 5,
  CLARKE_1880           = 6,
  EVEREST_INDIA_1830    = 7,
  EVEREST_SABAH_SARAWAK = 8,
  EVEREST_INDIA_1956    = 9,
  EVEREST_MALAYSIA_1969 = 10,
  EVEREST_MALAY_SING    = 11,
  EVEREST_PAKISTAN      = 12,
  MODIFIED_FISCHER_1960 = 13,
 	FISCHER_1968          = 14,
  HELMERT_1906          = 15,
  HOUGH_1960            = 16,
  INDONESIAN_1974       = 17,
  INTERNATIONAL_1924    = 18,
  KRASSOVSKY_1940       = 19,
	GRS_1967              = 20,
  GRS_80                = 21,
  SOUTH_AMERICAN_1969   = 22,
  WGS_60                = 23,
  WGS_66                = 24,
  WGS_72                = 25,
  WGS_84                = 26,
  IAU_1976              = 27
};

CReferenceEllipsoid g_ReferenceEllipsoids[] = 
{
  { "Airy 1830",               6377563.396, 299.3249646   },
  { "Modified Airy",           6377340.189, 299.3249646   },
  { "Australian National",     6378160,     298.25        },
  { "Bessel 1841 (Namibia)",   6377483.865, 299.1528128   },
  { "Bessel 1841",             6377397.155, 299.1528128   },
  { "Clarke 1866",             6378206.4,   294.9786982   },
  { "Clarke 1880",             6378249.145, 293.465       },
  { "Everest (India 1830)",    6377276.345, 300.8017      },
  { "Everest (Sabah Sarawak)", 6377298.556, 300.8017      },
  { "Everest (India 1956)",    6377301.243, 300.8017      },
  { "Everest (Malaysia 1969)", 6377295.664, 300.8017      },
  { "Everest (Malay. & Sing)", 6377304.063, 300.8017      },
  { "Everest (Pakistan)",      6377309.613, 300.8017      },
  { "Modified Fischer 1960",   6378155,     298.3         },
 	{ "Fischer 1968",            6378150,     298.3         },
  { "Helmert 1906",            6378200,     298.3         },
  { "Hough 1960",              6378270,     297           }, 
  { "Indonesian 1974",         6378160,     298.247       }, 
  { "International 1924",      6378388,     297           },
  { "Krassovsky 1940",         6378245,     298.3         },
	{ "GRS 1967",                6378160,     298.247182089 },
  { "GRS 80",                  6378137,     298.257222101 },
  { "South American 1969",     6378160,     298.25        },
  { "WGS 60",                  6378165,     298.3         }, 
  { "WGS 66",                  6378145,     298.25        },
  { "WGS 72",                  6378135,     298.26        },
  { "WGS 84",                  6378137,     298.257223563 },
  { "IAU 1976",                6378140,     298.257       },
};

struct CReferenceDatum
{
  char*              Name;       //The name of the datum
  ReferenceEllipsoid elipsoid;   //The elipsoid to use for this datum
  double             dx;         //Change in X value from to WGS-84 datum
  double             dy;         //Change in Y value from to WGS-84 datum
  double             dz;         //Change in Z value from to WGS-84 datum
  double             ex;         //Estimate of the error in dx in meters (A value of -1 means that the error estimate is unknown)
  double             ey;         //Estimate of the error in dx in meters (A value of -1 means that the error estimate is unknown) 
  double             ez;         //Estimate of the error in dy in meters (A value of -1 means that the error estimate is unknown)  
  long               Satellites; //Number of satellite measurement stations
};

enum ReferenceDatum //These enums correspond to the index of the corresponding values in g_ReferenceDatums
{                   //You can use this list as a reference when calling GIS XPs which require a Datum ID
  ADINDAN_BURKINA_FASO = 0,
  ADINDAN_CAMEROON = 1,
  ADINDAN_ETHIOPIA = 2,
  ADINDAN_MALI = 3,
  ADINDAN_MEAN_FOR_ETHIOPIA_SUDAN = 4,
  ADINDAN_SENEGAL = 5,
  ADINDAN_SUDAN = 6,
  AFGOOYE_SOMALIA = 7,
  AIN_EL_ABD_1970_BAHRAIN = 8,
  AIN_EL_ABD_1970_SAUDI_ARABIA = 9,
  AMERICAN_SAMOA_1962_AMERICAN_SAMOA_ISLANDS = 10,
  ANNA_1_ASTRO_1965_COCOS_ISLANDS = 11,
  ANTIGUA_ISLAND_ASTRO_1943_ANTIGUA_LEEWARD_ISLANDS = 12,
  ARC_1950_BOTSWANA = 13,
  ARC_1950_BURUNDI = 14,
  ARC_1950_LESOTHO = 15,
  ARC_1950_MALAWI = 16,
  ARC_1950_MEAN_FOR_BOTSWANA_LESOTHO_MALAWI_SWAZILAND_ZAIRE_ZAMBIA_ZIMBABWE = 17,
  ARC_1950_SWAZILAND = 18,
  ARC_1950_ZAIRE = 19,
  ARC_1950_ZAMBIA = 20,
  ARC_1950_ZIMBABWE = 21,
  ARC_1960_MEAN_FOR_KENYA_TANZANIA = 22,
  ARC_1960_KENYA = 23,
  ARC_1960_TAANZANIA = 24,
  ASCENSION_ISLAND_1958_ASCENSION_ISLAND = 25,
  ASTRO_BEACON_E_1945_IWO_JIMA = 26,
  ASTRO_DOS_71_4_ST_HELENA_ISLAND = 27,
  ASTRO_TERN_ISLAND_FRIG_1961_TERN_ISLAND = 28,
  ASTRONOMICAL_STATION_1952_MARCUS_ISLAND = 29,
  AUSTRALIAN_GEODETIC_1966_AUSTRALIA_TASMANIA = 30,
  AUSTRALIAN_GEODETIC_1984_AUSTRALIA_TASMANIA = 31,
  AYABELLE_LIGHTHOUSE_DJIBOUTI = 32,
  BELLEVUE_IGN_EFATE_ERROMANGO_ISLANDS = 33,
  BERMUDA_1957_BERMUDA = 34,
  BISSAU_GUINEA_BISSAU = 35,
  BOGOTA_OBSERVATORY_COLOMBIA = 36,
  BUKIT_RIMPAH_INDONESIA_BANGKA_BELITUNG_IDS = 37,
  CAMP_AREA_ASTRO_ANTARCTICA_MCMURDO_CAMP_AREA = 38,
  CAMPO_INCHAUSPE_ARGENTINA = 39,
  CANTON_ASTRO_1966_PHOENIX_ISLANDS = 40,
  CAPE_SOUTH_AFRICA = 41,
  CAPE_CANAVERAL_BAHAMAS_FLORIDA = 42,
  CARTHAGE_TUNISIA = 43,
  CHATHAM_ISLAND_ASTRO_1971_NEW_ZEALAND_CHATHAM_ISLAND = 44,
  CHUA_ASTRO_PARAGUAY = 45,
  CORREGO_ALEGRE_BRAZIL = 46,
  DABOLA_GUINEA = 47,
  DECEPTION_ISLAND_DECEPTION_ISLAND_ANTARCTIA = 48,
  DJAKARTA_BATAVIA_INDONESIA_SUMATRA = 49,
  DOS_1968_NEW_GEORGIA_ISLANDS_GIZO_ISLAND = 50,
  EASTER_ISLAND_1967_EASTER_ISLAND = 51,
  ESTONIA_COORDINATE_SYSTEM_1937_ESTONIA = 52,
  EUROPEAN_1950_CYPRUS = 53,
  EUROPEAN_1950_EGYPT = 54,
  EUROPEAN_1950_ENGLAND_CHANNEL_ISLANDS_SCOTLAND_SHETLAND_ISLANDS = 55,
  EUROPEAN_1950_ENGLAND_IRELAND_SCOTLAND_SHETLAND_ISLANDS = 56,
  EUROPEAN_1950_FINLAND_NORWAY = 57,
  EUROPEAN_1950_GREECE = 58,
  EUROPEAN_1950_IRAN = 59,
  EUROPEAN_1950_ITALY_SARDINIA = 60,
  EUROPEAN_1950_ITALY_SICILY = 61,
  EUROPEAN_1950_MALTA = 62,
  EUROPEAN_1950_MEAN_FOR_AUSTRIA_BELGIUM_DENMARK_FINLAND_FRANCE_W_GERMANY_GIBRALTAR_GREECE_ITALY_LUXEMBOURG_NETHERLANDS_NORWAY_PORTUGAL_SPAIN_SWEDEN_SWITZERLAND = 63,
  EUROPEAN_1950_MEAN_FOR_AUSTRIA_DENMARK_FRANCE_W_GERMANY_NETHERLANDS_SWITZERLAND = 64,
  EUROPEAN_1950_MEAN_FOR_IRAQ_ISRAEL_JORDAN_LEBANON_KUWAIT_SAUDI_ARABIA_SYRIA = 65,
  EUROPEAN_1950_PORTUGAL_SPAIN = 66,
  EUROPEAN_1950_TUNISIA = 67,
  EUROPEAN_1979_MEAN_FOR_AUSTRIA_FINLAND_NETHERLANDS_NORWAY_SPAIN_SWEDEN_SWITZERLAND = 68,
  FORT_THOMAS_1955_NEVIS_ST_KITTS_LEEWARD_ISLANDS = 69,
  GAN_1970_REPUBLIC_OF_MALDIVES = 70,
  GEODETIC_DATUM_1949_NEW_ZEALAND = 71,
  GRACIOSA_BASE_SW_1948_AZORES_FAIAL_GRACIOSA_PICO_SAO_JORGE_TERCEIRA = 72,
  GUAM_1963_GUAM = 73,
  GUNUNG_SEGARA_INDONESIA_KALIMANTAN = 74,
  GUX_1_ASTRO_GUADALCANAL_ISLAND = 75,
  HERAT_NORTH_AFGHANISTAN = 76,
  HERMANNSKOGEL_DATUM_CROATIA_SERBIA_BOSNIA_HERZEGOVINA = 77,
  HJORSEY_1955_ICELAND = 78,
  HONG_KONG_1963_HONG_KONG = 79,
  HU_TZU_SHAN_TAIWAN= 80,
  INDIAN_BANGLADESH = 81,
  INDIAN_INDIA_NEPAL = 82,
  INDIAN_PAKISTAN = 83,
  INDIAN_1954_THAILAND = 84,
  INDIAN_1960_VIETNAM_CON_SON_ISLAND = 85,
  INDIAN_1960_VIETNAM_NEAR_16_N = 86,
  INDIAN_1975_THAILAND = 87,
  INDONESIAN_1974_INDONESIA = 88,
  IRELAND_1965_IRELAND = 89,
  ISTS_061_ASTRO_1968_SOUTH_GEORGIA_ISLANDS = 90,
  ISTS_073_ASTRO_1969_DIEGO_GARCIA = 91,
  JOHNSTON_ISLAND_1961_JOHNSTON_ISLAND = 92,
  KANDAWALA_SRI_LANKA = 93,
  KERGUELEN_ISLAND_1949_KERGUELEN_ISLAND = 94,
  KERTAU_1948_WEST_MALAYSIA_SINGAPORE = 95,
  KUSAIE_ASTRO_1951_CAROLINE_ISLANDS = 96,
  KOREAN_GEODETIC_SYSTEM_SOUTH_KOREA = 97,
  L_C_5_ASTRO_1961_CAYMAN_BRAC_ISLAND = 98,
  LEIGON_GHANA = 99,
  LIBERIA_1964_LIBERIA = 100,
  LUZON_PHILIPPINES_EXCLUDING_MINDANAO = 101,
  LUZON_PHILIPPINES_MINDANAO = 102,
  M_PORALOKO_GABON = 103,
  MAHE_1971_MAHE_ISLAND = 104,
  MASSAWA_ETHIOPIA_ERITREA = 105,
  MERCHICH_MOROCCO = 106,
  MIDWAY_ASTRO_1961_MIDWAY_ISLANDS = 107,
  MINNA_CAMEROON = 108,
  MINNA_NIGERIA = 109,
  MONTSERRAT_ISLAND_ASTRO_1958_MONTSERRAT_LEEWARD_ISLANDS = 110,
  NAHRWAN_OMAN_MASIRAH_ISLAND = 111,
  NAHRWAN_SAUDI_ARABIA = 112,
  NAHRWAN_UNITED_ARAB_EMIRATES = 113,
  NAPARIMA_BWI_TRINIDAD_TOBAGO = 114,
  NORTH_AMERICAN_1927_ALASKA_EXCLUDING_ALEUTIAN_IDS = 115,
  NORTH_AMERICAN_1927_ALASKA_ALEUTIAN_IDS_EAST_OF_180_W = 116,
  NORTH_AMERICAN_1927_ALASKA_ALEUTIAN_IDS_WEST_OF_180_W = 117,
  NORTH_AMERICAN_1927_BAHAMAS_EXCEPT_SAN_SALVADOR_ID = 118,
  NORTH_AMERICAN_1927_BAHAMAS_SAN_SALVADOR_ISLAND = 119,
  NORTH_AMERICAN_1927_CANADA_ALBERTA_BRITISH_COLUMBIA = 120,
  NORTH_AMERICAN_1927_CANADA_MANITOBA_ONTARIO = 121,
  NORTH_AMERICAN_1927_CANADA_NEW_BRUNSWICK_NEWFOUNDLAND_NOVA_SCOTIA_QUEBEC = 122,
  NORTH_AMERICAN_1927_CANADA_NORTHWEST_TERRITORIES_SASKATCHEWAN = 123,
  NORTH_AMERICAN_1927_CANADA_YUKON = 124,
  NORTH_AMERICAN_1927_CANAL_ZONE = 125,
  NORTH_AMERICAN_1927_CUBA = 126,
  NORTH_AMERICAN_1927_GREENLAND_HAYES_PENINSULA = 127,
  NORTH_AMERICAN_1927_MEAN_FOR_ANTIGUA_BARBADOS_BARBUDA_CAICOS_ISLANDS_CUBA_DOMINICAN_REPUBLIC_GRAND_CAYMAN_JAMAICA_TURKS_ISLANDS = 128,
  NORTH_AMERICAN_1927_MEAN_FOR_BELIZE_COSTA_RICA_EL_SALVADOR_GUATEMALA_HONDURAS_NICARAGUA = 129,
  NORTH_AMERICAN_1927_MEAN_FOR_CANADA = 130,
  NORTH_AMERICAN_1927_MEAN_FOR_CONUS = 131,
  NORTH_AMERICAN_1927_MEAN_FOR_CONUS_EAST_OF_MISSISSIPPI_RIVER_INCLUDING_LOUISIANA_MISSOURI_MINNESOTA = 132,
  NORTH_AMERICAN_1927_MEAN_FOR_CONUS_WEST_OF_MISSISSIPPI_RIVER_EXCLUDING_LOUISIANA_MINNESOTA_MISSOURI = 133,
  NORTH_AMERICAN_1927_MEXICO = 134,
  NORTH_AMERICAN_1983_ALASKA_EXCLUDING_ALEUTIAN_IDS = 135,
  NORTH_AMERICAN_1983_ALEUTIAN_IDS = 136,
  NORTH_AMERICAN_1983_CANADA = 137,
  NORTH_AMERICAN_1983_CONUS = 138,
  NORTH_AMERICAN_1983_HAWAII = 139,
  NORTH_AMERICAN_1983_MEXICO_CENTRAL_AMERICA = 140,
  NORTH_SAHARA_1959_ALGERIA = 141,
  OBSERVATORIO_METEOROLOGICO_1939_AZORES_CORVO_FLORES_ISLANDS = 142,
  OLD_EGYPTIAN_1907_EGYPT = 143,
  OLD_HAWAIIAN_HAWAII = 144,
  OLD_HAWAIIAN_KAUAI = 145,
  OLD_HAWAIIAN_MAUI = 146,
  OLD_HAWAIIAN_MEAN_FOR_HAWAII_KAUAI_MAUI_OAHU = 147,
  OLD_HAWAIIAN_OAHU = 148,
  OMAN_OMAN = 149,
  ORDNANCE_SURVEY_GREAT_BRITAIN_1936_ENGLAND = 150,
  ORDNANCE_SURVEY_GREAT_BRITAIN_1936_ENGLAND_ISLE_OF_MAN_WALES = 151,
  ORDNANCE_SURVEY_GREAT_BRITAIN_1936_MEAN_FOR_ENGLAND_ISLE_OF_MAN_SCOTLAND_SHETLAND_ISLANDS_WALES = 152,
  ORDNANCE_SURVEY_GREAT_BRITAIN_1936_SCOTLAND_SHETLAND_ISLANDS = 153,
  ORDNANCE_SURVEY_GREAT_BRITAIN_1936_WALES = 154,
  PICO_DE_LAS_NIEVES_CANARY_ISLANDS = 155,
  PITCAIRN_ASTRO_1967_PITCAIRN_ISLAND = 156,
  POINT_58_MEAN_FOR_BURKINA_FASO_NIGER = 157,
  POINTE_NOIRE_1948_CONGO = 158,
  PORTO_SANTO_1936_PORTO_SANTO_MADEIRA_ISLANDS = 159,
  PROVISIONAL_SOUTH_AMERICAN_1956_BOLIVIA = 160,
  PROVISIONAL_SOUTH_AMERICAN_1956_CHILE_NORTHERN_NEAR_19_S = 161,
  PROVISIONAL_SOUTH_AMERICAN_1956_CHILE_SOUTHERN_NEAR_43_S = 162,
  PROVISIONAL_SOUTH_AMERICAN_1956_COLOMBIA = 163,
  PROVISIONAL_SOUTH_AMERICAN_1956_ECUADOR = 164,
  PROVISIONAL_SOUTH_AMERICAN_1956_GUYANA = 165,
  PROVISIONAL_SOUTH_AMERICAN_1956_MEAN_FOR_BOLIVIA_CHILE_COLOMBIA_ECUADOR_GUYANA_PERU_VENEZUELA = 166,
  PROVISIONAL_SOUTH_AMERICAN_1956_PERU = 167,
  PROVISIONAL_SOUTH_AMERICAN_1956_VENEZUELA = 168,
  PROVISIONAL_SOUTH_CHILEAN_1963_CHILE_NEAR_53_S_HITO_XVIII = 169,
  PUERTO_RICO_PUERTO_RICO_VIRGIN_ISLANDS = 170,
  PULKOVO_1942_RUSSIA = 171,
  QATAR_NATIONAL_QATAR = 172,
  QORNOQ_GREENLAND_SOUTH = 173,
  REUNION_MASCARENE_ISLANDS = 174,
  ROME_1940_ITALY_SARDINIA = 175,
  S_42_PULKOVO_1942_HUNGARY = 176,
  S_42_PULKOVO_1942_POLAND = 177,
  S_42_PULKOVO_1942_CZECHOSLAVAKIA = 178,
  S_42_PULKOVO_1942_LATVIA = 179,
  S_42_PULKOVO_1942_KAZAKHSTAN = 180,
  S_42_PULKOVO_1942_ALBANIA = 181,
  S_42_PULKOVO_1942_ROMANIA = 182,
  S_JTSK_CZECHOSLAVAKIA_PRIOR_1_JAN_1993 = 183,
  SANTO_DOS_1965_ESPIRITO_SANTO_ISLAND = 184,
  SAO_BRAZ_AZORES_SAO_MIGUEL_SANTA_MARIA_IDS = 185,
  SAPPER_HILL_1943_EAST_FALKLAND_ISLAND = 186,
  SCHWARZECK_NAMIBIA = 187,
  SELVAGEM_GRANDE_1938_SALVAGE_ISLANDS = 188,
  SIERRA_LEONE_1960_SIERRA_LEONE = 189,
  SOUTH_AMERICAN_1969_ARGENTINA = 190,
  SOUTH_AMERICAN_1969_BOLIVIA = 191,
  SOUTH_AMERICAN_1969_BRAZIL = 192,
  SOUTH_AMERICAN_1969_CHILE = 193,
  SOUTH_AMERICAN_1969_COLOMBIA = 194,
  SOUTH_AMERICAN_1969_ECUADOR = 195,
  SOUTH_AMERICAN_1969_ECUADOR_BALTRA_GALAPAGOS = 196,
  SOUTH_AMERICAN_1969_GUYANA = 197,
  SOUTH_AMERICAN_1969_MEAN_FOR_ARGENTINA_BOLIVIA_BRAZIL_CHILE_COLOMBIA_ECUADOR_GUYANA_PARAGUAY_PERU_TRINIDAD_TOBAGO_VENEZUELA = 198,
  SOUTH_AMERICAN_1969_PARAGUAY = 199,
  SOUTH_AMERICAN_1969_PERU = 200,
  SOUTH_AMERICAN_1969_TRINIDAD_TOBAGO = 201,
  SOUTH_AMERICAN_1969_VENEZUELA = 202,
  SOUTH_ASIA_SINGAPORE = 203,
  TANANARIVE_OBSERVATORY_1925_MADAGASCAR = 204,
  TIMBALAI_1948_BRUNEI_E_MALAYSIA_SABAH_SARAWAK = 205,
  TOKYO_JAPAN = 206,
  TOKYO_MEAN_FOR_JAPAN_SOUTH_KOREA_OKINAWA = 207,
  TOKYO_OKINAWA = 208,
  TOKYO_SOUTH_KOREA = 209,
  TRISTAN_ASTRO_1968_TRISTAN_DA_CUNHA = 210,
  VITI_LEVU_1916_FIJI_VITI_LEVU_ISLAND = 211,
  VOIROL_1960_ALGERIA = 212,
  WAKE_ISLAND_ASTRO_1952_WAKE_ATOLL = 213,
  WAKE_ENIWETOK_1960_MARSHALL_ISLANDS = 214,
  WGS_1972_GLOBAL_DEFINITION = 215,
  WGS_1984_GLOBAL_DEFINITION = 216,
  YACARE_URUGUAY = 217,
  ZANDERIJ_SURINAME = 218
};

CReferenceDatum g_ReferenceDatums[] =
{
  { "Adindan (Burkina Faso)",CLARKE_1880,-118,-14,218,25,25,25,1 },
  { "Adindan (Cameroon)",CLARKE_1880,-134,-2,210,25,25,25,1 },
  { "Adindan (Ethiopia)",CLARKE_1880,-165,-11,206,3,3,3,8 },
  { "Adindan (Mali)",CLARKE_1880,-123,-20,220,25,25,25,1 },
  { "Adindan (MEAN FOR Ethiopia; Sudan)",CLARKE_1880,-166,-15,204,5,5,3,2 },
  { "Adindan (Senegal)",CLARKE_1880,-128,-18,224,25,25,25,2 },
  { "Adindan (Sudan)",CLARKE_1880,-161,-14,205,3,5,3,4 },
  { "Afgooye (Somalia)",KRASSOVSKY_1940,-43,-163,45,25,25,25,1 },
  { "Ain el Abd 1970 (Bahrain)",INTERNATIONAL_1924,-150,-250,-1,25,25,25,2 },
  { "Ain el Abd 1970 (Saudi Arabia)",INTERNATIONAL_1924,-143,-236,7,10,10,10,9 },
  { "American Samoa 1962 (American Samoa Islands)",CLARKE_1866,-115,118,426,25,25,25,2 },
  { "Anna 1 Astro 1965 (Cocos Islands)",AUSTRALIAN_NATIONAL,-491,-22,435,25,25,25,1 },
  { "Antigua Island Astro 1943 (Antigua (Leeward Islands))",CLARKE_1880,-270,13,62,25,25,25,1 },
  { "Arc 1950 (Botswana)",CLARKE_1880,-138,-105,-289,3,5,3,9 },
  { "Arc 1950 (Burundi)",CLARKE_1880,-153,-5,-292,20,20,20,3 },
  { "Arc 1950 (Lesotho)",CLARKE_1880,-125,-108,-295,3,3,8,5 },
  { "Arc 1950 (Malawi)",CLARKE_1880,-161,-73,-317,9,24,8,6 },
  { "Arc 1950 (MEAN FOR Botswana; Lesotho; Malawi; Swaziland; Zaire; Zambia; Zimbabwe)",CLARKE_1880,-143,-90,-294,20,33,20,1 },
  { "Arc 1950 (Swaziland)",CLARKE_1880,-134,-105,-295,15,15,15,4 },
  { "Arc 1950 (Zaire)",CLARKE_1880,-169,-19,-278,25,25,25,2 },
  { "Arc 1950 (Zambia)",CLARKE_1880,-147,-74,-283,21,21,27,5 },
  { "Arc 1950 (Zimbabwe)",CLARKE_1880,-142,-96,-293,5,8,11,0 },
  { "Arc 1960 (MEAN FOR Kenya; Tanzania)",CLARKE_1880,-160,-6,-302,20,20,20,5 },
  { "Arc 1960 (Kenya)",CLARKE_1880,-157,-2,-299,4,3,3,4 },
  { "Arc 1960 (Taanzania)",CLARKE_1880,-175,-23,-303,6,9,10,2 },
  { "Ascension Island 1958 (Ascension Island)",INTERNATIONAL_1924,-205,107,53,25,25,25,2 },
  { "Astro Beacon E 1945 (Iwo Jima)",INTERNATIONAL_1924,145,75,-272,25,25,25,1 },
  { "Astro DOS 71/4 (St Helena Island)",INTERNATIONAL_1924,-320,550,-494,25,25,25,1 },
  { "Astro Tern Island (FRIG) 1961 (Tern Island)",INTERNATIONAL_1924,114,-116,-333,25,25,25,1 },
  { "Astronomical Station 1952 (Marcus Island)",INTERNATIONAL_1924,124,-234,-25,25,25,25,1 },
  { "Australian Geodetic 1966 (Australia; Tasmania)",AUSTRALIAN_NATIONAL,-133,-48,148,3,3,3,5 },
  { "Australian Geodetic 1984 (Australia; Tasmania)",AUSTRALIAN_NATIONAL,-134,-48,149,2,2,2,0 },
  { "Ayabelle Lighthouse (Djibouti)",CLARKE_1880,-79,-129,145,25,25,25,1 },
  { "Bellevue (IGN) (Efate & Erromango Islands)",INTERNATIONAL_1924,-127,-769,472,20,20,20,3 },
  { "Bermuda 1957 (Bermuda)",CLARKE_1866,-73,213,296,20,20,20,3 },
  { "Bissau (Guinea-Bissau)",INTERNATIONAL_1924,-173,253,27,25,25,25,2 },
  { "Bogota Observatory (Colombia)",INTERNATIONAL_1924,307,304,-318,6,5,6,7 },
  { "Bukit Rimpah (Indonesia (Bangka & Belitung Ids))",BESSEL_1841,-384,664,-48,-1,-1,-1,0 },
  { "Camp Area Astro (Antarctica (McMurdo Camp Area))",INTERNATIONAL_1924,-104,-129,239,-1,-1,-1,0 },
  { "Campo Inchauspe (Argentina)",INTERNATIONAL_1924,-148,136,90,5,5,5,0 },
  { "Canton Astro 1966 (Phoenix Islands)",INTERNATIONAL_1924,298,-304,-375,15,15,15,4 },
  { "Cape (South Africa)",CLARKE_1880,-136,-108,-292,3,6,6,5 },
  { "Cape Canaveral (Bahamas; Florida)",CLARKE_1866,-2,151,181,3,3,3,9 },
  { "Carthage (Tunisia)",CLARKE_1880,-263,6,431,6,9,8,5 },
  { "Chatham Island Astro 1971 (New Zealand (Chatham Island))",INTERNATIONAL_1924,175,-38,113,15,15,15,4 },
  { "Chua Astro (Paraguay)",INTERNATIONAL_1924,-134,229,-29,6,9,5,6 },
  { "Corrego Alegre (Brazil)",INTERNATIONAL_1924,-206,172,-6,5,3,5,7 },
  { "Dabola (Guinea)",CLARKE_1880,-83,37,124,15,15,15,4 },
  { "Deception Island (Deception Island; Antarctia)",CLARKE_1880,260,12,-147,20,20,20,3 },
  { "Djakarta (Batavia) (Indonesia (Sumatra))",BESSEL_1841,-377,681,-50,3,3,3,5 },
  { "DOS 1968 (New Georgia Islands (Gizo Island))",INTERNATIONAL_1924,230,-199,-752,25,25,25,1 },
  { "Easter Island 1967 (Easter Island)",INTERNATIONAL_1924,211,147,111,25,25,25,1 },
  { "Estonia; Coordinate System 1937 (Estonia)",BESSEL_1841,374,150,588,2,3,3,9 },
  { "European 1950 (Cyprus)",INTERNATIONAL_1924,-104,-101,-140,15,15,15,4 },
  { "European 1950 (Egypt)",INTERNATIONAL_1924,-130,-117,-151,6,8,8,4 },
  { "European 1950 (England; Channel Islands; Scotland; Shetland Islands)",INTERNATIONAL_1924,-86,-96,-120,3,3,3,0 },
  { "European 1950 (England; Ireland; Scotland; Shetland Islands)",INTERNATIONAL_1924,-86,-96,-120,3,3,3,7 },
  { "European 1950 (Finland; Norway)",INTERNATIONAL_1924,-87,-95,-120,3,5,3,0 },
  { "European 1950 (Greece)",INTERNATIONAL_1924,-84,-95,-130,25,25,25,2 },
  { "European 1950 (Iran)",INTERNATIONAL_1924,-117,-132,-164,9,12,11,7 },
  { "European 1950 (Italy (Sardinia))",INTERNATIONAL_1924,-97,-103,-120,25,25,25,2 },
  { "European 1950 (Italy (Sicily))",INTERNATIONAL_1924,-97,-88,-135,20,20,20,3 },
  { "European 1950 (Malta)",INTERNATIONAL_1924,-107,-88,-149,25,25,25,1 },
  { "European 1950 (MEAN FOR Austria; Belgium; Denmark; Finland; France; W Germany; Gibraltar; Greece; Italy; Luxembourg; Netherlands; Norway; Portugal; Spain; Sweden; Switzerland)",INTERNATIONAL_1924,-87,-98,-121,3,8,5,5 },
  { "European 1950 (MEAN FOR Austria; Denmark; France; W Germany; Netherlands; Switzerland)",INTERNATIONAL_1924,-87,-96,-120,3,3,3,2 },
  { "European 1950 (MEAN FOR Iraq; Israel; Jordan; Lebanon; Kuwait; Saudi Arabia; Syria)",INTERNATIONAL_1924,-103,-106,-141,-1,-1,-1,0 },
  { "European 1950 (Portugal; Spain)",INTERNATIONAL_1924,-84,-107,-120,5,6,3,8 },
  { "European 1950 (Tunisia)",INTERNATIONAL_1924,-112,-77,-145,25,25,25,4 },
  { "European 1979 (MEAN FOR Austria; Finland; Netherlands; Norway; Spain; Sweden; Switzerland)",INTERNATIONAL_1924,-86,-98,-119,3,3,3,2 },
  { "Fort Thomas 1955 (Nevis; St. Kitts (Leeward Islands))",CLARKE_1880,-7,215,225,25,25,25,2 },
  { "Gan 1970 (Republic of Maldives)",INTERNATIONAL_1924,-133,-321,50,25,25,25,1 },
  { "Geodetic Datum 1949 (New Zealand)",INTERNATIONAL_1924,84,-22,209,5,3,5,4 },
  { "Graciosa Base SW 1948 (Azores (Faial; Graciosa; Pico; Sao Jorge; Terceira))",INTERNATIONAL_1924,-104,167,-38,3,3,3,5 },
  { "Guam 1963 (Guam)",CLARKE_1866,-100,-248,259,3,3,3,5 },
  { "Gunung Segara (Indonesia (Kalimantan))",BESSEL_1841,-403,684,41,-1,-1,-1,0 },
  { "GUX 1 Astro (Guadalcanal Island)",INTERNATIONAL_1924,252,-209,-751,25,25,25,1 },
  { "Herat North (Afghanistan)",INTERNATIONAL_1924,-333,-222,114,-1,-1,-1, },
  { "Hermannskogel Datum (\"Croatia -Serbia, Bosnia-Herzegovina\")",BESSEL_1841_NAMIBIA,653,-212,449,-1,-1,-1,0 },
  { "Hjorsey 1955 (Iceland)",INTERNATIONAL_1924,-73,46,-86,3,3,6,6 },
  { "Hong Kong 1963 (Hong Kong)",INTERNATIONAL_1924,-156,-271,-189,25,25,25,2 },
  { "Hu-Tzu-Shan (Taiwan)",INTERNATIONAL_1924,-637,-549,-203,15,15,15,4 },
  { "Indian (Bangladesh)",EVEREST_INDIA_1830,282,726,254,10,8,12,6 },
  { "Indian (India; Nepal)",EVEREST_INDIA_1956,295,736,257,12,10,15,7 },
  { "Indian (Pakistan)",EVEREST_PAKISTAN,283,682,231,-1,-1,-1,0 },
  { "Indian 1954 (Thailand)",EVEREST_INDIA_1830,217,823,299,15,6,12,1 },
  { "Indian 1960 (Vietnam (Con Son Island))",EVEREST_INDIA_1830,182,915,344,25,25,25,1 },
  { "Indian 1960 (Vietnam (Near 16 N))",EVEREST_INDIA_1830,198,881,317,25,25,25,2 },
  { "Indian 1975 (Thailand)",EVEREST_INDIA_1830,210,814,289,3,2,3,2 },
  { "Indonesian 1974 (Indonesia)",INDONESIAN_1974,-24,-15,5,25,25,25,1 },
  { "Ireland 1965 (Ireland)",MODIFIED_AIRY,506,-122,611,3,3,3,7 },
  { "ISTS 061 Astro 1968 (South Georgia Islands)",INTERNATIONAL_1924,-794,119,-298,25,25,25,1 },
  { "ISTS 073 Astro 1969 (Diego Garcia)",INTERNATIONAL_1924,208,-435,-229,25,25,25,2 },
  { "Johnston Island 1961 (Johnston Island)",INTERNATIONAL_1924,189,-79,-202,25,25,25,1 },
  { "Kandawala (Sri Lanka)",EVEREST_INDIA_1830,-97,787,86,20,20,20,3 },
  { "Kerguelen Island 1949 (Kerguelen Island)",INTERNATIONAL_1924,145,-187,103,25,25,25,1 },
  { "Kertau 1948 (West Malaysia & Singapore)",EVEREST_MALAY_SING,-11,851,5,10,8,6,6 },
  { "Kusaie Astro 1951 (Caroline Islands)",INTERNATIONAL_1924,647,1777,-1124,25,25,25,1 },
  { "Korean Geodetic System (South Korea)",GRS_80,0,0,0,2,2,2,2 },
  { "L. C. 5 Astro 1961 (Cayman Brac Island)",CLARKE_1866,42,124,147,25,25,25,1 },
  { "Leigon (Ghana)",CLARKE_1880,-130,29,364,2,3,2,8 },
  { "Liberia 1964 (Liberia)",CLARKE_1880,-90,40,88,15,15,15,4 },
  { "Luzon (Philippines (Excluding Mindanao))",CLARKE_1866,-133,-77,-51,8,11,9,6 },
  { "Luzon (Philippines (Mindanao))",CLARKE_1866,-133,-79,-72,25,25,25,1 },
  { "M'Poraloko (Gabon)",CLARKE_1880,-74,-130,42,25,25,25,1 },
  { "Mahe 1971 (Mahe Island)",CLARKE_1880,41,-220,-134,25,25,25,1 },
  { "Massawa (Ethiopia (Eritrea))",BESSEL_1841,639,405,60,25,25,25,1 },
  { "Merchich (Morocco)",CLARKE_1880,31,146,47,5,3,3,9 },
  { "Midway Astro 1961 (Midway Islands)",INTERNATIONAL_1924,912,-58,1227,25,25,25,1 },
  { "Minna (Cameroon)",CLARKE_1880,-81,-84,115,25,25,25,2 },
  { "Minna (Nigeria)",CLARKE_1880,-92,-93,122,3,6,5,6 },
  { "Montserrat Island Astro 1958 (Montserrat (Leeward Islands))",CLARKE_1880,174,359,365,25,25,25,1 },
  { "Nahrwan (Oman (Masirah Island))",CLARKE_1880,-247,-148,369,25,25,25,2 },
  { "Nahrwan (Saudi Arabia)",CLARKE_1880,-243,-192,477,20,20,20,3 },
  { "Nahrwan (United Arab Emirates)",CLARKE_1880,-249,-156,381,25,25,25,2 },
  { "Naparima BWI (Trinidad & Tobago)",INTERNATIONAL_1924,-10,375,165,15,15,15,4 },
  { "North American 1927 (Alaska (Excluding Aleutian Ids))",CLARKE_1866,-5,135,172,5,9,5,7 },
  { "North American 1927 (Alaska (Aleutian Ids East of 180 W))",CLARKE_1866,-2,152,149,6,8,10, },
  { "North American 1927 (Alaska (Aleutian Ids West of 180 W))",CLARKE_1866,2,204,105,10,10,10, },
  { "North American 1927 (Bahamas (Except San Salvador Id))",CLARKE_1866,-4,154,178,5,3,5,1 },
  { "North American 1927 (Bahamas (San Salvador Island))",CLARKE_1866,1,140,165,25,25,25,1 },
  { "North American 1927 (Canada (Alberta; British Columbia))",CLARKE_1866,-7,162,188,8,8,6,5 },
  { "North American 1927 (Canada (Manitoba; Ontario))",CLARKE_1866,-9,157,184,9,5,5,5 },
  { "North American 1927 (Canada (New Brunswick; Newfoundland; Nova Scotia; Quebec))",CLARKE_1866,-22,160,190,6,6,3,7 },
  { "North American 1927 (Canada (Northwest Territories; Saskatchewan))",CLARKE_1866,4,159,188,5,5,3,7 },
  { "North American 1927 (Canada (Yukon))",CLARKE_1866,-7,139,181,5,8,3,8 },
  { "North American 1927 (Canal Zone)",CLARKE_1866,0,125,201,20,20,20,3 },
  { "North American 1927 (Cuba)",CLARKE_1866,-9,152,178,25,25,25,1 },
  { "North American 1927 (Greenland (Hayes Peninsula))",CLARKE_1866,11,114,195,25,25,25,2 },
  { "North American 1927 (MEAN FOR Antigua; Barbados; Barbuda; Caicos Islands; Cuba; Dominican Republic; Grand Cayman; Jamaica; Turks Islands)",CLARKE_1866,-3,142,183,3,9,12,5 },
  { "North American 1927 (MEAN FOR Belize; Costa Rica; El Salvador; Guatemala; Honduras; Nicaragua)",CLARKE_1866,0,125,194,8,3,5,9 },
  { "North American 1927 (MEAN FOR Canada)",CLARKE_1866,-10,158,187,15,11,6,2 },
  { "North American 1927 (MEAN FOR CONUS)",CLARKE_1866,-8,160,176,5,5,6,5 },
  { "North American 1927 (MEAN FOR CONUS (East of Mississippi; River Including Louisiana; Missouri; Minnesota))",CLARKE_1866,-9,161,179,5,5,8,9 },
  { "North American 1927 (MEAN FOR CONUS (West of Mississippi; River Excluding Louisiana; Minnesota; Missouri))",CLARKE_1866,-8,159,175,5,3,3,6 },
  { "North American 1927 (Mexico)",CLARKE_1866,-12,130,190,8,6,6,2 },
  { "North American 1983 (Alaska (Excluding Aleutian Ids))",GRS_80,0,0,0,2,2,2,2 },
  { "North American 1983 (Aleutian Ids)",GRS_80,-2,0,4,5,2,5,4 },
  { "North American 1983 (Canada)",GRS_80,0,0,0,2,2,2,6 },
  { "North American 1983 (CONUS)",GRS_80,0,0,0,2,2,2,6 },
  { "North American 1983 (Hawaii)",GRS_80,1,1,-1,2,2,2,6 },
  { "North American 1983 (Mexico; Central America)",GRS_80,0,0,0,2,2,2,5 },
  { "North Sahara 1959 (Algeria)",CLARKE_1880,-186,-93,310,25,25,25,3 },
  { "Observatorio Meteorologico 1939 (Azores (Corvo & Flores Islands))",INTERNATIONAL_1924,-425,-169,81,20,20,20,3 },
  { "Old Egyptian 1907 (Egypt)",HELMERT_1906,-130,110,-13,3,6,8,4 },
  { "Old Hawaiian (Hawaii)",CLARKE_1866,89,-279,-183,25,25,25,2 },
  { "Old Hawaiian (Kauai)",CLARKE_1866,45,-290,-172,20,20,20,3 },
  { "Old Hawaiian (Maui)",CLARKE_1866,65,-290,-190,25,25,25,2 },
  { "Old Hawaiian (MEAN FOR Hawaii; Kauai; Maui; Oahu)",CLARKE_1866,61,-285,-181,25,20,20,5 },
  { "Old Hawaiian (Oahu)",CLARKE_1866,58,-283,-182,10,6,6,8 },
  { "Oman (Oman)",CLARKE_1880,-346,-1,224,3,3,9,7 },
  { "Ordnance Survey Great Britain 1936 (England)",AIRY_1830,371,-112,434,5,5,6,1 },
  { "Ordnance Survey Great Britain 1936 (England; Isle of Man; Wales)",AIRY_1830,371,-111,434,10,10,15,5 },
  { "Ordnance Survey Great Britain 1936 (MEAN FOR England; Isle of Man; Scotland; Shetland Islands; Wales)",AIRY_1830,375,-111,431,10,10,15,8 },
  { "Ordnance Survey Great Britain 1936 (Scotland; Shetland Islands)",AIRY_1830,384,-111,425,10,10,10,3 },
  { "Ordnance Survey Great Britain 1936 (Wales)",AIRY_1830,370,-108,434,20,20,20,3 },
  { "Pico de las Nieves (Canary Islands)",INTERNATIONAL_1924,-307,-92,127,25,25,25,1 },
  { "Pitcairn Astro 1967 (Pitcairn Island)",INTERNATIONAL_1924,185,165,42,25,25,25,1 },
  { "Point 58 (MEAN FOR Burkina Faso & Niger)",CLARKE_1880,-106,-129,165,25,25,25,1 },
  { "Pointe Noire 1948 (Congo)",CLARKE_1880,-148,51,-291,25,25,25,1 },
  { "Porto Santo 1936 (Porto Santo; Madeira Islands)",INTERNATIONAL_1924,-499,-249,314,25,25,25,2 },
  { "Provisional South American 1956 (Bolivia)",INTERNATIONAL_1924,-270,188,-388,5,11,14,5 },
  { "Provisional South American 1956 (Chile (Northern; Near 19 S))",INTERNATIONAL_1924,-270,183,-390,25,25,25,1 },
  { "Provisional South American 1956 (Chile (Southern; Near 43 S))",INTERNATIONAL_1924,-305,243,-442,20,20,20,3 },
  { "Provisional South American 1956 (Colombia)",INTERNATIONAL_1924,-282,169,-371,15,15,15,4 },
  { "Provisional South American 1956 (Ecuador)",INTERNATIONAL_1924,-278,171,-367,3,5,3,1 },
  { "Provisional South American 1956 (Guyana)",INTERNATIONAL_1924,-298,159,-369,6,14,5,9 },
  { "Provisional South American 1956 (MEAN FOR Bolivia; Chile; Colombia; Ecuador; Guyana; Peru; Venezuela)",INTERNATIONAL_1924,-288,175,-376,17,27,27,3 },
  { "Provisional South American 1956 (Peru)",INTERNATIONAL_1924,-279,175,-379,6,8,12,6 },
  { "Provisional South American 1956 (Venezuela)",INTERNATIONAL_1924,-295,173,-371,9,14,15,4 },
  { "Provisional South Chilean 1963 (Chile (Near 53 S) (Hito XVIII))",INTERNATIONAL_1924,16,196,93,25,25,25,2 },
  { "Puerto Rico (Puerto Rico; Virgin Islands)",CLARKE_1866,11,72,-101,3,3,3,1 },
  { "Pulkovo 1942 (Russia)",KRASSOVSKY_1940,28,-130,-95,-1,-1,-1,0 },
  { "Qatar National (Qatar)",INTERNATIONAL_1924,-128,-283,22,20,20,20,3 },
  { "Qornoq (Greenland (South))",INTERNATIONAL_1924,164,138,-189,25,25,32,2 },
  { "Reunion (Mascarene Islands)",INTERNATIONAL_1924,94,-948,-1262,25,25,25,1 },
  { "Rome 1940 (Italy (Sardinia))",INTERNATIONAL_1924,-225,-65,9,25,25,25,1 },
  { "S-42 (Pulkovo 1942) (Hungary)",KRASSOVSKY_1940,28,-121,-77,2,2,2,5 },
  { "S-42 (Pulkovo 1942) (Poland)",KRASSOVSKY_1940,23,-124,-82,4,2,4,1 },
  { "S-42 (Pulkovo 1942) (Czechoslavakia)",KRASSOVSKY_1940,26,-121,-78,3,3,2,6 },
  { "S-42 (Pulkovo 1942) (Latvia)",KRASSOVSKY_1940,24,-124,-82,2,2,2,5 },
  { "S-42 (Pulkovo 1942) (Kazakhstan)",KRASSOVSKY_1940,15,-130,-84,25,25,25,2 },
  { "S-42 (Pulkovo 1942) (Albania)",KRASSOVSKY_1940,24,-130,-92,3,3,3,7 },
  { "S-42 (Pulkovo 1942) (Romania)",KRASSOVSKY_1940,28,-121,-77,3,5,3,4 },
  { "S-JTSK (Czechoslavakia (Prior 1 JAN 1993))",BESSEL_1841,589,76,480,4,2,3,6 },
  { "Santo (DOS) 1965 (Espirito Santo Island)",INTERNATIONAL_1924,170,42,84,25,25,25,1 },
  { "Sao Braz (Azores (Sao Miguel; Santa Maria Ids))",INTERNATIONAL_1924,-203,141,53,25,25,25,2 },
  { "Sapper Hill 1943 (East Falkland Island)",INTERNATIONAL_1924,-355,21,72,1,1,1,5 },
  { "Schwarzeck (Namibia)",BESSEL_1841_NAMIBIA,616,97,-251,20,20,20,3 },
  { "Selvagem Grande 1938 (Salvage Islands)",INTERNATIONAL_1924,-289,-124,60,25,25,25,1 },
  { "Sierra Leone 1960 (Sierra Leone)",CLARKE_1880,-88,4,101,15,15,15,8 },
  { "South American 1969 (Argentina)",SOUTH_AMERICAN_1969,-62,-1,-37,5,5,5,0 },
  { "South American 1969 (Bolivia)",SOUTH_AMERICAN_1969,-61,2,-48,15,15,15,4 },
  { "South American 1969 (Brazil)",SOUTH_AMERICAN_1969,-60,-2,-41,3,5,5,2 },
  { "South American 1969 (Chile)",SOUTH_AMERICAN_1969,-75,-1,-44,15,8,11,9 },
  { "South American 1969 (Colombia)",SOUTH_AMERICAN_1969,-44,6,-36,6,6,5,7 },
  { "South American 1969 (Ecuador)",SOUTH_AMERICAN_1969,-48,3,-44,3,3,3,1 },
  { "South American 1969 (Ecuador (Baltra; Galapagos))",SOUTH_AMERICAN_1969,-47,26,-42,25,25,25,1 },
  { "South American 1969 (Guyana)",SOUTH_AMERICAN_1969,-53,3,-47,9,5,5,5 },
  { "South American 1969 (MEAN FOR Argentina; Bolivia; Brazil; Chile; Colombia; Ecuador; Guyana; Paraguay; Peru; Trinidad & Tobago; Venezuela)",SOUTH_AMERICAN_1969,-57,1,-41,15,6,9,4 },
  { "South American 1969 (Paraguay)",SOUTH_AMERICAN_1969,-61,2,-33,15,15,15,4 },
  { "South American 1969 (Peru)",SOUTH_AMERICAN_1969,-58,0,-44,5,5,5,6 },
  { "South American 1969 (Trinidad & Tobago)",SOUTH_AMERICAN_1969,-45,12,-33,25,25,25,1 },
  { "South American 1969 (Venezuela)",SOUTH_AMERICAN_1969,-45,8,-33,3,6,3,5 },
  { "South Asia (Singapore)",MODIFIED_FISCHER_1960,7,-10,-26,25,25,25,1 },
  { "Tananarive Observatory 1925 (Madagascar)",INTERNATIONAL_1924,-189,-242,-91,-1,-1,-1,0 },
  { "Timbalai 1948 (Brunei; E. Malaysia (Sabah Sarawak))",EVEREST_SABAH_SARAWAK,-679,669,-48,10,10,12,8 },
  { "Tokyo (Japan)",BESSEL_1841,-148,507,685,8,5,8,6 },
  { "Tokyo (MEAN FOR Japan; South Korea; Okinawa)",BESSEL_1841,-148,507,685,20,5,20,1 },
  { "Tokyo (Okinawa)",BESSEL_1841,-158,507,676,20,5,20,3 },
  { "Tokyo (South Korea)",BESSEL_1841,-147,506,687,2,2,2,9 },
  { "Tristan Astro 1968 (Tristan da Cunha)",INTERNATIONAL_1924,-632,438,-609,25,25,25,1 },
  { "Viti Levu 1916 (Fiji (Viti Levu Island))",CLARKE_1880,51,391,-36,25,25,25,1 },
  { "Voirol 1960 (Algeria)",CLARKE_1880,-123,-206,219,25,25,25,2 },
  { "Wake Island Astro 1952 (Wake Atoll)",INTERNATIONAL_1924,276,-57,149,25,25,25,2 },
  { "Wake-Eniwetok 1960 (Marshall Islands)",HOUGH_1960,102,52,-38,3,3,3,0 },
  { "WGS 1972 (Global Definition)",WGS_72,0,0,0,-1,-1,-1,0 },
  { "WGS 1984 (Global Definition)",WGS_84,0,0,0,-1,-1,-1,0 },
  { "Yacare (Uruguay)",INTERNATIONAL_1924,-155,171,37,-1,-1,-1,0 },
  { "Zanderij (Suriname)",INTERNATIONAL_1924,-265,120,-358,5,5,8,5 },
};

struct CGeodeticPosition
{
  double lon;
  double lat;
  double h;
};



/////////////////////////////// Implementation ////////////////////////////////


// DatumTransform: Implements the Standard Molodensky Datum Transform
//
// Parameters:
//     from:     The geodetic position to be translated.
//     from_a:   The semi-major axis of the "from" ellipsoid.
//     from_f:   Flattening of the "from" ellipsoid.
//     from_esq: Eccentricity-squared of the "from" ellipsoid.
//     da:       Change in semi-major axis length (meters); "to" minus "from"    
//     df:       Change in flattening; "to" minus "from"
//     dx:       Change in x between "from" and "to" datum.
//     dy:       Change in y between "from" and "to" datum.
//     dz:       Change in z between "from" and "to" datum.
CGeodeticPosition DatumTransform(const CGeodeticPosition& from, const double& from_a, const double& from_f, const double& from_esq, 
                                 const double& da, const double& df, const double& dx, const double& dy, double dz)
{
  double slat = sin(from.lat);
  double clat = cos(from.lat);
  double slon = sin(from.lon);
  double clon = cos(from.lon);
  double ssqlat = slat * slat;
  double adb = 1.0 / (1.0 - from_f);  // "a divided by b"

  double rn = from_a / sqrt(1.0 - from_esq * ssqlat);
  double rm = from_a * (1.0 - from_esq) / pow((1.0 - from_esq * ssqlat), 1.5);
  double dlat = (((((-dx * slat * clon - dy * slat * slon) + dz * clat) + (da * ((rn * from_esq * slat * clat) / from_a))) + (df * (rm * adb + rn / adb) * slat * clat))) / (rm + from.h);
  double dlon = (-dx * slon + dy * clon) / ((rn + from.h) * clat);
  double dh = (dx * clat * clon) + (dy * clat * slon) + (dz * slat) - (da * (from_a / rn)) + ((df * rn * ssqlat) / adb);

  CGeodeticPosition position;
  position.lat = from.lat + dlat;
  position.lon = from.lon + dlon;
  position.h = from.h + dh;

  return position;
}

//This routine determines the correct UTM letter designator for the given latitude
//returns 'Z' if latitude is outside the UTM limits of 84N to 80S
//Written by Chuck Gantz- chuck.gantz@globalstar.com
char UTMLetterDesignator(double Lat)
{
	char LetterDesignator;
	if ((84 >= Lat) && (Lat >= 72)) 
    LetterDesignator = 'X';
	else if ((72 > Lat) && (Lat >= 64)) 
    LetterDesignator = 'W';
	else if ((64 > Lat) && (Lat >= 56)) 
    LetterDesignator = 'V';
	else if ((56 > Lat) && (Lat >= 48)) 
    LetterDesignator = 'U';
	else if ((48 > Lat) && (Lat >= 40)) 
    LetterDesignator = 'T';
	else if ((40 > Lat) && (Lat >= 32)) 
    LetterDesignator = 'S';
	else if ((32 > Lat) && (Lat >= 24)) 
    LetterDesignator = 'R';
	else if ((24 > Lat) && (Lat >= 16)) 
    LetterDesignator = 'Q';
	else if ((16 > Lat) && (Lat >= 8)) 
    LetterDesignator = 'P';
	else if (( 8 > Lat) && (Lat >= 0)) 
    LetterDesignator = 'N';
	else if (( 0 > Lat) && (Lat >= -8)) 
    LetterDesignator = 'M';
	else if ((-8> Lat) && (Lat >= -16)) 
    LetterDesignator = 'L';
	else if ((-16 > Lat) && (Lat >= -24)) 
    LetterDesignator = 'K';
	else if ((-24 > Lat) && (Lat >= -32)) 
    LetterDesignator = 'J';
	else if ((-32 > Lat) && (Lat >= -40)) 
    LetterDesignator = 'H';
	else if ((-40 > Lat) && (Lat >= -48)) 
    LetterDesignator = 'G';
	else if ((-48 > Lat) && (Lat >= -56)) 
    LetterDesignator = 'F';
	else if ((-56 > Lat) && (Lat >= -64)) 
    LetterDesignator = 'E';
	else if ((-64 > Lat) && (Lat >= -72)) 
    LetterDesignator = 'D';
	else if ((-72 > Lat) && (Lat >= -80)) 
    LetterDesignator = 'C';
	else 
    LetterDesignator = 'Z'; //This is here as an error flag to show that the Latitude is outside the UTM limits

	return LetterDesignator;
}

//converts lat/long to UTM coords.  Equations from USGS Bulletin 1532 
//East Longitudes are positive, West longitudes are negative. 
//North latitudes are positive, South latitudes are negative
//Lat and Long are in decimal degrees
//Written by Chuck Gantz- chuck.gantz@globalstar.com
void LLtoUTM(int ReferenceEllipsoid, const double& Lat, const double& Long, double &UTMNorthing, double &UTMEasting, char* UTMZone)
{
	double a = g_ReferenceEllipsoids[ReferenceEllipsoid].EquatorialRadius;
  double f = 1 / g_ReferenceEllipsoids[ReferenceEllipsoid].OneOverF;
  double eccSquared = 2*f - f*f;
	double k0 = 0.9996;
	
  //Make sure the longitude is between -180.00 .. 179.9
	double LongTemp = (Long+180)-int((Long+180)/360)*360-180; // -180.00 .. 179.9;

  //Convert to radians
	double LatRad = Lat*g_deg2rad;
	double LongRad = LongTemp*g_deg2rad;

	int ZoneNumber = int((LongTemp + 180)/6) + 1;
  
	if (Lat >= 56.0 && Lat < 64.0 && LongTemp >= 3.0 && LongTemp < 12.0)
		ZoneNumber = 32;

  //Special zones for Svalbard
	if (Lat >= 72.0 && Lat < 84.0) 
	{
	  if (LongTemp >= 0.0  && LongTemp < 9.0) 
      ZoneNumber = 31;
	  else if (LongTemp >= 9.0  && LongTemp < 21.0)
      ZoneNumber = 33;
	  else if (LongTemp >= 21.0 && LongTemp < 33.0) 
      ZoneNumber = 35;
	  else if (LongTemp >= 33.0 && LongTemp < 42.0) 
      ZoneNumber = 37;
	}
	double LongOrigin = (ZoneNumber - 1)*6 - 180 + 3;  //+3 puts origin in middle of zone
	double LongOriginRad = LongOrigin*g_deg2rad;

	//compute the UTM Zone from the latitude and longitude
	sprintf(UTMZone, "%d%c", ZoneNumber, UTMLetterDesignator(Lat));

	double eccPrimeSquared = (eccSquared)/(1-eccSquared);

	double N = a/sqrt(1-eccSquared*sin(LatRad)*sin(LatRad));
	double T = tan(LatRad)*tan(LatRad);
	double C = eccPrimeSquared*cos(LatRad)*cos(LatRad);
	double A = cos(LatRad)*(LongRad-LongOriginRad);

	double M = a*((1 - eccSquared/4	- 3*eccSquared*eccSquared/64 - 5*eccSquared*eccSquared*eccSquared/256)*LatRad 
				     - (3*eccSquared/8	+ 3*eccSquared*eccSquared/32 + 45*eccSquared*eccSquared*eccSquared/1024)*sin(2*LatRad)
						 + (15*eccSquared*eccSquared/256 + 45*eccSquared*eccSquared*eccSquared/1024)*sin(4*LatRad) 
						 - (35*eccSquared*eccSquared*eccSquared/3072)*sin(6*LatRad));
	
	UTMEasting = (double)(k0*N*(A+(1-T+C)*A*A*A/6
              					+ (5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120)
					              + 500000.0);

	UTMNorthing = (double)(k0*(M+N*tan(LatRad)*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24
                				 + (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720)));
	if (Lat < 0)
		UTMNorthing += 10000000.0; //10000000 meter offset for southern hemisphere
}

//converts UTM coords to lat/long.  Equations from USGS Bulletin 1532 
//East Longitudes are positive, West longitudes are negative. 
//North latitudes are positive, South latitudes are negative
//Lat and Long are in decimal degrees. 
//Written by Chuck Gantz- chuck.gantz@globalstar.com
void UTMtoLL(int ReferenceEllipsoid, const double& UTMNorthing, const double& UTMEasting, const char* UTMZone, double& Lat, double& Long)
{
	double k0 = 0.9996;
	double a = g_ReferenceEllipsoids[ReferenceEllipsoid].EquatorialRadius;
  double f = 1 / g_ReferenceEllipsoids[ReferenceEllipsoid].OneOverF;
  double eccSquared = 2*f - f*f;
	double e1 = (1-sqrt(1-eccSquared))/(1+sqrt(1-eccSquared));

	double x = UTMEasting - 500000.0; //remove 500,000 meter offset for longitude
	double y = UTMNorthing;

	char* ZoneLetter;
	int ZoneNumber = strtoul(UTMZone, &ZoneLetter, 10);
	int NorthernHemisphere; //1 for northern hemisphere, 0 for southern
	if ((*ZoneLetter - 'N') >= 0)
		NorthernHemisphere = 1;//point is in northern hemisphere
	else
	{
		NorthernHemisphere = 0;//point is in southern hemisphere
		y -= 10000000.0;//remove 10,000,000 meter offset used for southern hemisphere
	}

	double LongOrigin = (ZoneNumber - 1)*6 - 180 + 3;  //+3 puts origin in middle of zone

	double eccPrimeSquared = (eccSquared)/(1-eccSquared);

	double M = y / k0;
	double mu = M/(a*(1-eccSquared/4-3*eccSquared*eccSquared/64-5*eccSquared*eccSquared*eccSquared/256));

	double phi1Rad = mu	+ (3*e1/2-27*e1*e1*e1/32)*sin(2*mu) 
            			 + (21*e1*e1/16-55*e1*e1*e1*e1/32)*sin(4*mu)
				           +(151*e1*e1*e1/96)*sin(6*mu);

	double N1 = a/sqrt(1-eccSquared*sin(phi1Rad)*sin(phi1Rad));
	double T1 = tan(phi1Rad)*tan(phi1Rad);
	double C1 = eccPrimeSquared*cos(phi1Rad)*cos(phi1Rad);
	double R1 = a*(1-eccSquared)/pow(1-eccSquared*sin(phi1Rad)*sin(phi1Rad), 1.5);
	double D = x/(N1*k0);

	Lat = phi1Rad - (N1*tan(phi1Rad)/R1)*(D*D/2-(5+3*T1+10*C1-4*C1*C1-9*eccPrimeSquared)*D*D*D*D/24
					+(61+90*T1+298*C1+45*T1*T1-252*eccPrimeSquared-3*C1*C1)*D*D*D*D*D*D/720);
	Lat = Lat * g_rad2deg;

	Long = (D-(1+2*T1+C1)*D*D*D/6+(5-2*C1+28*T1-3*C1*C1+8*eccPrimeSquared+24*T1*T1)
					*D*D*D*D*D/120)/cos(phi1Rad);
	Long = LongOrigin + Long * g_rad2deg;
}

double DistanceBetweenPoints(double f, double a, double GeographicalLatitude1, double GeographicalLongitude1, double GeographicalLatitude2, double GeographicalLongitude2)
{
  double F = (GeographicalLatitude1 + GeographicalLatitude2) / 2;
  double G = (GeographicalLatitude1 - GeographicalLatitude2) / 2;
  double lambda = (GeographicalLongitude1 - GeographicalLongitude2) / 2;
  double sinG = sin(G);
  double cosG = cos(G);
  double cosF = cos(F);
  double sinF = sin(F);
  double sinLambda = sin(lambda);
  double cosLambda = cos(lambda);
  double S = (sinG*sinG*cosLambda*cosLambda) + (cosF*cosF*sinLambda*sinLambda);
  double C = (cosG*cosG*cosLambda*cosLambda) + (sinF*sinF*sinLambda*sinLambda);
  double w = atan(sqrt(S/C));
  double R = sqrt(S*C)/w;
  double D = 2*w*a;
  double Hprime = (3*R - 1) / (2*C);
  double Hprime2 = (3*R + 1) / (2*S);

  return D * (1 + (f*Hprime*sinF*sinF*cosG*cosG) - (f*Hprime2*cosF*cosF*sinG*sinG));
}



#define GIS_ELLIPSOIDINFO_ELIPSOID          0
#define GIS_ELLIPSOIDINFO_NAME              1
#define GIS_ELLIPSOIDINFO_EQUATORIAL_RADIUS 2
#define GIS_ELLIPSOIDINFO_ONEOVERF          3

BEGIN_IMPLEMENT_XP_PARAMETERS(CEllipsoidInfoExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(3, NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CEllipsoidInfoExtendedStoredProcedure)

void CEllipsoidInfoExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_ELLIPSOIDINFO @ReferenceEllipsoidID, @EllipsoidName OUTPUT, @EquatorialRadius OUTPUT, @OneOverF OUTPUT");
}

BOOL CEllipsoidInfoExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Validate the Datum value
  DBINT Ellipsoid = XP_INT4(m_pParameterData[GIS_ELLIPSOIDINFO_ELIPSOID].m_Data);
  if (Ellipsoid < 0 || Ellipsoid >= sizeof(g_ReferenceEllipsoids) / sizeof(CReferenceEllipsoid))
  {
    sprintf(szErrorMsg, "%s: Invalid ReferenceEllipsoidID value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Return the data in the output parameters
  ParamSetOutput(GIS_ELLIPSOIDINFO_NAME+1,              (BYTE*) g_ReferenceEllipsoids[Ellipsoid].Name,              (ULONG) strlen(g_ReferenceEllipsoids[Ellipsoid].Name),     FALSE);
  ParamSetOutput(GIS_ELLIPSOIDINFO_EQUATORIAL_RADIUS+1, (BYTE*) &g_ReferenceEllipsoids[Ellipsoid].EquatorialRadius, sizeof(g_ReferenceEllipsoids[Ellipsoid].EquatorialRadius), FALSE);
  ParamSetOutput(GIS_ELLIPSOIDINFO_ONEOVERF+1,          (BYTE*) &g_ReferenceEllipsoids[Ellipsoid].OneOverF,         sizeof(g_ReferenceEllipsoids[Ellipsoid].OneOverF),         FALSE);

  return TRUE;
}



#define GIS_DATUMINFO_DATUM      0
#define GIS_DATUMINFO_NAME       1
#define GIS_DATUMINFO_ELIPSOID   2
#define GIS_DATUMINFO_DX         3
#define GIS_DATUMINFO_DY         4
#define GIS_DATUMINFO_DZ         5
#define GIS_DATUMINFO_EX         6
#define GIS_DATUMINFO_EY         7
#define GIS_DATUMINFO_EZ         8
#define GIS_DATUMINFO_SATELLITES 9

BEGIN_IMPLEMENT_XP_PARAMETERS(CDatumInfoExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1,  NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(2,  NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(3,  NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                             TRUE, FALSE) 
  XP_OUTPUT_PARAMETER(4,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(5,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(6,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(7,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(8,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(9,  NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(10, NULL, XP_ALLOW_INT4,                                                             TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CDatumInfoExtendedStoredProcedure)

void CDatumInfoExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_DATUMINFO @ReferenceDatumID, @DatumName OUTPUT, @ReferenceEllipsoidID OUTPUT, @DX OUTPUT, @DY OUTPUT, @DZ OUTPUT, @EX OUTPUT, @EY OUTPUT, @EZ OUTPUT, @Satellites OUTPUT");
}

BOOL CDatumInfoExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Validate the Datum value
  DBINT Datum = XP_INT4(m_pParameterData[GIS_DATUMINFO_DATUM].m_Data);
  if (Datum < 0 || Datum >= sizeof(g_ReferenceDatums) / sizeof(CReferenceDatum))
  {
    sprintf(szErrorMsg, "%s: Invalid ReferenceDatumID value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Return the data in the output parameters
  ParamSetOutput(GIS_DATUMINFO_NAME+1,       (BYTE*) g_ReferenceDatums[Datum].Name,        (ULONG) strlen(g_ReferenceDatums[Datum].Name), FALSE);
  ParamSetOutput(GIS_DATUMINFO_ELIPSOID+1,   (BYTE*) &g_ReferenceDatums[Datum].elipsoid,   sizeof(g_ReferenceDatums[Datum].elipsoid),     FALSE);
  ParamSetOutput(GIS_DATUMINFO_DX+1,         (BYTE*) &g_ReferenceDatums[Datum].dx,         sizeof(g_ReferenceDatums[Datum].dx),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_DY+1,         (BYTE*) &g_ReferenceDatums[Datum].dy,         sizeof(g_ReferenceDatums[Datum].dy),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_DZ+1,         (BYTE*) &g_ReferenceDatums[Datum].dz,         sizeof(g_ReferenceDatums[Datum].dz),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_EX+1,         (BYTE*) &g_ReferenceDatums[Datum].ex,         sizeof(g_ReferenceDatums[Datum].ex),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_EY+1,         (BYTE*) &g_ReferenceDatums[Datum].ey,         sizeof(g_ReferenceDatums[Datum].ey),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_EZ+1,         (BYTE*) &g_ReferenceDatums[Datum].ez,         sizeof(g_ReferenceDatums[Datum].ez),           FALSE);
  ParamSetOutput(GIS_DATUMINFO_SATELLITES+1, (BYTE*) &g_ReferenceDatums[Datum].Satellites, sizeof(g_ReferenceDatums[Datum].Satellites),   FALSE);

  return TRUE;
}



#define GIS_LL2UTM_LATITUDE  0
#define GIS_LL2UTM_LONGITUDE 1
#define GIS_LL2UTM_DATUM     2
#define GIS_LL2UTM_NORTHING  3
#define GIS_LL2UTM_EASTING   4
#define GIS_LL2UTM_ZONE      5

BEGIN_IMPLEMENT_XP_PARAMETERS(CLL2UTMExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_FLT8,                                 FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_FLT8,                                 FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(5, NULL, XP_ALLOW_FLT8,                                                             TRUE, FALSE)
  XP_OUTPUT_PARAMETER(6, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CLL2UTMExtendedStoredProcedure)

void CLL2UTMExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_LL2UTM @Latitude, @Longitude, @ReferenceEllipsoidID @Northing OUTPUT, @Easting OUTPUT, @UTMZone OUTPUT");
}

BOOL CLL2UTMExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Pull out the parameters
  double Lat = XP_FLT8(m_pParameterData[GIS_LL2UTM_LATITUDE].m_Data);
  double Long = XP_FLT8(m_pParameterData[GIS_LL2UTM_LONGITUDE].m_Data);
  DBINT ReferenceEllipsoid = XP_INT4(m_pParameterData[GIS_LL2UTM_DATUM].m_Data);
  if (ReferenceEllipsoid < 0 || ReferenceEllipsoid >= sizeof(g_ReferenceEllipsoids) / sizeof(CReferenceEllipsoid))
  {
    //Local buffer we use for reporting messages back to ODS
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Invalid ReferenceEllipsoidID value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Call the helper function which does the work 
  double UTMNorthing;
  double UTMEasting;
  char UTMZone[4];
  LLtoUTM(ReferenceEllipsoid, Lat, Long, UTMNorthing, UTMEasting, UTMZone);

  //Return the data in the 3 output parameters
	ParamSetOutput(GIS_LL2UTM_NORTHING+1, (BYTE*) &UTMNorthing, sizeof(UTMNorthing),     FALSE);
  ParamSetOutput(GIS_LL2UTM_EASTING+1,  (BYTE*) &UTMEasting,  sizeof(UTMEasting),      FALSE);
  ParamSetOutput(GIS_LL2UTM_ZONE+1,     (BYTE*) UTMZone,      (ULONG) strlen(UTMZone), FALSE);

  return TRUE;
}



#define GIS_UTM2LL_NORTHING  0
#define GIS_UTM2LL_EASTING   1
#define GIS_UTM2LL_ZONE      2
#define GIS_UTM2LL_DATUM     3
#define GIS_UTM2LL_LATITUDE  4
#define GIS_UTM2LL_LONGITUDE 5

BEGIN_IMPLEMENT_XP_PARAMETERS(CUTM2LLExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_FLT8,                                                                                                  FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_FLT8,                                                                                                  FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_ASCIITEXT)
  XP_INPUT_PARAMETER (4, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(5, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(6, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CUTM2LLExtendedStoredProcedure)

void CUTM2LLExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_UTM2LL @Northing, @Easting, @UTMZone, @ReferenceEllipsoidID, @Latitude OUTPUT, @Longitude OUTPUT");
}

BOOL CUTM2LLExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Pull out the parameters
  double UTMNorthing = XP_FLT8(m_pParameterData[GIS_UTM2LL_NORTHING].m_Data);
  double UTMEasting = XP_FLT8(m_pParameterData[GIS_UTM2LL_EASTING].m_Data);
  char*  pszUTMZone = XP_ASCIITEXT(m_pParameterData[GIS_UTM2LL_ZONE].m_Data);
  if (strlen(pszUTMZone) < 2)
  {
    sprintf(szErrorMsg, "%s: Invalid UTMZone value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }
  DBINT ReferenceEllipsoid = XP_INT4(m_pParameterData[GIS_UTM2LL_DATUM].m_Data);
  if (ReferenceEllipsoid < 0 || ReferenceEllipsoid >= sizeof(g_ReferenceEllipsoids) / sizeof(CReferenceEllipsoid))
  {
    sprintf(szErrorMsg, "%s: Invalid ReferenceEllipsoidID value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Call the helper function which does the work 
  double Lat;
  double Long;
  UTMtoLL(ReferenceEllipsoid, UTMNorthing, UTMEasting, pszUTMZone, Lat, Long);

  //Return the data in the 2 output parameters
	ParamSetOutput(GIS_UTM2LL_LATITUDE+1,  (BYTE*) &Lat,  sizeof(Lat),  FALSE);
  ParamSetOutput(GIS_UTM2LL_LONGITUDE+1, (BYTE*) &Long, sizeof(Long), FALSE);

  return TRUE;
}



#define GIS_DATUMTRANSFORM_INPUT_LATITUDE   0
#define GIS_DATUMTRANSFORM_INPUT_LONGITUDE  1
#define GIS_DATUMTRANSFORM_INPUT_HEIGHT     2
#define GIS_DATUMTRANSFORM_FROMDATUM        3
#define GIS_DATUMTRANSFORM_TODATUM          4
#define GIS_DATUMTRANSFORM_OUTPUT_LATITUDE  5
#define GIS_DATUMTRANSFORM_OUTPUT_LONGITUDE 6
#define GIS_DATUMTRANSFORM_OUTPUT_HEIGHT    7

BEGIN_IMPLEMENT_XP_PARAMETERS(CDatumTransformExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_FLT8,                                 FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_FLT8,                                 FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_FLT8,                                 FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (4, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_INPUT_PARAMETER (5, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(6, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(7, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
  XP_OUTPUT_PARAMETER(8, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CDatumTransformExtendedStoredProcedure)

void CDatumTransformExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_DATUMTRANSFORM @Latitude, @Longitude, @Height, @FromReferenceDatum, @ToReferenceDatum, @Latitude OUTPUT, @Longitude OUTPUT, @Height OUTPUT");
}

BOOL CDatumTransformExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Validate the Froma and To Datum values
  DBINT FromDatum = XP_INT4(m_pParameterData[GIS_DATUMTRANSFORM_FROMDATUM].m_Data);
  if (FromDatum < 0 || FromDatum >= sizeof(g_ReferenceDatums) / sizeof(CReferenceDatum))
  {
    sprintf(szErrorMsg, "%s: Invalid FromReferenceDatum value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }
  DBINT ToDatum = XP_INT4(m_pParameterData[GIS_DATUMTRANSFORM_TODATUM].m_Data);
  if (ToDatum < 0 || ToDatum >= sizeof(g_ReferenceDatums) / sizeof(CReferenceDatum))
  {
    sprintf(szErrorMsg, "%s: Invalid ToReferenceDatum value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Pull out the other parameters
  CGeodeticPosition FromPosition;
  FromPosition.lat = XP_FLT8(m_pParameterData[GIS_DATUMTRANSFORM_INPUT_LATITUDE].m_Data) * g_deg2rad;
  FromPosition.lon = XP_FLT8(m_pParameterData[GIS_DATUMTRANSFORM_INPUT_LONGITUDE].m_Data) * g_deg2rad;
  FromPosition.h = XP_FLT8(m_pParameterData[GIS_DATUMTRANSFORM_INPUT_HEIGHT].m_Data);

  //First convert from the FromDatum to the Reference Datum of WGS 84 (assuming the FromDatum is not WGS 84 to begin with)
  double from_a;
  double from_f;
  double from_esq;
  CGeodeticPosition ToPosition;
  ToPosition.h = 0;
  ToPosition.lat = 0;
  ToPosition.lon = 0;
  if (FromDatum != WGS_1984_GLOBAL_DEFINITION)
  {
    from_a = g_ReferenceEllipsoids[g_ReferenceDatums[FromDatum].elipsoid].EquatorialRadius;
    from_f = 1 / g_ReferenceEllipsoids[g_ReferenceDatums[FromDatum].elipsoid].OneOverF;
    from_esq = 2*from_f - from_f*from_f;

    double da = g_ReferenceEllipsoids[g_ReferenceDatums[WGS_1984_GLOBAL_DEFINITION].elipsoid].EquatorialRadius - from_a;
    double df = (1.0 / g_ReferenceEllipsoids[g_ReferenceDatums[WGS_1984_GLOBAL_DEFINITION].elipsoid].OneOverF) - from_f;
    double dx = g_ReferenceDatums[FromDatum].dx;
    double dy = g_ReferenceDatums[FromDatum].dy;
    double dz = g_ReferenceDatums[FromDatum].dz;

    FromPosition = DatumTransform(FromPosition, from_a, from_f, from_esq, da, df, dx, dy, dz);
    ToPosition.h = FromPosition.h;
    ToPosition.lat = FromPosition.lat;
    ToPosition.lon = FromPosition.lon;
  }
  
  //Now convert from the Reference Datum of WGS 84 to the ToDatum (Assuming to the ToDatum is not WGS 84 to begin with)
  if (ToDatum != WGS_1984_GLOBAL_DEFINITION)
  {
    from_a = g_ReferenceEllipsoids[g_ReferenceDatums[WGS_1984_GLOBAL_DEFINITION].elipsoid].EquatorialRadius;
    from_f = 1 / g_ReferenceEllipsoids[g_ReferenceDatums[WGS_1984_GLOBAL_DEFINITION].elipsoid].OneOverF;
    from_esq = 2*from_f - from_f*from_f;

    double da = g_ReferenceEllipsoids[g_ReferenceDatums[ToDatum].elipsoid].EquatorialRadius - from_a;
    double df = (1.0 / g_ReferenceEllipsoids[g_ReferenceDatums[ToDatum].elipsoid].OneOverF) - from_f;
    double dx = -g_ReferenceDatums[ToDatum].dx;
    double dy = -g_ReferenceDatums[ToDatum].dy;
    double dz = -g_ReferenceDatums[ToDatum].dz;

    //Finally call the helper function
    ToPosition = DatumTransform(FromPosition, from_a, from_f, from_esq, da, df, dx, dy, dz);
  }

  //Return the data in the 3 output parameters (not forgetting to convert back to degrees)
  ToPosition.lat *= g_rad2deg;
  ToPosition.lon *= g_rad2deg;
	ParamSetOutput(GIS_DATUMTRANSFORM_OUTPUT_LATITUDE+1,  (BYTE*) &ToPosition.lat, sizeof(ToPosition.lat), FALSE);
	ParamSetOutput(GIS_DATUMTRANSFORM_OUTPUT_LONGITUDE+1, (BYTE*) &ToPosition.lon, sizeof(ToPosition.lon), FALSE);
  ParamSetOutput(GIS_DATUMTRANSFORM_OUTPUT_HEIGHT+1,    (BYTE*) &ToPosition.h,   sizeof(ToPosition.h),   FALSE);

  return TRUE;
}



#define GIS_DISTANCE_BETWEEN_LATITUDE1   0
#define GIS_DISTANCE_BETWEEN_LONGITUDE1  1
#define GIS_DISTANCE_BETWEEN_LATITUDE2   2
#define GIS_DISTANCE_BETWEEN_LONGITUDE2  3
#define GIS_DISTANCE_BETWEEN_ELLIPSOID   4
#define GIS_DISTANCE_DISTANCE            5

BEGIN_IMPLEMENT_XP_PARAMETERS(CDistanceBetweenExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (4, NULL, XP_ALLOW_FLT8, FALSE, FALSE, XP_PT_FLT8)
  XP_INPUT_PARAMETER (5, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_OUTPUT_PARAMETER(6, NULL, XP_ALLOW_FLT8, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CDistanceBetweenExtendedStoredProcedure)

void CDistanceBetweenExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_GIS_DISTANCE_BETWEEN @Latitude1, @Longitude1, @Latitude2, @Longitude2, @ReferenceEllipsoidID, @DistanceBetween OUTPUT");
}

BOOL CDistanceBetweenExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Pull out the parameters
  double Lat1  = XP_FLT8(m_pParameterData[GIS_DISTANCE_BETWEEN_LATITUDE1].m_Data) * g_deg2rad;
  double Long1 = XP_FLT8(m_pParameterData[GIS_DISTANCE_BETWEEN_LONGITUDE1].m_Data) * g_deg2rad;
  double Lat2  = XP_FLT8(m_pParameterData[GIS_DISTANCE_BETWEEN_LATITUDE2].m_Data) * g_deg2rad;
  double Long2 = XP_FLT8(m_pParameterData[GIS_DISTANCE_BETWEEN_LONGITUDE2].m_Data) * g_deg2rad;
  DBINT ReferenceEllipsoid = XP_INT4(m_pParameterData[GIS_DISTANCE_BETWEEN_ELLIPSOID].m_Data);
  if (ReferenceEllipsoid < 0 || ReferenceEllipsoid >= sizeof(g_ReferenceEllipsoids) / sizeof(CReferenceEllipsoid))
  {
    sprintf(szErrorMsg, "%s: Invalid ReferenceEllipsoidID value", m_szFunctionName);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get the values required by the helper function
  double a = g_ReferenceEllipsoids[ReferenceEllipsoid].EquatorialRadius;
  double f = 1 / g_ReferenceEllipsoids[ReferenceEllipsoid].OneOverF;

  //And call the helper function
  double Distance = DistanceBetweenPoints(f, a, Lat1, Long1, Lat2, Long2);

  //Return the data in the output parameter
	ParamSetOutput(GIS_DISTANCE_DISTANCE+1, (BYTE*) &Distance, sizeof(Distance), FALSE);

  return TRUE;
}



