/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_HTMLENCODE.h"



/////////////////////////////// Implementation ////////////////////////////////

#define DataIn_PARAMETER_INDEX  0
#define DataOut_PARAMETER_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CXP_HTMLENCODE)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE, XP_PT_ASCIITEXT)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CXP_HTMLENCODE)

void CXP_HTMLENCODE::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_HTMLENCODE @DataIn, @DataOut OUTPUT");
}

BOOL CXP_HTMLENCODE::Run()
{
  //Validate our parameters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter, note that this XP only operates on ASCII data and not UNICODE
  char* szBuffer = XP_ASCIITEXT(m_pParameterData[DataIn_PARAMETER_INDEX].m_Data);

  //replace the special characters with their encoded equivalents
  //namely:
  // &	&amp;
  // <	&lt;
  // >	&gt;
  // '  	&apos;
  // "  	&quot;
  //Note this example is not meant to be a comprehensive HTML encoder and
  //other encodings are of course possible
  char szBufferOut[8000]; //The local variable buffer for the output parameter
  int nMaxOutputLen = sizeof(szBufferOut);
  int nInputLen = strlen(szBuffer);
  int nOutputPos = 0;
  int nInputPos = 0;
  BOOL bTruncated = FALSE;
  while ((nOutputPos < nMaxOutputLen) && (nInputPos < nInputLen) && !bTruncated)
  {
    //Pull out the current character in the input buffer
    char tCurrentChar = szBuffer[nInputPos];
    switch (tCurrentChar)
    {
      case '>':
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+4) > nMaxOutputLen)
		      bTruncated = TRUE; //Cause a break out of the loop
        else
        {
		      szBufferOut[nOutputPos++]='&';
		      szBufferOut[nOutputPos++]='g';
		      szBufferOut[nOutputPos++]='t';
		      szBufferOut[nOutputPos++]=';';
        }
        break;
      }
      case '<':
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+4) > nMaxOutputLen)
          bTruncated = TRUE; //Cause a break out of the loop
		    else
        {
		      szBufferOut[nOutputPos++]='&';
		      szBufferOut[nOutputPos++]='l';
		      szBufferOut[nOutputPos++]='t';
		      szBufferOut[nOutputPos++]=';';
        }
        break;
      }
      case '&':
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+4) > nMaxOutputLen)
		      bTruncated = TRUE; //Cause a break out of the loop
        else
        {
		      szBufferOut[nOutputPos++]='&';
		      szBufferOut[nOutputPos++]='a';
		      szBufferOut[nOutputPos++]='m';
		      szBufferOut[nOutputPos++]='p';
 		      szBufferOut[nOutputPos++]=';';
        }
        break;
      }
      case '\'':
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+5) > nMaxOutputLen)
		      bTruncated = TRUE; //Cause a break out of the loop
        else
        {
		      szBufferOut[nOutputPos++]='&';
		      szBufferOut[nOutputPos++]='a';
		      szBufferOut[nOutputPos++]='p';
		      szBufferOut[nOutputPos++]='o';
 		      szBufferOut[nOutputPos++]='s';
 		      szBufferOut[nOutputPos++]=';';
        }
        break;
      }
      case '"':
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+5) > nMaxOutputLen)
		      bTruncated = TRUE; //Cause a break out of the loop
        else
        {
		      szBufferOut[nOutputPos++]='&';
		      szBufferOut[nOutputPos++]='q';
		      szBufferOut[nOutputPos++]='u';
		      szBufferOut[nOutputPos++]='o';
 		      szBufferOut[nOutputPos++]='t';
 		      szBufferOut[nOutputPos++]=';';
        }
        break;
      }
      default:
      {
        //Check to see if we have run out of output buffer
		    if ((nOutputPos+1) > nMaxOutputLen)
          bTruncated = TRUE; //Cause a break out of the loop
        else
  		    szBufferOut[nOutputPos++] = tCurrentChar; //Simply transfer the input character to the output array
        break;
      }
	  }
    nInputPos++; //Move onto the next input parameter
  }

  //Return the data in the output parameter
  ParamSetOutput(DataOut_PARAMETER_INDEX + 1, (unsigned char *)szBufferOut, nOutputPos, FALSE);

  //We got here, so the running of the XP was successful
  return TRUE;
}
