/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include <atlenc.h>
#include "XP_HTMLENCODE2Proc.h"



/////////////////////////////// Implementation ////////////////////////////////

#define HTMLENCODE2_INPUT_PARAMETER_INDEX  0
#define HTMLENCODE2_OUTPUT_PARAMETER_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CHTMLENCODE2)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, TRUE, FALSE, XP_PT_UNICODETEXT)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CHTMLENCODE2)

void CHTMLENCODE2::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_HTMLENCODE2 @DataIn, @DataOut OUTPUT");
}

BOOL CHTMLENCODE2::Run()
{
  USES_CONVERSION;

  //Validate our parameters  
  XP_ASSERT(m_pParameterData);

  //Get the input parameter
  wchar_t* pwszInputBuffer = XP_UNICODETEXT(m_pParameterData[HTMLENCODE2_INPUT_PARAMETER_INDEX].m_Data);
  int nInputLen = (int) wcslen(pwszInputBuffer);
  int nOutputSize = (nInputLen * 6) + 1;
  wchar_t* pwszOutputBuffer = new wchar_t[nOutputSize]; //Allocate an output buffer big enough to contain the largest possible encoded output buffer
  if (pwszOutputBuffer == NULL)
  {
    //Local buffer we use for reporting messages back to ODS
    char szErrorMsg[1024];

    sprintf(szErrorMsg, "%s: Memory allocation Error", m_szFunctionName);
    SendErrorMsg(szErrorMsg);
  
    //Return the error to indicate failure
    return FALSE;
  }

  //Get the ATL function which does the work for us
  int nEscapedLen = ATL::EscapeXML(pwszInputBuffer, nInputLen, pwszOutputBuffer, nOutputSize, ATL_ESC_FLAG_ATTR);
  pwszOutputBuffer[nEscapedLen] = L'\0';
  
  //Return the output parameter in the appropiate format
  if ((m_pParameterData[HTMLENCODE2_OUTPUT_PARAMETER_INDEX].m_bType == SRVNVARCHAR) || (m_pParameterData[HTMLENCODE2_OUTPUT_PARAMETER_INDEX].m_bType == SRVNTEXT))
  {
    //Return the data in the output parameter
    ParamSetOutput(HTMLENCODE2_OUTPUT_PARAMETER_INDEX + 1, (unsigned char *)pwszOutputBuffer, nEscapedLen*sizeof(wchar_t), FALSE);
  }
  else
  {
    char* pszOutput = W2A(pwszOutputBuffer);
    int nOutputLength = (int) strlen(pszOutput);

    //Return the data in the output parameter
    ParamSetOutput(HTMLENCODE2_OUTPUT_PARAMETER_INDEX + 1, (unsigned char *)pszOutput, nOutputLength, FALSE);
  }

  //Tidy up the heap memory before we exit
  delete [] pwszOutputBuffer;

  //We got here, so the running of the XP was successful
  return TRUE;
}
