/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "xp_iniprocs.h"



/////////////////////////////// Implementation ////////////////////////////////

#define WRITESTRING_INIFILE_INDEX 0
#define WRITESTRING_SECTION_INDEX 1
#define WRITESTRING_KEY_INDEX     2
#define WRITESTRING_VALUE_INDEX   3

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniWriteStringExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(4, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniWriteStringExtendedStoredProcedure)

BOOL CIniWriteStringExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[WRITESTRING_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[WRITESTRING_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[WRITESTRING_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);
  TCHAR* pszValue = XP_TTEXT(m_pParameterData[WRITESTRING_VALUE_INDEX].m_Data);
  XP_ASSERT(pszValue != NULL);

  BOOL bSuccess = WritePrivateProfileString(pszSection, pszKey, pszValue, pszIniFile);
  if (!bSuccess)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: Error calling WritePrivateProfileString, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CIniWriteStringExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_WRITESTRING @IniFile, @Section, @Key, @Value");
}



#define WRITEINTEGER_INIFILE_INDEX 0
#define WRITEINTEGER_SECTION_INDEX 1
#define WRITEINTEGER_KEY_INDEX     2
#define WRITEINTEGER_VALUE_INDEX   3

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniWriteIntegerExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(4, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE, XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CIniWriteIntegerExtendedStoredProcedure)

BOOL CIniWriteIntegerExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[WRITEINTEGER_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[WRITEINTEGER_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[WRITEINTEGER_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);
  DBINT lVal = XP_INT4(m_pParameterData[WRITEINTEGER_VALUE_INDEX].m_Data);

  //For some reason there is no WritePrivateProfileInt, so we just 
  //convert to a string
  TCHAR pszValue[32];
  _stprintf(pszValue, _T("%d"), lVal);

  BOOL bSuccess = WritePrivateProfileString(pszSection, pszKey, pszValue, pszIniFile);
  if (!bSuccess)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: Error calling WritePrivateProfileString, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CIniWriteIntegerExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_WRITEINTEGER @IniFile, @Section, @Key, @Value");
}



#define READSTRING_INIFILE_INDEX      0
#define READSTRING_SECTION_INDEX      1
#define READSTRING_KEY_INDEX          2
#define READSTRING_DEFAULTVALUE_INDEX 3
#define READSTRING_VALUE_INDEX        4

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniReadStringExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (4, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_OUTPUT_PARAMETER(5, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CIniReadStringExtendedStoredProcedure)

BOOL CIniReadStringExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[READSTRING_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[READSTRING_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[READSTRING_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);
  TCHAR* pszDefault = XP_TTEXT(m_pParameterData[READSTRING_DEFAULTVALUE_INDEX].m_Data);
  XP_ASSERT(pszDefault != NULL);

  DWORD dwSize = 1024;
  BOOL bSuccess = FALSE;
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    LPTSTR pszValue = NULL;
    try
    {
      pszValue = new TCHAR[dwSize];
    }
#ifdef _AFX
    catch(CMemoryException* pEx)
    {
      pEx->Delete();
    }
#endif
    catch(std::bad_alloc&)
    {
    }

    if (pszValue)
    {
      //Call the SDK function
      DWORD dwRetreived = GetPrivateProfileString(pszSection, pszKey, pszDefault, pszValue, dwSize, pszIniFile);
      if (dwRetreived == (dwSize - 1))
      {
        //Realloc the array by doubling its size ready for the next loop around
        delete [] pszValue;
        dwSize *= 2;
      }
      else if ((dwRetreived == 0) && _tcslen(pszDefault))
      {
        //Tidy up the heap memory we have used
        delete [] pszValue;

        sprintf(szErrorMsg, "%s: Error calling GetPrivateProfileString, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);

        return FALSE;
      }
      else
      {
        bSuccess = TRUE;

        //Return the data in the output parameter
        if (m_pParameterData[READSTRING_VALUE_INDEX].m_bType == SRVNTEXT || m_pParameterData[READSTRING_VALUE_INDEX].m_bType == SRVNVARCHAR)
        {
          wchar_t* pszwValue = T2W(pszValue);
	        ParamSetOutput(READSTRING_VALUE_INDEX+1, (BYTE*) pszwValue, (ULONG) ((wcslen(pszwValue)+1)*sizeof(wchar_t)), FALSE);
        }
        else
        {
          char* pszAsciiValue = T2A(pszValue);
	        ParamSetOutput(READSTRING_VALUE_INDEX+1, (BYTE*) pszAsciiValue, (ULONG) ((strlen(pszAsciiValue)+1)*sizeof(char)), FALSE);
        }

        //Tidy up the heap memory we have used
        delete [] pszValue;
      }
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to allocate memory to store the retrieved value in, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }
		
  return TRUE;
}

void CIniReadStringExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_READSTRING @IniFile, @Section, @Key, @DefaultValue, @Value OUTPUT");
}



#define READINTEGER_INIFILE_INDEX      0
#define READINTEGER_SECTION_INDEX      1
#define READINTEGER_KEY_INDEX          2
#define READINTEGER_DEFAULTVALUE_INDEX 3
#define READINTEGER_VALUE_INDEX        4

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniReadIntegerExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (4, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE, XP_PT_INT4)
  XP_OUTPUT_PARAMETER(5, NULL, XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CIniReadIntegerExtendedStoredProcedure)

BOOL CIniReadIntegerExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[READINTEGER_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[READINTEGER_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[READINTEGER_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);
  DBINT lDefault = XP_INT4(m_pParameterData[READINTEGER_DEFAULTVALUE_INDEX].m_Data);

  //Call the SDK function
  UINT nValue = GetPrivateProfileInt(pszSection, pszKey, lDefault, pszIniFile);

  //Return the output parameter
	return (ParamSetOutput(READINTEGER_VALUE_INDEX+1, (BYTE*) &nValue, sizeof(nValue), FALSE) != 0);
}

void CIniReadIntegerExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_READINTEGER @IniFile, @Section, @Key, @DefaultValue, @Value OUTPUT");
}



#define GETSECTIONS_INIFILE_INDEX      0

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniGetSectionsExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniGetSectionsExtendedStoredProcedure)

void CIniGetSectionsExtendedStoredProcedure::ReturnRecordSet(LPTSTR pszString)
{
  //First work out how many strings there are in the multi sz string as well as the maximum length
  int nStrings = 0;
  int nMaxStringLengthInBytes = -1;
  LPTSTR pszCurrentString = pszString;
  while (pszCurrentString[0] != _T('\0'))
  {
    ++nStrings;
    int nCurrentLength = (int) _tcslen(pszCurrentString);
    int nCurrentLengthInBytes = nCurrentLength * sizeof(TCHAR);
    if (nCurrentLengthInBytes > nMaxStringLengthInBytes)
      nMaxStringLengthInBytes = nCurrentLengthInBytes;

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }


  //Describe the 1 and only column
#ifdef _UNICODE
  Describe(1, "Section", SRV_NULLTERM, SRVNVARCHAR, nMaxStringLengthInBytes, SRVNVARCHAR, nMaxStringLengthInBytes, NULL);
#else
  Describe(1, "Section", SRV_NULLTERM, SRVVARCHAR, nMaxStringLengthInBytes, SRVVARCHAR, nMaxStringLengthInBytes, NULL);
#endif

  //return each row in the recordset
  pszCurrentString = pszString;
  for (int i=1; i<=nStrings; i++)
  {
    int nCurrentLength = (int) _tcslen(pszCurrentString);
    int nCurrentLengthInBytes = nCurrentLength * sizeof(TCHAR);

    //Set the row data
    SetColumnLength(1, nCurrentLengthInBytes);
    SetColumnData(1, pszCurrentString);
    
    //Send the completed row back to the client
    SendRow();

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }

  if (nStrings)
    SendDone(SRV_DONE_MORE | SRV_DONE_COUNT, 0, nStrings);
  else
    SendDone(SRV_DONE_MORE, 0, 0);
}

BOOL CIniGetSectionsExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[GETSECTIONS_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);

  DWORD dwSize = 1024;
  BOOL bSuccess = FALSE;
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    LPTSTR pszSections = NULL;
    try
    {
      pszSections = new TCHAR[dwSize];
    }
#ifdef _AFX
    catch(CMemoryException* pEx)
    {
      pEx->Delete();
    }
#endif
    catch(std::bad_alloc&)
    {
    }

    if (pszSections)
    {
      //Call the SDK function
      DWORD dwRetreived = GetPrivateProfileSectionNames(pszSections, dwSize, pszIniFile);
      if (dwRetreived == (dwSize - 2))
      {
        //Realloc the array by doubling its size ready for the next loop around
        delete [] pszSections;
        dwSize *= 2;
      }
      else if (dwRetreived == 0)
      {
        DWORD dwError = GetLastError();

        //Tidy up the heap memory we have used
        delete [] pszSections;

        //Only consider it an error if no data was retreived and GetLastError flagged an error
        if (dwError)
        {
          sprintf(szErrorMsg, "%s: Error calling GetPrivateProfileSectionNames, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);

          return FALSE;
        }
        else
        {
          bSuccess = TRUE;
          ReturnRecordSet(_T("\0\0"));
        }
      }
      else
      {
        bSuccess = TRUE;
        ReturnRecordSet(pszSections);

        delete [] pszSections;
      }
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to allocate memory to store the retrieved sections in, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }

  return TRUE;
}

void CIniGetSectionsExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_GETSECTIONS @IniFile");
}



#define READSECTION_INIFILE_INDEX 0
#define READSECTION_SECTION_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniReadSectionExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniReadSectionExtendedStoredProcedure)

void CIniReadSectionExtendedStoredProcedure::ReturnRecordSet(LPTSTR pszString)
{
  //First work out how many strings there are in the multi sz string as well as the maximum length of the keys and values
  int nStrings = 0;
  int nMaxKeyLengthInBytes = -1;
  int nMaxValueLengthInBytes = -1;
  LPTSTR pszCurrentString = pszString;
  while (pszCurrentString[0] != _T('\0'))
  {
    ++nStrings;
    int nCurrentLength = (int) _tcslen(pszCurrentString);

    TCHAR* pszSeparator = _tcschr(pszCurrentString, _T('='));
    int nCurrentKeyLengthInBytes = nCurrentLength * sizeof(TCHAR);
    int nCurrentValueLengthInBytes = 0;
    if (pszSeparator)
    {
      int nKeyLength = (int) (pszSeparator - pszCurrentString);
      nCurrentKeyLengthInBytes = nKeyLength * sizeof(TCHAR);
      nCurrentValueLengthInBytes = (nCurrentLength - nKeyLength - 1) * sizeof(TCHAR);
    }

    if (nCurrentKeyLengthInBytes > nMaxKeyLengthInBytes)
      nMaxKeyLengthInBytes = nCurrentKeyLengthInBytes;
    if (nCurrentValueLengthInBytes > nMaxValueLengthInBytes)
      nMaxValueLengthInBytes = nCurrentValueLengthInBytes;

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }

  //Describe the 2 Columns
#ifdef _UNICODE
  Describe(1, "Key", SRV_NULLTERM, SRVNVARCHAR, nMaxKeyLengthInBytes, SRVNVARCHAR, nMaxKeyLengthInBytes, NULL);
  Describe(2, "Value", SRV_NULLTERM, SRVNVARCHAR, nMaxValueLengthInBytes, SRVNVARCHAR, nMaxValueLengthInBytes, NULL);
#else
  Describe(1, "Key", SRV_NULLTERM, SRVVARCHAR, nMaxKeyLengthInBytes, SRVVARCHAR, nMaxKeyLengthInBytes, NULL);
  Describe(2, "Value", SRV_NULLTERM, SRVVARCHAR, nMaxValueLengthInBytes, SRVVARCHAR, nMaxValueLengthInBytes, NULL);
#endif

  //return each row in the recordset
  pszCurrentString = pszString;
  for (int i=1; i<=nStrings; i++)
  {
    TCHAR* pszSeparator = _tcschr(pszCurrentString, _T('='));
    TCHAR* pszValue = NULL;
    int nCurrentLength = (int) _tcslen(pszCurrentString);
    int nCurrentKeyLengthInBytes = nCurrentLength * sizeof(TCHAR);
    int nCurrentValueLengthInBytes = 0;
    if (pszSeparator)
    {
      int nKeyLength = (int) (pszSeparator - pszCurrentString);
      nCurrentKeyLengthInBytes = nKeyLength * sizeof(TCHAR);
      nCurrentValueLengthInBytes = (nCurrentLength - nKeyLength - 1) * sizeof(TCHAR);
      pszValue = pszSeparator + 1;
    }

    //Set the row data
    SetColumnLength(1, nCurrentKeyLengthInBytes);
    SetColumnData(1, pszCurrentString);

    SetColumnLength(2, nCurrentValueLengthInBytes);
    SetColumnData(2, pszValue);

    //Send the completed row back to the client
    SendRow();

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }

  if (nStrings)
    SendDone(SRV_DONE_MORE | SRV_DONE_COUNT, 0, nStrings);
  else
    SendDone(SRV_DONE_MORE, 0, 0);
}

BOOL CIniReadSectionExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[READSECTION_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[READSECTION_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);

  DWORD dwSize = 1024;
  BOOL bSuccess = FALSE;
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    LPTSTR pszKeyValues = NULL;
    try
    {
      pszKeyValues = new TCHAR[dwSize];
    }
#ifdef _AFX
    catch(CMemoryException* pEx)
    {
      pEx->Delete();
    }
#endif
    catch(std::bad_alloc&)
    {
    }

    if (pszKeyValues)
    {
      //Call the SDK function
      DWORD dwRetreived = GetPrivateProfileSection(pszSection, pszKeyValues, dwSize, pszIniFile);
      if (dwRetreived == (dwSize - 2))
      {
        //Realloc the array by doubling its size ready for the next loop around
        delete [] pszKeyValues;
        dwSize *= 2;
      }
      else if (dwRetreived == 0)
      {
        DWORD dwError = GetLastError();

        //Tidy up the heap memory we have used
        delete [] pszKeyValues;

        //Only consider it an error if no data was retreived and GetLastError flagged an error
        if (dwError)
        {
          sprintf(szErrorMsg, "%s: Error calling GetPrivateProfileSection, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);

          return FALSE;
        }
        else
        {
          bSuccess = TRUE;
          ReturnRecordSet(_T("\0\0"));
        }
      }
      else
      {
        bSuccess = TRUE;
        ReturnRecordSet(pszKeyValues);

        delete [] pszKeyValues;
      }
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to allocate memory to store the retrieved Key Value pairs in, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }
		
  return TRUE;
}

void CIniReadSectionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_READSECTION @IniFile, @Section");
}



#define DELETESECTION_INIFILE_INDEX 0
#define DELETESECTION_SECTION_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniDeleteSectionExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniDeleteSectionExtendedStoredProcedure)

BOOL CIniDeleteSectionExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[DELETESECTION_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[DELETESECTION_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);

  //Call the SDK function to do the work		
  BOOL bSuccess = WritePrivateProfileString(pszSection, NULL, NULL, pszIniFile);
  if (!bSuccess)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: Error calling WritePrivateProfileString, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CIniDeleteSectionExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_DELETESECTION @IniFile, @Section");
}



#define DELETEKEY_INIFILE_INDEX 0
#define DELETEKEY_SECTION_INDEX 1
#define DELETEKEY_KEY_INDEX     2

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniDeleteKeyExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniDeleteKeyExtendedStoredProcedure)

BOOL CIniDeleteKeyExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[DELETEKEY_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[DELETEKEY_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[DELETEKEY_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);

  //Call the SDK function to do the work	
  BOOL bSuccess = WritePrivateProfileString(pszSection, pszKey, NULL, pszIniFile);
  if (!bSuccess)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: Error calling WritePrivateProfileString, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CIniDeleteKeyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_DELETEKEY @IniFile, @Section, @Key");
}



#define GETKEYS_INIFILE_INDEX 0
#define GETKEYS_SECTION_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniGetKeysExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CIniGetKeysExtendedStoredProcedure)

void CIniGetKeysExtendedStoredProcedure::ReturnRecordSet(LPTSTR pszString)
{
  //First work out how many strings there are in the multi sz string as well as the maximum length
  int nStrings = 0;
  int nMaxStringLengthInBytes = -1;
  LPTSTR pszCurrentString = pszString;
  while (pszCurrentString[0] != _T('\0'))
  {
    ++nStrings;
    int nCurrentLength = (int) _tcslen(pszCurrentString);
    int nCurrentLengthInBytes = nCurrentLength * sizeof(TCHAR);
    if (nCurrentLengthInBytes > nMaxStringLengthInBytes)
      nMaxStringLengthInBytes = nCurrentLengthInBytes;

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }


  //Describe the 1 and only column
#ifdef _UNICODE
  Describe(1, "Key", SRV_NULLTERM, SRVNVARCHAR, nMaxStringLengthInBytes, SRVNVARCHAR, nMaxStringLengthInBytes, NULL);
#else
  Describe(1, "Key", SRV_NULLTERM, SRVVARCHAR, nMaxStringLengthInBytes, SRVVARCHAR, nMaxStringLengthInBytes, NULL);
#endif

  //return each row in the recordset
  pszCurrentString = pszString;
  for (int i=1; i<=nStrings; i++)
  {
    int nCurrentLength = (int) _tcslen(pszCurrentString);
    int nCurrentLengthInBytes = nCurrentLength * sizeof(TCHAR);

    //Set the row data
    SetColumnLength(1, nCurrentLengthInBytes);
    SetColumnData(1, pszCurrentString);
    
    //Send the completed row back to the client
    SendRow();

    //Move onto the next string
    pszCurrentString += (nCurrentLength + 1);
  }

  if (nStrings)
    SendDone(SRV_DONE_MORE | SRV_DONE_COUNT, 0, nStrings);
  else
    SendDone(SRV_DONE_MORE, 0, 0);
}

BOOL CIniGetKeysExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[GETKEYS_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[GETKEYS_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);

  DWORD dwSize = 1024;
  BOOL bSuccess = FALSE;
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    LPTSTR pszKeys = NULL; 
    try
    {
      pszKeys = new TCHAR[dwSize];
    }
#ifdef _AFX
    catch(CMemoryException* pEx)
    {
      pEx->Delete();
    }
#endif
    catch(std::bad_alloc&)
    {
    }

    if (pszKeys)
    {
      pszKeys[0] = _T('\0');

      //Call the SDK function
      DWORD dwRetreived = GetPrivateProfileString(pszSection, NULL, _T(""), pszKeys, dwSize, pszIniFile);
      if (dwRetreived == (dwSize - 2))
      {
        //Realloc the array by doubling its size ready for the next loop around
        delete [] pszKeys;
        dwSize *= 2;
      }
      else
      {
        bSuccess = TRUE;
        ReturnRecordSet(pszKeys);

        delete [] pszKeys;
      }
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to allocate memory to store the retrieved Key Value pairs in, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }
	
  return TRUE;
}

void CIniGetKeysExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_GETKEYS @IniFile, @Section");
}



#define WRITE_INIFILE_INDEX 0
#define WRITE_SECTION_INDEX 1
#define WRITE_KEY_INDEX     2
#define WRITE_VALUE_INDEX   3

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniWriteExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(4, NULL, XP_ALLOW_ANY,                                                                                                   FALSE, FALSE, XP_PT_EMPTY)
END_IMPLEMENT_XP_PARAMETERS(CIniWriteExtendedStoredProcedure)

BOOL CIniWriteExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[WRITE_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[WRITE_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[WRITE_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);

  BOOL bSuccess = WritePrivateProfileStruct(pszSection, pszKey, m_pParameterData[WRITE_VALUE_INDEX].m_pData, m_pParameterData[WRITE_VALUE_INDEX].m_cbActualLen, pszIniFile);
  if (!bSuccess)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: Error calling WritePrivateProfileStruct, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CIniWriteExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_WRITE @IniFile, @Section, @Key, @Value");
}



#define READ_INIFILE_INDEX      0
#define READ_SECTION_INDEX      1
#define READ_KEY_INDEX          2
#define READ_VALUE_INDEX        3

BEGIN_IMPLEMENT_XP_PARAMETERS(CIniReadExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_ANY, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CIniReadExtendedStoredProcedure)

int CIniReadExtendedStoredProcedure::HexToInt(TCHAR ch)
{
  //character is in range of 0-9 subtract ascii value of '0' to get integer
  if (ch >= _T('0')  && ch <= _T('9'))
    return (ch - _T('0'));

  //character is in range of a-z or A-Z subtract ascii value of 'a' and add 10 to get it converted into int
  if ((ch >= _T('a') && ch <= _T('f')) || (ch >= _T('A') && ch <= _T('F')))
    return (tolower(ch) - _T('a') + 10);

  //Character is not a Hex Digit
  return -1;
}

BOOL CIniReadExtendedStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get our parameters
  TCHAR* pszIniFile = XP_TTEXT(m_pParameterData[READ_INIFILE_INDEX].m_Data);
  XP_ASSERT(pszIniFile != NULL);
  TCHAR* pszSection = XP_TTEXT(m_pParameterData[READ_SECTION_INDEX].m_Data);
  XP_ASSERT(pszSection != NULL);
  TCHAR* pszKey = XP_TTEXT(m_pParameterData[READ_KEY_INDEX].m_Data);
  XP_ASSERT(pszKey != NULL);

  DWORD dwSize = 1024;
  BOOL bSuccess = FALSE;
  while (!bSuccess)
  {
    //Allocate some heap memory for the SDK call
    TCHAR* pszValue = NULL;
    BYTE* pbyValue = NULL;
    try
    {
      pszValue = new TCHAR[dwSize];
      pbyValue = new BYTE[dwSize];
    }
#ifdef _AFX
    catch(CMemoryException* pEx)
    {
      pEx->Delete();
    }
#endif
    catch(std::bad_alloc&)
    {
    }

    if (pszValue && pbyValue)
    {
      //Call the SDK function, Note we need to use GetPrivateProfileString because GetPrivateProfileStruct
      //needs to know the size of the structure to read beforehand where in our case we do not know this
      DWORD dwRetreived = GetPrivateProfileString(pszSection, pszKey, _T(""), pszValue, dwSize, pszIniFile);
      if (dwRetreived == (dwSize - 1))
      {
        //Realloc the array by doubling its size ready for the next loop around
        delete [] pbyValue;
        delete [] pszValue;
        dwSize *= 2;
      }
      else if (dwRetreived == 0)
      {
        //Tidy up the heap memory we have used
        delete [] pbyValue;
        delete [] pszValue;

        //Return the data in the output parameter (A Null value)
        ParamSetOutput(READ_VALUE_INDEX+1, NULL, 0, TRUE);

        return FALSE;
      }
      else if ((dwRetreived < 4) || (dwRetreived%2 == 1))
      {
        //Tidy up the heap memory we have used
        delete [] pbyValue;
        delete [] pszValue;

        sprintf(szErrorMsg, "%s: Encountered invalid data length in ini value", m_szFunctionName);
        SendErrorMsg(szErrorMsg);

        return FALSE;
      }
      else
      {
        bSuccess = TRUE;

        //Convert from the TCHAR string to the byte data
        DWORD dwBytesToReturn = ((dwRetreived-2)/2);
        for (DWORD i=0; i<dwBytesToReturn; i++)
        {
          int nByteValue1 = HexToInt(pszValue[i*2]);
          if (nByteValue1 == -1)
          {
            //Tidy up the heap memory we have used
            delete [] pbyValue;
            delete [] pszValue;

            sprintf(szErrorMsg, "%s: Encountered invalid data in ini value", m_szFunctionName);
            SendErrorMsg(szErrorMsg);

            return FALSE;
          }
          int nByteValue2 = HexToInt(pszValue[(i*2)+1]);
          if (nByteValue2 == -1)
          {
            //Tidy up the heap memory we have used
            delete [] pbyValue;
            delete [] pszValue;

            sprintf(szErrorMsg, "%s: Encountered invalid data in ini value", m_szFunctionName);
            SendErrorMsg(szErrorMsg);

            return FALSE;
          }

          pbyValue[i] = (BYTE) ((nByteValue1 << 4) + nByteValue2);
        }

        //Return the data in the output parameter
        ParamSetOutput(READ_VALUE_INDEX+1, pbyValue, dwBytesToReturn, FALSE);
 
        //Tidy up the heap memory we have used
        delete [] pbyValue;
        delete [] pszValue;
      }
    }
    else
    {
      //Tidy up the heap memory we have used
      if (pbyValue)
        delete [] pbyValue;
      if (pszValue)
        delete [] pszValue;

      sprintf(szErrorMsg, "%s: Failed to allocate memory to store the retrieved value in, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }
		
  return TRUE;
}

void CIniReadExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_INI_READ @IniFile, @Section, @Key, @DefaultValue, @Value OUTPUT");
}
