/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "xp_ipconfigproc.h"




/////////////////////////////// Implementation ////////////////////////////////

#define VERBOSE_PARAMETER_INDEX 0

BEGIN_IMPLEMENT_XP_PARAMETERS(CIPConfigExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT, FALSE, TRUE, XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CIPConfigExtendedStoredProcedure)

void CIPConfigExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RAND @VerboseOutput");
}

char CIPConfigExtendedStoredProcedure::NibbleToChar(int Nibble)
{
  if (Nibble >= 0 && Nibble < 10)
    return (char) (Nibble + '0');
  else if (Nibble < 16)
    return (char) (Nibble - 10 + 'A');
  else
    return '?';
}


BOOL CIPConfigExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Assume the worst
  BOOL bSuccess = FALSE;

  //Should we do verbose output
  BOOL bVerbose = FALSE;
  if (m_pParameterData[VERBOSE_PARAMETER_INDEX].m_bType != SRVNULL)
    bVerbose = (XP_INT4(m_pParameterData[VERBOSE_PARAMETER_INDEX].m_Data) != 0);

  ULONG nOutputBufferSize = 0;
  DWORD dwError = GetAdaptersInfo(NULL, &nOutputBufferSize);
  if (dwError == ERROR_BUFFER_OVERFLOW)
  {
    PIP_ADAPTER_INFO pAdaptersInfo = (PIP_ADAPTER_INFO) _alloca(nOutputBufferSize);
    dwError = GetAdaptersInfo(pAdaptersInfo, &nOutputBufferSize);
    if (dwError == ERROR_SUCCESS)
    {
      //Get the network params
      nOutputBufferSize = 0;
      dwError = GetNetworkParams(NULL, &nOutputBufferSize);
      if (dwError == ERROR_BUFFER_OVERFLOW)
      {
        PFIXED_INFO pFixedInfo = (PFIXED_INFO) _alloca(nOutputBufferSize);
        dwError = GetNetworkParams(pFixedInfo, &nOutputBufferSize);
        if (dwError == ERROR_SUCCESS)
        {
          //If we got here then consider ourselves successful
          bSuccess = TRUE;

          //Describe the two columns in the recordset we return
          Describe(1, "Attribute", SRV_NULLTERM, SRVVARCHAR, 255, SRVVARCHAR, 255, NULL);
          Describe(2, "Value", SRV_NULLTERM, SRVVARCHAR, 255, SRVVARCHAR, 255, NULL);
          char* pszYes = "Yes";
          char* pszNo = "No";

          if (bVerbose) //Only return the fixed info when we want a verbose listing
          {
            //Return the Host Name
            char* pszHostName = "Host Name";
            SetColumnLength(1, (int) strlen(pszHostName));
            SetColumnData(1, pszHostName);
            SetColumnLength(2, (int) strlen(pFixedInfo->HostName));
            SetColumnData(2, pFixedInfo->HostName);
            SendRow();

            //Return the Primary Dns suffix
            char* pszPrimaryDnsSuffix = "Primary Dns Suffix";
            SetColumnLength(1, (int) strlen(pszPrimaryDnsSuffix));
            SetColumnData(1, pszPrimaryDnsSuffix);
            SetColumnLength(2, (int) strlen(pFixedInfo->DomainName));
            SetColumnData(2, pFixedInfo->DomainName);
            SendRow();

            //Return the node type
            char* pszNodeType = "Node Type";
            char* pszBroadcast = "Broadcast";
            char* pszPeerToPeer = "Peer To Peer";
            char* pszMixed = "Mixed";
            char* pszHybrid = "Hybrid";
            char* pszUnknown = "Unknown";
            SetColumnLength(1, (int) strlen(pszNodeType));
            SetColumnData(1, pszNodeType);
            switch (pFixedInfo->NodeType)
            {
              case BROADCAST_NODETYPE:
              {
                SetColumnLength(2, (int) strlen(pszBroadcast));
                SetColumnData(2, pszBroadcast);
                break;
              }
              case PEER_TO_PEER_NODETYPE:
              {
                SetColumnLength(2, (int) strlen(pszPeerToPeer));
                SetColumnData(2, pszPeerToPeer);
                break;
              }
              case MIXED_NODETYPE:
              {
                SetColumnLength(2, (int) strlen(pszMixed));
                SetColumnData(2, pszMixed);
                break;
              }
              case HYBRID_NODETYPE:
              {
                SetColumnLength(2, (int) strlen(pszHybrid));
                SetColumnData(2, pszHybrid);
                break;
              }
              default:
              {
                SetColumnLength(2, (int) strlen(pszUnknown));
                SetColumnData(2, pszUnknown);
                break;
              }

            }
            SendRow();

            //Return IP Routing Enabled
            char* pszIPRoutingEnabled = "IP Routing Enabled";
            SetColumnLength(1, (int) strlen(pszIPRoutingEnabled));
            SetColumnData(1, pszIPRoutingEnabled);
            if (pFixedInfo->EnableRouting)
            {
              SetColumnLength(2, (int) strlen(pszYes));
              SetColumnData(2, pszYes);
            }
            else
            {
              SetColumnLength(2, (int) strlen(pszNo));
              SetColumnData(2, pszNo);
            }
            SendRow();

            //Return WINS Proxy Enabled
            char* pszWINSProxyEnabled = "WINS Proxy Enabled";
            SetColumnLength(1, (int) strlen(pszWINSProxyEnabled));
            SetColumnData(1, pszWINSProxyEnabled);
            if (pFixedInfo->EnableProxy)
            {
              SetColumnLength(2, (int) strlen(pszYes));
              SetColumnData(2, pszYes);
            }
            else
            {
              SetColumnLength(2, (int) strlen(pszNo));
              SetColumnData(2, pszNo);
            }
            SendRow();
          }

          //Iterate thro all the adapters
          PIP_ADAPTER_INFO pCurrentAdapter = pAdaptersInfo;
          do
          {
            //Use GetPerAdapterInfo to determine some of the attribute values
            nOutputBufferSize = 0;
            dwError = GetPerAdapterInfo(pCurrentAdapter->Index, NULL, &nOutputBufferSize);
            if (dwError == ERROR_BUFFER_OVERFLOW)
            {
              PIP_PER_ADAPTER_INFO pPerAdapterInfo = (PIP_PER_ADAPTER_INFO) _alloca(nOutputBufferSize);
              dwError = GetPerAdapterInfo(pCurrentAdapter->Index, pPerAdapterInfo, &nOutputBufferSize);
              if (dwError == ERROR_SUCCESS)
              {
                //Return the Adapter name
                char* pszAdapterName = "Adapter Name";
                SetColumnLength(1, (int) strlen(pszAdapterName));
                SetColumnData(1, pszAdapterName);
                SetColumnLength(2, (int) strlen(pCurrentAdapter->AdapterName));
                SetColumnData(2, pCurrentAdapter->AdapterName);
                SendRow();
              
                //Note that we do not return the connection-specific DNS suffix as this requires the 
                //GetAdaptersAddresses rather than GetAdaptersInfo IPHelper function which is only 
                //available on XP or 2003. We will leave it as an exercise for the reader, to extend the 
                //code to dynamically handle the lack of the function when run on down level operating systems

                if (bVerbose)
                {
                  //Return the description
                  char* pszDescription = "Description";
                  SetColumnLength(1, (int) strlen(pszDescription));
                  SetColumnData(1, pszDescription);
                  SetColumnLength(2, (int) strlen(pCurrentAdapter->Description));
                  SetColumnData(2, pCurrentAdapter->Description);
                  SendRow();

                  //Return the hardware address (In a readable format)
                  char ReadableAddress[MAX_ADAPTER_ADDRESS_LENGTH*3];
                  memset(ReadableAddress, 0, sizeof(ReadableAddress));
                  int nCurrentOutputIndex = 0;
                  for (UINT i=0; i<pCurrentAdapter->AddressLength; i++)
                  {
                    ReadableAddress[nCurrentOutputIndex++] = NibbleToChar((pCurrentAdapter->Address[i] & 0xF0) >> 4);
                    ReadableAddress[nCurrentOutputIndex++] = NibbleToChar(pCurrentAdapter->Address[i] & 0x0F);
                    if (i<(pCurrentAdapter->AddressLength-1))
                      ReadableAddress[nCurrentOutputIndex++] = '-';
                  }

                  char* pszPhysicalAddress = "Physical Address";
                  SetColumnLength(1, (int) strlen(pszPhysicalAddress));
                  SetColumnData(1, pszPhysicalAddress);
                  SetColumnLength(2, (int) strlen(ReadableAddress));
                  SetColumnData(2, ReadableAddress);
                  SendRow();

                  //Return whether or not DHCP is enabled
                  char* pszDHCPEnabled = "DHCP Enabled";
                  SetColumnLength(1, (int) strlen(pszDHCPEnabled));
                  SetColumnData(1, pszDHCPEnabled);
                  if (pCurrentAdapter->DhcpEnabled)
                  {
                    SetColumnLength(2, (int) strlen(pszYes));
                    SetColumnData(2, pszYes);
                  }
                  else
                  {
                    SetColumnLength(2, (int) strlen(pszNo));
                    SetColumnData(2, pszNo);
                  }
                  SendRow();

                  //Return the Autoconfiguration enabled attribute
                  //Return whether or not DHCP is enabled
                  char* pszAutoconfigurationEnabled = "Autoconfiguration Enabled";
                  SetColumnLength(1, (int) strlen(pszAutoconfigurationEnabled));
                  SetColumnData(1, pszAutoconfigurationEnabled);
                  if (pPerAdapterInfo->AutoconfigEnabled)
                  {
                    SetColumnLength(2, (int) strlen(pszYes));
                    SetColumnData(2, pszYes);
                  }
                  else
                  {
                    SetColumnLength(2, (int) strlen(pszNo));
                    SetColumnData(2, pszNo);
                  }
                  SendRow();
                }


                //Return the IP addresses  
                char* pszIPAddress = "IP Address";
                char* pszSubnetMask = "Subnet Mask";
                IP_ADDR_STRING* pCurrentIPAddress = &pCurrentAdapter->IpAddressList;
                do
                {
                  //First the IP address
                  SetColumnLength(1, (int) strlen(pszIPAddress));
                  SetColumnData(1, pszIPAddress);
                  SetColumnLength(2, (int) strlen(pCurrentIPAddress->IpAddress.String));
                  SetColumnData(2, pCurrentIPAddress->IpAddress.String);
                  SendRow();

                  //Then the subnet mask
                  SetColumnLength(1, (int) strlen(pszSubnetMask));
                  SetColumnData(1, pszSubnetMask);
                  SetColumnLength(2, (int) strlen(pCurrentIPAddress->IpMask.String));
                  SetColumnData(2, pCurrentIPAddress->IpMask.String);
                  SendRow();

                  pCurrentIPAddress = pCurrentIPAddress->Next;
                }
                while (pCurrentIPAddress);

                //return the Default gateway(s)
                char* pszDefaultGateway = "Default Gateway";
                SetColumnLength(1, (int) strlen(pszDefaultGateway));
                SetColumnData(1, pszDefaultGateway);
                IP_ADDR_STRING* pCurrentDefaultGateway = &pCurrentAdapter->GatewayList;
                do
                {
                  SetColumnLength(2, (int) strlen(pCurrentDefaultGateway->IpAddress.String));
                  SetColumnData(2, pCurrentDefaultGateway->IpAddress.String);
                  SendRow();
                  pCurrentDefaultGateway = pCurrentDefaultGateway->Next;
                }
                while (pCurrentDefaultGateway);

                if (bVerbose)
                {
                  //return the DHCP server
                  char* pszDHCPServer = "DHCP Server";
                  SetColumnLength(1, (int) strlen(pszDHCPServer));
                  SetColumnData(1, pszDHCPServer);
                  SetColumnLength(2, (int) strlen(pCurrentAdapter->DhcpServer.IpAddress.String));
                  SetColumnData(2, pCurrentAdapter->DhcpServer.IpAddress.String);
                  SendRow();

                  //return the DNS Server(s)
                  char* pszDNSServer = "DNS Server";
                  SetColumnLength(1, (int) strlen(pszDNSServer));
                  SetColumnData(1, pszDNSServer);
                  IP_ADDR_STRING* pCurrentDNSServer = &pPerAdapterInfo->DnsServerList;
                  do
                  {
                    SetColumnLength(2, (int) strlen(pCurrentDNSServer->IpAddress.String));
                    SetColumnData(2, pCurrentDNSServer->IpAddress.String);
                    SendRow();
                    pCurrentDNSServer = pCurrentDNSServer->Next;
                  }
                  while (pCurrentDNSServer);
      
                  //return the Primary Wins server
                  char* pszPrimaryWinsServer = "Primary Wins Server";
                  SetColumnLength(1, (int) strlen(pszPrimaryWinsServer));
                  SetColumnData(1, pszPrimaryWinsServer);
                  SetColumnLength(2, (int) strlen(pCurrentAdapter->PrimaryWinsServer.IpAddress.String));
                  SetColumnData(2, pCurrentAdapter->PrimaryWinsServer.IpAddress.String);
                  SendRow();

                  //return the Secondary Wins server
                  char* pszSecondaryWinsServer = "Secondary Wins Server";
                  SetColumnLength(1, (int) strlen(pszSecondaryWinsServer));
                  SetColumnData(1, pszSecondaryWinsServer);
                  SetColumnLength(2, (int) strlen(pCurrentAdapter->SecondaryWinsServer.IpAddress.String));
                  SetColumnData(2, pCurrentAdapter->SecondaryWinsServer.IpAddress.String);
                  SendRow();
      
                  if (pCurrentAdapter->DhcpEnabled)
                  {
                    //Return when the Lease was obtained
                    char* pszLeaseObtained = "Lease Obtained";
                    SetColumnLength(1, (int) strlen(pszLeaseObtained));
                    SetColumnData(1, pszLeaseObtained);
                    char* pszTime = ctime(&pCurrentAdapter->LeaseObtained);
                    SetColumnLength(2, (int) strlen(pszTime));
                    SetColumnData(2, pszTime);
                    SendRow();

                    //Retun when the Lease will expire
                    char* pszLeaseExpires = "Lease Expires";
                    SetColumnLength(1, (int) strlen(pszLeaseExpires));
                    SetColumnData(1, pszLeaseExpires);
                    pszTime = ctime(&pCurrentAdapter->LeaseExpires);
                    SetColumnLength(2, (int) strlen(pszTime));
                    SetColumnData(2, pszTime);
                    SendRow();
                  }
                }
              }
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed in call to GetPerAdapterInfo, Error:%d", m_szFunctionName, dwError);
              SendErrorMsg(szErrorMsg);

              return FALSE;
            }

            //Move on to the next item in the linked list
            pCurrentAdapter = pCurrentAdapter->Next;
          }
          while (pCurrentAdapter);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed in call to GetNetworkParams, Error:%d", m_szFunctionName, dwError);
        SendErrorMsg(szErrorMsg);
      }
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed in call to GetAdaptersInfo, Error:%d", m_szFunctionName, dwError);
      SendErrorMsg(szErrorMsg);
    }
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed in call to GetAdaptersInfo, Error:%d", m_szFunctionName, dwError);
    SendErrorMsg(szErrorMsg);
  }
  
  return bSuccess;
}

