/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_NOTIFYProcs.h"



/////////////////////////////// Implementation ////////////////////////////////

BOOL CMessageBeepExtendedStoredProcedure::Run()
{
  BOOL bSuccess = MessageBeep(m_nType);
  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to MessageBeep, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



CSimpleMessageBeepExtendedStoredProcedure::CSimpleMessageBeepExtendedStoredProcedure()
{
  m_nType = (UINT) -1;
}

void CSimpleMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SIMPLE_MESSAGEBEEP");
}



CIconAsteriskMessageBeepExtendedStoredProcedure::CIconAsteriskMessageBeepExtendedStoredProcedure()
{
  m_nType = MB_ICONASTERISK;
}

void CIconAsteriskMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ICONASTERISK_MESSAGEBEEP");
}



CIconExclamationMessageBeepExtendedStoredProcedure::CIconExclamationMessageBeepExtendedStoredProcedure()
{
  m_nType = MB_ICONEXCLAMATION;
}

void CIconExclamationMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ICONEXCLAMATION_MESSAGEBEEP");
}



CIconHandMessageBeepExtendedStoredProcedure::CIconHandMessageBeepExtendedStoredProcedure()
{
  m_nType = MB_ICONHAND;
}

void CIconHandMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ICONHAND_MESSAGEBEEP");
}



CIconQuestionMessageBeepExtendedStoredProcedure::CIconQuestionMessageBeepExtendedStoredProcedure()
{
  m_nType = MB_ICONQUESTION;
}

void CIconQuestionMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_ICONQUESTION_MESSAGEBEEP");
}



COkMessageBeepExtendedStoredProcedure::COkMessageBeepExtendedStoredProcedure()
{
  m_nType = MB_OK;
}

void COkMessageBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_OK_MESSAGEBEEP");
}



#define BEEP_FREQUENCY 0
#define BEEP_DURATION  1

BEGIN_IMPLEMENT_XP_PARAMETERS(CBeepExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_INT4) 
END_IMPLEMENT_XP_PARAMETERS(CBeepExtendedStoredProcedure)

BOOL CBeepExtendedStoredProcedure::Run()
{
  DBINT nFrequency = XP_INT4(m_pParameterData[BEEP_FREQUENCY].m_Data);
  DBINT nDuration = XP_INT4(m_pParameterData[BEEP_DURATION].m_Data);

  BOOL bSuccess = Beep(nFrequency, nDuration);
  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to Beep, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CBeepExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_BEEP Frequency Duration");
}



#define PLAYSOUNDFILE_FILE        0
#define PLAYSOUNDFILE_SYNCHRONOUS 1
#define PLAYSOUNDFILE_LOOP        2

BEGIN_IMPLEMENT_XP_PARAMETERS(CPlaySoundFileExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CPlaySoundFileExtendedStoredProcedure)

BOOL CPlaySoundFileExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszFileName = XP_TTEXT(m_pParameterData[PLAYSOUNDFILE_FILE].m_Data);
  XP_ASSERT(pszFileName);

  //Should we play the sound synchronously, by default we do not
  BOOL bSync = FALSE;
  if (m_pParameterData[PLAYSOUNDFILE_SYNCHRONOUS].m_bType != SRVNULL)
    bSync = (XP_INT4(m_pParameterData[PLAYSOUNDFILE_SYNCHRONOUS].m_Data) != 0);

  //Should we play the sound synchronously, by default we do not
  BOOL bLoop = FALSE;
  if (m_pParameterData[PLAYSOUNDFILE_LOOP].m_bType != SRVNULL)
    bLoop = (XP_INT4(m_pParameterData[PLAYSOUNDFILE_LOOP].m_Data) != 0);

  DWORD dwFlags = SND_FILENAME;
  if (bSync)
    dwFlags |= SND_SYNC;
  else
    dwFlags |= SND_ASYNC;
  if (bLoop)
    dwFlags |= SND_LOOP;

  //Call the SDK function
  BOOL bSuccess = PlaySound(pszFileName, NULL, dwFlags);
  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to PlaySound, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CPlaySoundFileExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_PLAYSOUNDFILE WaveFile [Synchronous] [Loop]");
}



#define PLAYSOUNDALIAS_ALIAS       0
#define PLAYSOUNDALIAS_SYNCHRONOUS 1
#define PLAYSOUNDALIAS_LOOP        2

BEGIN_IMPLEMENT_XP_PARAMETERS(CPlaySoundAliasExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CPlaySoundAliasExtendedStoredProcedure)

BOOL CPlaySoundAliasExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszAlias = XP_TTEXT(m_pParameterData[PLAYSOUNDALIAS_ALIAS].m_Data);
  XP_ASSERT(pszAlias);

  //Should we play the sound synchronously, by default we do not
  BOOL bSync = FALSE;
  if (m_pParameterData[PLAYSOUNDFILE_SYNCHRONOUS].m_bType != SRVNULL)
    bSync = (XP_INT4(m_pParameterData[PLAYSOUNDFILE_SYNCHRONOUS].m_Data) != 0);

  //Should we play the sound synchronously, by default we do not
  BOOL bLoop = FALSE;
  if (m_pParameterData[PLAYSOUNDFILE_LOOP].m_bType != SRVNULL)
    bLoop = (XP_INT4(m_pParameterData[PLAYSOUNDFILE_LOOP].m_Data) != 0);

  DWORD dwFlags = SND_ALIAS;
  if (bSync)
    dwFlags |= SND_SYNC;
  else
    dwFlags |= SND_ASYNC;
  if (bLoop)
    dwFlags |= SND_LOOP;

  //Call the SDK function
  BOOL bSuccess = PlaySound(pszAlias, NULL, dwFlags);
  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to PlaySound, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CPlaySoundAliasExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_PLAYSOUNDALIAS Alias [Synchronous] [Loop]");
}



BOOL CStopSoundExtendedStoredProcedure::Run()
{
  //Call the SDK function
  BOOL bSuccess = PlaySound(NULL, NULL, SND_PURGE);

  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to PlaySound, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CStopSoundExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_STOPSOUND");
}



#define PLAYSOUNDMEMORY_FIELD 0

BEGIN_IMPLEMENT_XP_PARAMETERS(CPlaySoundMemoryExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE, FALSE, FALSE, XP_PT_EMPTY)
END_IMPLEMENT_XP_PARAMETERS(CPlaySoundMemoryExtendedStoredProcedure)

BOOL CPlaySoundMemoryExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameter
  BYTE* pbyWaveFileMemory = m_pParameterData[PLAYSOUNDMEMORY_FIELD].m_pData;
  XP_ASSERT(pbyWaveFileMemory);

  //Call the SDK function
  BOOL bSuccess = PlaySound((LPCTSTR) pbyWaveFileMemory, NULL, SND_MEMORY | SND_SYNC); //Wave file must be played synchronously from memory because allowing asychronous behaviour would cause problems since the 
  if (!bSuccess)                                                                       //memory could get deallocated before the playing of the sound finishes, which would result in an access violation. This
  {                                                                                    //also explains why this XP does not allow looping
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed in call to PlaySound, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

void CPlaySoundMemoryExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_PLAYSOUNDMEMORY WaveFileData");
}
