/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_NTLogProcs.h"
#include "XP_NTLogMsg.h"



/////////////////////////////// Implementation ////////////////////////////////

BOOL CNTLogExtendedStoredProcedure::Report(LPCTSTR pszSourceName, LPCTSTR pszMessage, WORD wType, DWORD dwEventID)
{
  //The buffer we use for reporting any errors
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First we need to register the event source
  HANDLE hEventSource = RegisterEventSource(NULL, pszSourceName);
  if (hEventSource)
  {
    //Report the entry to the event log
    bSuccess = ReportEvent(hEventSource, wType, 0, dwEventID, NULL, 1, 0, &pszMessage, NULL);

    //Report the error if any
    if (!bSuccess)
    {
      sprintf(szErrorMsg, "%s: Error calling ReportEvent, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }
    
    //Deregister the event source
    DeregisterEventSource(hEventSource);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Error calling RegisterEventSource, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



#define LOG_SOURCENAME_INDEX 0
#define LOG_MESSAGE_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CNTLogErrorExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CNTLogErrorExtendedStoredProcedure)

void CNTLogErrorExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_NTLOG_ERROR @SourceName @Message");
}

BOOL CNTLogErrorExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszSourceName = XP_TTEXT(m_pParameterData[LOG_SOURCENAME_INDEX].m_Data);
  XP_ASSERT(pszSourceName);
  const TCHAR* pszMessage = XP_TTEXT(m_pParameterData[LOG_MESSAGE_INDEX].m_Data);
  XP_ASSERT(pszMessage);

  return Report(pszSourceName, pszMessage, EVENTLOG_ERROR_TYPE, NTLOG_FULLSTRING_ERROR);
}



BEGIN_IMPLEMENT_XP_PARAMETERS(CNTLogWarningExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CNTLogWarningExtendedStoredProcedure)

void CNTLogWarningExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_NTLOG_WARNING @SourceName @Message");
}

BOOL CNTLogWarningExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszSourceName = XP_TTEXT(m_pParameterData[LOG_SOURCENAME_INDEX].m_Data);
  XP_ASSERT(pszSourceName);
  const TCHAR* pszMessage = XP_TTEXT(m_pParameterData[LOG_MESSAGE_INDEX].m_Data);
  XP_ASSERT(pszMessage);

  return Report(pszSourceName, pszMessage, EVENTLOG_WARNING_TYPE, NTLOG_FULLSTRING_WARNING);
}




BEGIN_IMPLEMENT_XP_PARAMETERS(CNTLogInformationalExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CNTLogInformationalExtendedStoredProcedure)

void CNTLogInformationalExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_NTLOG_INFORMATIONAL @SourceName @Message");
}

BOOL CNTLogInformationalExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszSourceName = XP_TTEXT(m_pParameterData[LOG_SOURCENAME_INDEX].m_Data);
  XP_ASSERT(pszSourceName);
  const TCHAR* pszMessage = XP_TTEXT(m_pParameterData[LOG_MESSAGE_INDEX].m_Data);
  XP_ASSERT(pszMessage);

  return Report(pszSourceName, pszMessage, EVENTLOG_INFORMATION_TYPE, NTLOG_FULLSTRING_INFORMATIONAL);
}



#define INSTALL_SOURCENAME_INDEX 0

BEGIN_IMPLEMENT_XP_PARAMETERS(CNTLogInstallExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CNTLogInstallExtendedStoredProcedure)

void CNTLogInstallExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_NTLOG_INSTALL @SourceName");
}

BOOL CNTLogInstallExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //The buffer we use for reporting any errors
  char szErrorMsg[2048];

  //Get the parameters
  TCHAR* pszSourceName = XP_TTEXT(m_pParameterData[INSTALL_SOURCENAME_INDEX].m_Data);
  XP_ASSERT(pszSourceName);

  //Get the full path to this dll
  HMODULE hThisDLL = GetModuleHandle(_T("XP_NTLOG"));
  XP_ASSERT(hThisDLL);
  TCHAR szEventMessageFile[_MAX_PATH];
  GetModuleFileName(hThisDLL, szEventMessageFile, _MAX_PATH);

  //Make the necessary updates to the registry
  BOOL bSuccess = FALSE;
  HKEY hAppKey = NULL;
  LONG nSuccess = RegOpenKeyEx(HKEY_LOCAL_MACHINE, _T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"), 0, KEY_WRITE|KEY_READ, &hAppKey);
  if (nSuccess == ERROR_SUCCESS)
  {
		DWORD dwDisposition = 0;
    HKEY hSourceKey = NULL;
    nSuccess = RegCreateKeyEx(hAppKey, pszSourceName, 0, REG_NONE, REG_OPTION_NON_VOLATILE, KEY_WRITE|KEY_READ, NULL, &hSourceKey, &dwDisposition);
		if (nSuccess == ERROR_SUCCESS)
		{
      //Write the Message file string
      nSuccess = RegSetValueEx(hSourceKey, _T("EventMessageFile"), NULL, REG_SZ, (LPBYTE)szEventMessageFile, (lstrlen(szEventMessageFile)+1)*sizeof(TCHAR));
  		bSuccess = (nSuccess == ERROR_SUCCESS);
      if (!bSuccess)
      {
        //Report the error if there is one
        sprintf(szErrorMsg, "%s: Failed to set the EventMessageFile registry value, SourceName:%s, Error:%d", m_szFunctionName, pszSourceName, nSuccess);
        SendErrorMsg(szErrorMsg);
      }

      //Write the Types supported dword
      DWORD dwTypesSupported = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;
      nSuccess = RegSetValueEx(hSourceKey, _T("TypesSupported"), NULL, REG_DWORD,	(LPBYTE)&dwTypesSupported, sizeof(dwTypesSupported));
      if (nSuccess != ERROR_SUCCESS)
      {
        //Report the error if there is one
        sprintf(szErrorMsg, "%s: Failed to set the TypesSupported registry value, SourceName:%s, Error:%d", m_szFunctionName, pszSourceName, nSuccess);
        SendErrorMsg(szErrorMsg);
      }

      bSuccess = bSuccess && (nSuccess == ERROR_SUCCESS);

      //Close the registry key we opened
      RegCloseKey(hSourceKey);
    }
    else
    {
      //Report the error if there is one
      sprintf(szErrorMsg, "%s: Failed to create the EventLog Application source key, SourceName:%s, Error:%d", m_szFunctionName, pszSourceName, nSuccess);
      SendErrorMsg(szErrorMsg);
    }

    //Close the registry key we opened
    RegCloseKey(hAppKey);
  }
  else
  {
    //Report the error if there is one
    sprintf(szErrorMsg, "%s: Failed to open the EventLog Application registry key, SourceName:%s, Error:%d", m_szFunctionName, pszSourceName, nSuccess);
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



#define UNINSTALL_SOURCENAME_INDEX 0

BEGIN_IMPLEMENT_XP_PARAMETERS(CNTLogUninstallExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CNTLogUninstallExtendedStoredProcedure)

void CNTLogUninstallExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_NTLOG_UNINSTALL @SourceName");
}

BOOL CNTLogUninstallExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszSourceName = XP_TTEXT(m_pParameterData[INSTALL_SOURCENAME_INDEX].m_Data);
  XP_ASSERT(pszSourceName);

  //Get the full path to this dll
  HMODULE hThisDLL = GetModuleHandle(_T("XP_NTLOG"));
  XP_ASSERT(hThisDLL);
  TCHAR szEventMessageFile[_MAX_PATH];
  GetModuleFileName(hThisDLL, szEventMessageFile, _MAX_PATH);

  //Remove the settings from the registry
  TCHAR szSubKey[1024];
  _tcscpy(szSubKey, _T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"));
  _tcscat(szSubKey, pszSourceName);
  BOOL bSuccess = (RegDeleteKey(HKEY_LOCAL_MACHINE, szSubKey) == ERROR_SUCCESS);

  //Report the error if there is one
  if (!bSuccess)
  {
    //The buffer we use for reporting any errors
    char szErrorMsg[2048];

    sprintf(szErrorMsg, "%s: Failed to delete SourceName registry entry, SourceName:%s, Error:%d", m_szFunctionName, pszSourceName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}
