/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "xp_randproc.h"




/////////////////////////////// Implementation ////////////////////////////////

#define OUTPUT_PARAMETER_INDEX 0

BEGIN_IMPLEMENT_XP_PARAMETERS(CRandExtendedStoredProcedure)
  XP_OUTPUT_PARAMETER(1, NULL, XP_ALLOW_ANY, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CRandExtendedStoredProcedure)

void CRandExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RAND @OutputValue OUTPUT");
}

BOOL CRandExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Assume the worst
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT | CRYPT_SILENT)) 
  {
    //Call the function to generate the random data
    bSuccess = CryptGenRandom(hProv, m_pParameterData[OUTPUT_PARAMETER_INDEX].m_cbMaxLen, m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData);
    if (!bSuccess)
    {
      sprintf(szErrorMsg, "%s: Failed to generate random data, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }
    else
    {
      //Handles the special cases
      switch (m_pParameterData[OUTPUT_PARAMETER_INDEX].m_bType)
      {
        case SRVNTEXT:
        case SRVNVARCHAR:
        case SRVNCHAR:
        {
          //Generated Unicode data is really just ASCII data promoted to the Unicode code page. This avoids issues
          //such as restricted ranges etc which Unicode has
          int nUnicodeChars = m_pParameterData[OUTPUT_PARAMETER_INDEX].m_cbMaxLen / sizeof(wchar_t);
          for (int i=0; i<nUnicodeChars; i++)
            m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData[i*2] = '\0';
          break;
        }
        case SRVBIT:  //deliberate fallthrough
        case SRVBITN:
        {
          if (m_pParameterData[OUTPUT_PARAMETER_INDEX].m_cbMaxLen)
          {
            if (m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData[0] > 127)
              m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData[0] = 1;
            else
              m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData[0] = 0;
          }
          break;
        }
        case SRVNUMERIC:  //deliberate fallthrough
        case SRVNUMERICN: 
        {
          //We always return a 8 byte integer (bigint) in these cases
          DBNUMERIC* pDBNumeric = (DBNUMERIC*) m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData;
          if (m_pParameterData[OUTPUT_PARAMETER_INDEX].m_cbMaxLen == sizeof(DBNUMERIC))
          {
            pDBNumeric->precision = sizeof(DBNUMERIC);
            memset(&pDBNumeric->val[8], 0, 8); //0 out the remaining 8 bytes of the numeric value
            pDBNumeric->scale = 0;             //Always return a non fractional value

            //Treat the 8 random bytes just created as a signed __int64.
            if (pDBNumeric->val[7] & 0x80)
              pDBNumeric->sign = 0;
            else
              pDBNumeric->sign = 1;
            pDBNumeric->val[7] = (BYTE) (pDBNumeric->val[7] & 0x7F); //Avoid overflow of the integer
          }
          else
          {
            sprintf(szErrorMsg, "%s: Output parameter data is not the expected size", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
            bSuccess = FALSE;
          }
          break;
        }
        case SRVDECIMAL:  //deliberate fallthrough
        case SRVDECIMALN:
        {
          //Awlays return a default precision / default scale value which contains 7 bytes of actual data
          DBDECIMAL* pDBDecimal = (DBDECIMAL*) m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData;
          pDBDecimal->precision = DEFAULTPRECISION;
          pDBDecimal->scale = DEFAULTSCALE;
          pDBDecimal->sign = 1;
          memset(&pDBDecimal->val[7], 0, 9); //0 out the remaining 9 bytes of the numeric value
        }
        default:
        {
          break;
        }
      }

      //Return the data in the output parameter
	    ParamSetOutput(1, m_pParameterData[OUTPUT_PARAMETER_INDEX].m_pData, m_pParameterData[OUTPUT_PARAMETER_INDEX].m_cbMaxLen, FALSE);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}

