/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_REGEXPProc.h"


/////////////////////////////// Implementation ////////////////////////////////

#define REGEXP_ASCII_SOURCETEXT_INDEX    0
#define REGEXP_ASCII_REGEXP_INDEX        1
#define REGEXP_ASCII_CASESENSITIVE_INDEX 2
#define REGEXP_ASCII_MATCH_INDEX         3

BEGIN_IMPLEMENT_XP_PARAMETERS(CASCIIRegExpExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, FALSE, FALSE, XP_PT_ASCIITEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, FALSE, FALSE, XP_PT_ASCIITEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_BIT)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CASCIIRegExpExtendedStoredProcedure)

void CASCIIRegExpExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_REGEXP_ASCII @SourceText, @RegExp, @CaseSensitive, @Match OUTPUT");
}

BOOL CASCIIRegExpExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];
  
  //What will be the return value  
  BOOL bSuccess = FALSE;

  //Get the parameters
  ATL::CAtlRECharTraitsA::RECHARTYPE* pszSourceText = XP_ASCIITEXT(m_pParameterData[REGEXP_ASCII_SOURCETEXT_INDEX].m_Data);
  XP_ASSERT(pszSourceText);
  ATL::CAtlRECharTraitsA::RECHARTYPE* pszRegularExpression = XP_ASCIITEXT(m_pParameterData[REGEXP_ASCII_REGEXP_INDEX].m_Data);
  XP_ASSERT(pszRegularExpression);
  DBBOOL bCaseSensitive = XP_BIT(m_pParameterData[REGEXP_ASCII_CASESENSITIVE_INDEX].m_Data);

  //Create the regular expression parser
  ATL::CAtlRegExp<ATL::CAtlRECharTraitsA> regexp;

  //Parse the regular expression
  ATL::REParseError reParseError = regexp.Parse(pszRegularExpression, bCaseSensitive);
  if (reParseError == ATL::REPARSE_ERROR_OK)
  {  
    //Do the match
    ATL::CAtlREMatchContext<ATL::CAtlRECharTraitsA> mcRegExp;
    BOOL bMatch = regexp.Match(pszSourceText, &mcRegExp);

    //Return the data in the output parameter
    int nConvertedDataSize = Convert(SRVINT4, &bMatch, sizeof(bMatch), m_pParameterData[REGEXP_ASCII_MATCH_INDEX].m_bType, m_pParameterData[REGEXP_ASCII_MATCH_INDEX].m_pData, m_pParameterData[REGEXP_ASCII_MATCH_INDEX].m_cbMaxLen);
    if (nConvertedDataSize > 0)
      bSuccess = (ParamSetOutput(REGEXP_ASCII_MATCH_INDEX+1, m_pParameterData[REGEXP_ASCII_MATCH_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
    else
    {
      sprintf(szErrorMsg, "%s: Failed to convert the match value to the required output type, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed in call to match the regular expression, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;  
}


#define REGEXP_MBCS_SOURCETEXT_INDEX    0
#define REGEXP_MBCS_REGEXP_INDEX        1
#define REGEXP_MBCS_CASESENSITIVE_INDEX 2
#define REGEXP_MBCS_MATCH_INDEX         3

BEGIN_IMPLEMENT_XP_PARAMETERS(CMBCSRegExpExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, FALSE, FALSE, XP_PT_ASCIITEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, FALSE, FALSE, XP_PT_ASCIITEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_BIT)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CMBCSRegExpExtendedStoredProcedure)

void CMBCSRegExpExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_REGEXP_MBCS @SourceText, @RegExp, @CaseSensitive, @Match OUTPUT");
}

BOOL CMBCSRegExpExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];
  
  //What will be the return value  
  BOOL bSuccess = FALSE;

  //Get the parameters
  ATL::CAtlRECharTraitsMB::RECHARTYPE* pszSourceText = (ATL::CAtlRECharTraitsMB::RECHARTYPE*) XP_ASCIITEXT(m_pParameterData[REGEXP_MBCS_SOURCETEXT_INDEX].m_Data);
  XP_ASSERT(pszSourceText);
  ATL::CAtlRECharTraitsMB::RECHARTYPE* pszRegularExpression = (ATL::CAtlRECharTraitsMB::RECHARTYPE*) XP_ASCIITEXT(m_pParameterData[REGEXP_MBCS_REGEXP_INDEX].m_Data);
  XP_ASSERT(pszRegularExpression);
  DBBOOL bCaseSensitive = XP_BIT(m_pParameterData[REGEXP_MBCS_CASESENSITIVE_INDEX].m_Data);

  //Create the regular expression parser
  ATL::CAtlRegExp<ATL::CAtlRECharTraitsMB> regexp;

  //Parse the regular expression
  ATL::REParseError reParseError = regexp.Parse(pszRegularExpression, bCaseSensitive);
  if (reParseError == ATL::REPARSE_ERROR_OK)
  {  
    //Do the match
    ATL::CAtlREMatchContext<ATL::CAtlRECharTraitsMB> mcRegExp;
    BOOL bMatch = regexp.Match(pszSourceText, &mcRegExp);

    //Return the data in the output parameter
    int nConvertedDataSize = Convert(SRVINT4, &bMatch, sizeof(bMatch), m_pParameterData[REGEXP_MBCS_MATCH_INDEX].m_bType, m_pParameterData[REGEXP_MBCS_MATCH_INDEX].m_pData, m_pParameterData[REGEXP_MBCS_MATCH_INDEX].m_cbMaxLen);
    if (nConvertedDataSize > 0)
      bSuccess = (ParamSetOutput(REGEXP_MBCS_MATCH_INDEX+1, m_pParameterData[REGEXP_MBCS_MATCH_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
    else
    {
      sprintf(szErrorMsg, "%s: Failed to convert the match value to the required output type, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed in call to match the regular expression, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;  
}


#define REGEXP_UNICODE_SOURCETEXT_INDEX    0
#define REGEXP_UNICODE_REGEXP_INDEX        1
#define REGEXP_UNICODE_CASESENSITIVE_INDEX 2
#define REGEXP_UNICODE_MATCH_INDEX         3

BEGIN_IMPLEMENT_XP_PARAMETERS(CUnicodeRegExpExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_UNICODETEXT)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_UNICODETEXT)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, FALSE, FALSE, XP_PT_BIT)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CUnicodeRegExpExtendedStoredProcedure)

void CUnicodeRegExpExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_REGEXP_UNICODE @SourceText, @RegExp, @CaseSensitive, @Match OUTPUT");
}

BOOL CUnicodeRegExpExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];
  
  //What will be the return value  
  BOOL bSuccess = FALSE;

  //Get the parameters
  ATL::CAtlRECharTraitsW::RECHARTYPE* pszSourceText = XP_UNICODETEXT(m_pParameterData[REGEXP_UNICODE_SOURCETEXT_INDEX].m_Data);
  XP_ASSERT(pszSourceText);
  ATL::CAtlRECharTraitsW::RECHARTYPE* pszRegularExpression = XP_UNICODETEXT(m_pParameterData[REGEXP_UNICODE_REGEXP_INDEX].m_Data);
  XP_ASSERT(pszRegularExpression);
  DBBOOL bCaseSensitive = XP_BIT(m_pParameterData[REGEXP_UNICODE_CASESENSITIVE_INDEX].m_Data);

  //Create the regular expression parser
  ATL::CAtlRegExp<ATL::CAtlRECharTraitsW> regexp;

  //Parse the regular expression
  ATL::REParseError reParseError = regexp.Parse(pszRegularExpression, bCaseSensitive);
  if (reParseError == ATL::REPARSE_ERROR_OK)
  {  
    //Do the match
    ATL::CAtlREMatchContext<ATL::CAtlRECharTraitsW> mcRegExp;
    BOOL bMatch = regexp.Match(pszSourceText, &mcRegExp);

    //Return the data in the output parameter
    int nConvertedDataSize = Convert(SRVINT4, &bMatch, sizeof(bMatch), m_pParameterData[REGEXP_UNICODE_MATCH_INDEX].m_bType, m_pParameterData[REGEXP_UNICODE_MATCH_INDEX].m_pData, m_pParameterData[REGEXP_UNICODE_MATCH_INDEX].m_cbMaxLen);
    if (nConvertedDataSize > 0)
      bSuccess = (ParamSetOutput(REGEXP_UNICODE_MATCH_INDEX+1, m_pParameterData[REGEXP_UNICODE_MATCH_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
    else
    {
      sprintf(szErrorMsg, "%s: Failed to convert the match value to the required output type, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed in call to match the regular expression, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;  
}





