////////////////////////// Includes /////////////////////////////////////////

#include "stdafx.h"
#include "XP_RUNSCRIPT.h"
#include "XPParameterObj.h"
#include "XP_RUNSCRIPTProcs.h"



//////////////////////////// Macros / Defines ///////////////////////////////

const double ONE3HUNDREDSECONDS_IN_A_DAY = 25920000.0;
const double MINUTES_IN_A_DAY            = 1440.0;



//////////////////////////// Implementation /////////////////////////////////

CXPParameterObj::CXPParameterObj() : m_pESP(NULL), m_nParameterIndex(0), m_bType(SRVNULL), m_nMaxLen(0), m_nActualLen(0), m_bNull(TRUE), m_pData(NULL)
{
}

CXPParameterObj::~CXPParameterObj()
{
  //Tidy up the heap memory if required
  if (m_pData)
    delete [] m_pData;
}

STDMETHODIMP CXPParameterObj::get_Type(XP_DATATYPE* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);

  //Get the info for the specfied parameter before we hand it back (if it has not been got before)
  if (m_bType == SRVNULL)
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }  

  *pVal = (XP_DATATYPE) m_bType;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_MaxLen(long* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);

  //Get the info for the specfied parameter before we hand it back (if it has not been got before)
  if (m_bType == SRVNULL)
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }  

  *pVal = m_nMaxLen;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_ActualLen(long* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);

  //Get the info for the specfied parameter before we hand it back (if it has not been got before)
  if (m_bType == SRVNULL)
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }  

  *pVal = m_nActualLen;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_IsNull(VARIANT_BOOL* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);

  //Get the info for the specfied parameter before we hand it back (if it has not been got before)
  if (m_bType == SRVNULL)
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }  

  *pVal = m_bNull ? VARIANT_TRUE : VARIANT_FALSE;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_IsInput(VARIANT_BOOL* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);
  ATLASSERT(m_pESP);

  *pVal = m_pESP->IsReturnParam(m_nParameterIndex) ? VARIANT_FALSE : VARIANT_TRUE;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_IsOutput(VARIANT_BOOL* pVal)
{
  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);
  ATLASSERT(m_pESP);

  *pVal = m_pESP->IsReturnParam(m_nParameterIndex) ? VARIANT_TRUE : VARIANT_FALSE;
  return S_OK;
}

STDMETHODIMP CXPParameterObj::get_Data(VARIANT* pVal)
{
  USES_CONVERSION;

  //Validate our parameters
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameter, E_POINTER);

  //Get the info for the specfied parameter before we hand it back (if it has not been got before)
  if (m_bType == SRVNULL)
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }  

  //Allocate some heap memory for the parameters data
  if (m_pData == NULL)
    m_pData = new BYTE[max(m_nActualLen, m_nMaxLen) + 2]; //The +2 allows us to NULL terminate the data if needed

  //Handle the error
  if (m_pData == NULL)
    return Error(IDS_FAIL_MEMORY_ALLOCATION, IID_IXPParameter, E_OUTOFMEMORY);

  //Get the actual parameter data if it is not currently NULL
  if (!m_bNull) 
  {
    if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, m_pData, &m_bNull) != SUCCEED)
      return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
  }

  //Initialize the variant to a sane default
  pVal->vt = VT_EMPTY;

  if (!m_bNull)
  {
    switch (m_bType)
    {
      case SRVBIT:  //deliberate fallthrough
      case SRVBITN:
      {
        DBBOOL BoolVal = *((DBBOOL*) m_pData);
        pVal->boolVal = BoolVal ? VARIANT_TRUE : VARIANT_FALSE;
        pVal->vt = VT_BOOL;
        break;
      }
      case SRVBIGCHAR:    //deliberate fallthrough
      case SRVBIGVARCHAR: //deliberate fallthrough
      case SRVVARCHAR:    //deliberate fallthrough
      case SRVCHAR:       //deliberate fallthrough
      case SRVTEXT:
      {
        DBCHAR* pText = (DBCHAR*) m_pData;
        pText[m_nActualLen] = '\0';
        pVal->bstrVal = SysAllocString(A2W(pText));
        pVal->vt = VT_BSTR;
        break;
      }
      case SRVNTEXT: //Deliverate fallthrough
      case SRVNVARCHAR:
      case SRVNCHAR:
      {
        wchar_t* wszText = (wchar_t*) m_pData;
        wszText[m_nActualLen/sizeof(wchar_t)] = '\0';
        pVal->bstrVal = SysAllocString(wszText);
        pVal->vt = VT_BSTR;
        break;
      }
      case SRVINTN:
      {
        switch (m_nMaxLen)
        {
          case sizeof(DBTINYINT):
          {
            pVal->bVal = *((unsigned char*) m_pData);
            pVal->vt = VT_UI1;
            break;
          }
          case sizeof(DBSMALLINT):
          {
            pVal->iVal = *((short*) m_pData);
            pVal->vt = VT_I2;
            break;
          }
          case sizeof(DBINT):
          {
            pVal->lVal = *((long*) m_pData);
            pVal->vt = VT_I4;
            break;
          }
          default:
          {
            break;
          }
        }
        break;
      }
      case SRVINT1:
      {
        pVal->bVal = *((unsigned char*) m_pData);
        pVal->vt = VT_UI1;
        break;
      }
      case SRVINT2:
      {
        pVal->iVal = *((short*) m_pData);
        pVal->vt = VT_I2;
        break;
      }
      case SRVMONEY4:   //deliberate fallthrough
      case SRVINT4:
      {
        pVal->lVal = *((long*) m_pData);
        pVal->vt = VT_I4;
        break;
      }
      case SRVFLT4:
      {
        pVal->fltVal = *((float*) m_pData);
        pVal->vt = VT_R4;
        break;
      }
      case SRVFLT8:
      {
        pVal->dblVal = *((double*) m_pData);
        pVal->vt = VT_R8;
        break;
      }
      case SRVFLTN:
      {
        switch (m_nMaxLen)
        {
          case sizeof(DBFLT4):
          {
            pVal->fltVal = *((float*) m_pData);
            pVal->vt = VT_R4;
            break;
          }
          case sizeof(DBFLT8):
          {
            pVal->dblVal = *((double*) m_pData);
            pVal->vt = VT_R8;
            break;
          }
          default:
          {
            break;
          }
        }
        break;
      }
      case SRVGUID:
      {
        //Need to convert to a string as a VARIANT cannot directly reference a GUID
        wchar_t wszGuid[128];
        if (m_pESP->Convert(m_bType, m_pData, m_nActualLen, SRVNVARCHAR, wszGuid, sizeof(wszGuid)) == -1)
          return Error(IDS_FAIL_CALL_CONVERT, IID_IXPParameter, E_FAIL);
        pVal->bstrVal = SysAllocString(wszGuid);
        pVal->vt = VT_BSTR;
        break;
      }
      case SRVMONEY:
      {
        DBMONEY* pMoney = (DBMONEY*) m_pData;
        pVal->cyVal.Hi = pMoney->mnyhigh;
        pVal->cyVal.Lo = pMoney->mnylow; 
        pVal->vt = VT_CY;
        break;
      }
      case SRVDATETIME:
      {
        DBDATETIME* pDateTime = (DBDATETIME*) m_pData;
        pVal->date = pDateTime->dtdays + (pDateTime->dttime / (ONE3HUNDREDSECONDS_IN_A_DAY)) + 2; //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
        pVal->vt = VT_DATE;
        break;
      }
      case SRVDATETIM4:
      {
        DBDATETIM4* pDateTime = (DBDATETIM4*) m_pData;
        pVal->date = pDateTime->numdays + (pDateTime->nummins / (MINUTES_IN_A_DAY)) + 2; //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
        pVal->vt = VT_DATE;
        break;
      }
      case SRVDATETIMN:
      {
        switch (m_nMaxLen)
        {
          case sizeof(DBDATETIM4):
          {
            DBDATETIM4* pDateTime = (DBDATETIM4*) m_pData;
            pVal->date = pDateTime->numdays + (pDateTime->nummins / (MINUTES_IN_A_DAY)) + 2; //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
            pVal->vt = VT_DATE;
            break;
          }
          case sizeof(DBDATETIME):
          {
            DBDATETIME* pDateTime = (DBDATETIME*) m_pData;
            pVal->date = pDateTime->dtdays + (pDateTime->dttime / (ONE3HUNDREDSECONDS_IN_A_DAY)) + 2; //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
            pVal->vt = VT_DATE;
            break;
          }
          default:
          {
            break;
          }
        }
        break;
      }
      case SRVMONEYN:
      {
        switch (m_nMaxLen)
        {
          case sizeof(DBMONEY4):
          {
            pVal->lVal = *((long*) m_pData);
            pVal->vt = VT_I4;
            break;
          }
          case sizeof(DBMONEY):
          {
            DBMONEY* pMoney = (DBMONEY*) m_pData;
            pVal->cyVal.Hi = pMoney->mnyhigh;
            pVal->cyVal.Lo = pMoney->mnylow; 
            pVal->vt = VT_CY;
            break;
          }
          default:
          {
            break;
          }
        }
        break;
      }
      case SRVDECIMAL:  //deliberate fallthrough
      case SRVNUMERIC:  //deliberate fallthrough
      case SRVDECIMALN: //deliberate fallthrough
      case SRVNUMERICN:
      {
        //To get a decimal into a VARIANT we take the easy approach and first convert to a string 
        //via srv_convert and then use the COM function VarDecFromStr
        char szDecimal[1024];
        int nConverted = m_pESP->Convert(SRVNUMERIC, m_pData, m_nActualLen, SRVTEXT, szDecimal, sizeof(szDecimal));
        if (nConverted == -1)
          return Error(IDS_FAIL_CALL_CONVERT, IID_IXPParameter, E_FAIL);
        szDecimal[nConverted] = '\0';
        HRESULT hr = VarDecFromStr(A2W(szDecimal), LOCALE_SYSTEM_DEFAULT, 0, &pVal->decVal);
        if (FAILED(hr))
          return Error(IDS_FAIL_CONVERT_DECIMAL_TO_STRING, IID_IXPParameter, E_FAIL);
        pVal->vt = VT_DECIMAL;
        break;
      }
      case SRVINT8:
      {
        //Need to convert to a string as a VARIANT cannot directly reference a 8 byte int
        char szBigInt[1024];
        int nConverted = m_pESP->Convert(SRVINT8, m_pData, m_nActualLen, SRVNTEXT, szBigInt, sizeof(szBigInt));
        if (nConverted == -1)
          return Error(IDS_FAIL_CALL_CONVERT, IID_IXPParameter, E_FAIL);
        szBigInt[nConverted] = '\0';
        pVal->bstrVal = SysAllocString(A2W(szBigInt));
        pVal->vt = VT_BSTR;
        break;
      }
      case SRVVARBINARY:    //deliberate fallthrough
      case SRVBINARY:       //deliberate fallthrough 
      case SRVIMAGE:        //deliberate fallthrough 
      case SRVBIGVARBINARY: //deliberate fallthrough
      case SRVBIGBINARY:    //deliberate fallthrough
      {
        //Package the data into a safe array
        SAFEARRAYBOUND bound;
        bound.cElements = (unsigned long) m_nActualLen;
        bound.lLbound = 0;
        SAFEARRAY* pSa = SafeArrayCreate(VT_UI1, 1, &bound);
        if (pSa == NULL)
          return Error(IDS_FAIL_MEMORY_ALLOCATION, IID_IXPParameter, E_OUTOFMEMORY);

        //Get a pointer to the elements of the array and
        //transfer the data into it
        char* pData;
        HRESULT hr = SafeArrayAccessData(pSa, (void**) &pData);
        if (FAILED(hr))
        {
          SafeArrayDestroy(pSa);
          return hr;
        }
        CopyMemory(pData, m_pData, m_nActualLen);  
        SafeArrayUnaccessData(pSa);
        pVal->parray = pSa;
        pVal->vt = VT_ARRAY | VT_BSTR;
        break;
      }
      case SRVSSVARIANT:
      {
        //Create the rich error info
        return Error(IDS_UNHANDLED_PARAMETER_TYPE, IID_IXPParameter, MAKE_HRESULT(1, FACILITY_ITF, IDS_UNKNOWN_PARAMETER_TYPE));
        break;
      }
      default:
      {
        //Create the rich error info
        return Error(IDS_UNKNOWN_PARAMETER_TYPE, IID_IXPParameter, MAKE_HRESULT(1, FACILITY_ITF, IDS_UNKNOWN_PARAMETER_TYPE));
        break;
      }
    }
  }

  return S_OK;
}

STDMETHODIMP CXPParameterObj::SetOutputData(VARIANT Data)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  int nRet = SUCCEED;
  if (Data.vt == VT_EMPTY)
  {
    //call the srv_paramsetoutput
    nRet = m_pESP->ParamSetOutput(m_nParameterIndex, NULL, 0, TRUE);
  }
  else
  {
    //Get the info for the specfied parameter before we hand it back (if it has not been got before)
    if (m_bType == SRVNULL)
    {
      if (m_pESP->ParamInfo(m_nParameterIndex, &m_bType, &m_nMaxLen, &m_nActualLen, NULL, &m_bNull) != SUCCEED)
        return Error(IDS_FAIL_CALL_PARAMINFO, IID_IXPParameter, E_FAIL);
    }  

    //Now convert the VARIANT into ODS compatible data
    BYTE* pData;
    ULONG cbLen;
    HRESULT hr = VariantToXPData(Data, m_bType, m_nMaxLen, IID_IXPParameter, pData, cbLen);
    if (FAILED(hr))
      return hr;

    //Finally call srv_paramsetoutput
    nRet = m_pESP->ParamSetOutput(m_nParameterIndex, pData, cbLen, FALSE);

    //Tidy up the heap memory used before we return
    delete [] pData;
  }

  if (nRet != SUCCEED)
    return Error(IDS_FAIL_CALL_PARAMSETOUTPUT, IID_IXPParameter, E_FAIL);

	return S_OK;
}

HRESULT CXPParameterObj::VariantToXPData(const VARIANT& Data, DBINT desttype, ULONG nMaxLen, const IID& iid, BYTE*& pData, ULONG& cbLen)
{
  USES_CONVERSION;

  //Initialize the output parameters to the output parameters srv_convert
  pData = NULL;
  cbLen = 0;

  switch (desttype)
  {
    case SRVINT1:    //deliberate fallthrough
    case SRVBIT:     //deliberate fallthrough
    case SRVBITN:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_I1, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = sizeof(varConverted.bVal);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &varConverted.bVal, cbLen);
      break;
    }
    case SRVBIGCHAR:   //deliberate fallthrough
    case SRVCHAR:      //deliberate fallthrough
    case SRVTEXT:      //deliberate fallthrough  
    case SRVVARCHAR:   //deliberate fallthrough
    case SRVBIGVARCHAR:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_BSTR, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      char* pszAsciiData = W2A(varConverted.bstrVal);

      cbLen = strlen(pszAsciiData);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, pszAsciiData, cbLen);

      break;
    }
    case SRVINT2:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_I2, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = sizeof(varConverted.iVal);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &varConverted.iVal, cbLen);

      break;
    }
    case SRVMONEY4: //deliberate fallthrough
    case SRVINT4:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_I4, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = sizeof(varConverted.lVal);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &varConverted.lVal, cbLen);

      break;
    }    
    case SRVFLT4:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_R4, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = sizeof(varConverted.fltVal);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &varConverted.fltVal, cbLen);

      break;
    }    
    case SRVFLT8:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_R8, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = sizeof(varConverted.dblVal);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &varConverted.dblVal, cbLen);

      break;
    }
    case SRVMONEY:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_CY, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      DBMONEY dbMoney;
      dbMoney.mnyhigh = varConverted.cyVal.Hi;
      dbMoney.mnylow = varConverted.cyVal.Lo;

      cbLen = sizeof(dbMoney);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &dbMoney, cbLen);

      break;
    }
    case SRVDATETIME:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_DATE, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      double dbldays;
      double dblRemainder = modf(varConverted.date, &dbldays);
      DBDATETIME dbDatetime;
      dbDatetime.dtdays = (long) (dbldays - 2);  //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
      dbDatetime.dttime = (unsigned long) (dblRemainder * ONE3HUNDREDSECONDS_IN_A_DAY);

      cbLen = sizeof(dbDatetime);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &dbDatetime, cbLen);

      break;
    }
    case SRVDATETIM4:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_DATE, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      double dbldays;
      double dblRemainder = modf(varConverted.date, &dbldays);
      DBDATETIM4 dbDatetime;
      dbDatetime.numdays = (unsigned short) (dbldays - 2); //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
      dbDatetime.nummins = (unsigned short) (dblRemainder * MINUTES_IN_A_DAY);

      cbLen = sizeof(dbDatetime);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, &dbDatetime, cbLen);

      break;
    }
    case SRVNCHAR:    //deliberate fallthrough
    case SRVNVARCHAR: //deliberate fallthrough
    case SRVNTEXT:
    {
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_BSTR, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = SysStringLen(varConverted.bstrVal) * sizeof(OLECHAR);
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, varConverted.bstrVal, cbLen);

      break;
    }
    case SRVBIGVARBINARY: //deliberate fallthrough
    case SRVBINARY:       //deliberate fallthrough
    case SRVBIGBINARY:    //deliberate fallthrough
    case SRVIMAGE:        //deliberate fallthrough
    case SRVVARBINARY: 
    {  
      CComVariant varConverted;
      HRESULT hr = varConverted.ChangeType(VT_ARRAY | VT_UI1, &Data);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      long lBound;
      hr = SafeArrayGetLBound(varConverted.parray, 1, &lBound);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      long UBound;
      hr = SafeArrayGetUBound(varConverted.parray, 1, &UBound);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      BYTE* psaData;
      hr = SafeArrayAccessData(varConverted.parray, (void**) &psaData);
      if (FAILED(hr))
        return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

      cbLen = UBound - lBound;
      pData = new BYTE[cbLen];
      if (pData == NULL)
        return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
      memcpy(pData, psaData, cbLen);
      SafeArrayUnaccessData(varConverted.parray);

      break;
    }
    case SRVMONEYN:
    {
      switch (nMaxLen)
      {
        case sizeof(DBMONEY4):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_I4, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.lVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.lVal, cbLen);

          break;
        }
        case sizeof(DBMONEY):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_CY, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          DBMONEY dbMoney;
          dbMoney.mnyhigh = varConverted.cyVal.Hi;
          dbMoney.mnylow = varConverted.cyVal.Lo;

          cbLen = sizeof(dbMoney);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &dbMoney, cbLen);

          break;
        }
        default:
        {
          return Error(IDS_INVALID_DATA_TYPE_TO_CONVERT, iid, E_FAIL);
          break;
        }
      }
      break;
    }
    case SRVDATETIMN:
    {
      switch (nMaxLen)
      {
        case sizeof(DBDATETIM4):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_DATE, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          double dbldays;
          double dblRemainder = modf(varConverted.date, &dbldays);
          DBDATETIM4 dbDatetime;
          dbDatetime.numdays = (unsigned short) (dbldays - 2); //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
          dbDatetime.nummins = (unsigned short) (dblRemainder * MINUTES_IN_A_DAY);

          cbLen = sizeof(dbDatetime);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &dbDatetime, cbLen);

          break;
        }
        case sizeof(DBDATETIME):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_DATE, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          double dbldays;
          double dblRemainder = modf(varConverted.date, &dbldays);
          DBDATETIME dbDatetime;
          dbDatetime.dtdays = (long) (dbldays - 2); //DATE has Jan 1 1900 = 2, Jan 2 1900 = 3 etc etc
          dbDatetime.dttime = (unsigned long) (dblRemainder * ONE3HUNDREDSECONDS_IN_A_DAY);

          cbLen = sizeof(dbDatetime);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &dbDatetime, cbLen);

          break;
        }
        default:
        {
          return Error(IDS_INVALID_DATA_TYPE_TO_CONVERT, iid, E_FAIL);
          break;
        }
      }
      break;
    }
    case SRVINTN:
    {
      switch (nMaxLen)
      {
        case sizeof(DBTINYINT):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_I1, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.bVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.bVal, cbLen);

          break;
        }
        case sizeof(DBSMALLINT):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_I2, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.iVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.iVal, cbLen);

          break;
        }
        case sizeof(DBINT):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_I4, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.lVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.lVal, cbLen);

          break;
        }
        default:
        {
          return Error(IDS_INVALID_DATA_TYPE_TO_CONVERT, iid, E_FAIL);
          break;
        }
      }
      break;
    }
    case SRVFLTN:
    {
      switch (nMaxLen)
      {
        case sizeof(DBFLT4):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_R4, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.fltVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.fltVal, cbLen);

          break;
        }
        case sizeof(DBFLT8):
        {
          CComVariant varConverted;
          HRESULT hr = varConverted.ChangeType(VT_R8, &Data);
          if (FAILED(hr))
            return Error(IDS_FAIL_CONVERT_VARIANT_TO_XP, iid, hr);

          cbLen = sizeof(varConverted.dblVal);
          pData = new BYTE[cbLen];
          if (pData == NULL)
            return Error(IDS_FAIL_MEMORY_ALLOCATION, iid, E_OUTOFMEMORY);
          memcpy(pData, &varConverted.dblVal, cbLen);

          break;
        }
        default:
        {
          return Error(IDS_INVALID_DATA_TYPE_TO_CONVERT, iid, E_FAIL);
          break;
        }
      }
      break;
    }

    default:
    {
      return Error(IDS_INVALID_DATA_TYPE_TO_CONVERT, iid, E_FAIL);
      break;
    }
  }

  return S_OK;
}
