/////////////////////////// Includes  /////////////////////////////////////////

#include "stdafx.h"
#include "XP_RUNSCRIPT.h"
#include "XPParametersObj.h"
#include "XPParameterObj.h"
#include "XP_RUNSCRIPTProcs.h"



/////////////////////////// Macros / Defines //////////////////////////////////

typedef CComEnum < IEnumVARIANT, &IID_IEnumVARIANT, VARIANT, _Copy<VARIANT> > CComEnumVariant;



//////////////////////////// Implementation ///////////////////////////////////

CXPParametersObj::CXPParametersObj() : m_pESP(NULL)
{
}

STDMETHODIMP CXPParametersObj::get_Count(long* pVal)
{
  //Validate our parameters
  if (pVal == NULL)
    return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameters, E_POINTER);
  ATLASSERT(m_pESP);

  //Get the parameter count    
  *pVal = m_pESP->m_nParameters;
	return S_OK;
}

STDMETHODIMP CXPParametersObj::Item(long nIndex, IXPParameter** ppParameter)
{
  //Validate our parameters
  if (ppParameter == NULL)
    return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameters, E_POINTER);
  ATLASSERT(m_pESP);

  //Initialize out value to NULL
  *ppParameter = NULL;

  //Get the specified field
  if ((nIndex >= 1) && (nIndex <= m_pESP->m_nParameters))
  {
    ATLASSERT(m_pESP->m_pParametersCOMArray);

    //Query for the IXPParameter interface for this array element
    return m_pESP->m_pParametersCOMArray[nIndex-1].QueryInterface(IID_IXPParameter, (void**) ppParameter);
  }
  else
  {
    //Create the rich error info
    return Error(IDS_PARAMETER_NAME_NOT_FOUND, IID_IXPParameters, MAKE_HRESULT(1, FACILITY_ITF, IDS_INVALID_PARAMETER_INDEX));
  }

	return S_OK;
}

HRESULT CXPParametersObj::FindIndex(BSTR Name, long* pVal)
{
  USES_CONVERSION;

  //Validate our parameters
  if (Name == NULL)
    return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameters, E_POINTER);
  if (Name == NULL)
    return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameters, E_POINTER);
  ATLASSERT(m_pESP);

  //Try to find the specified field
  *pVal = m_pESP->ParamNumber(W2A(Name));

	return S_OK;
}

STDMETHODIMP CXPParametersObj::get__NewEnum(IUnknown** ppVal)
{
  //Validate our parameters
  if (ppVal == NULL)
    return Error(IDS_FAIL_NULL_POINTER, IID_IXPParameters, E_POINTER);
  ATLASSERT(m_pESP);

  //Set the Enumerator to NULL
  *ppVal = NULL;

  CComObject<CComEnumVariant>* pEnumerator = NULL;
  HRESULT hr = CComObject<CComEnumVariant>::CreateInstance(&pEnumerator);
  if (SUCCEEDED(hr))
  {
    //AddRef the object to prevent it auto destructing
    //while we party on it
    pEnumerator->AddRef();

    //Allocate a VARIANT array which will be taken over by the enumerator
    int nParameters = m_pESP->ParamCount();
    VARIANT* pVarData = new VARIANT[nParameters];
    if (pVarData)
    {
      //Copy each item into the VARIANT arry
      for (long i=0; i<nParameters; i++)
      {
        //Initialize each variant
        VariantInit(&pVarData[i]);

        //Fill in each variant's IDispatch pointer
        CComObject<CXPParameterObj>* pParameter = NULL;
        HRESULT hr = CComObject<CXPParameterObj>::CreateInstance(&pParameter);
        if (SUCCEEDED(hr))
        {
          ATLASSERT(pParameter);
          pParameter->m_pESP = m_pESP;
          pParameter->m_nParameterIndex = i+1;

          //Query for the IDispatch interface for this array element
          if (SUCCEEDED(pParameter->QueryInterface(IID_IDispatch, (void**)&pVarData[i].pdispVal)))
            pVarData[i].vt = VT_DISPATCH;
        }
      } 

      //Initialize the enumerator
      hr = pEnumerator->Init(&pVarData[0], &pVarData[nParameters], 0, AtlFlagTakeOwnership);
      if (SUCCEEDED(hr))
        hr = pEnumerator->QueryInterface(IID_IUnknown, (void**) ppVal);
    }
    else
      hr = Error(IDS_FAIL_CREATE_PARAMETER_ENUMERATOR, IID_IXPParameters, E_OUTOFMEMORY);

    //Decrement the COM reference now that we are finished with it
    pEnumerator->Release();
  }

	return hr;
}
