//////////////////////////// Includes /////////////////////////////////////////

#include "stdafx.h"
#include "XP_RUNSCRIPT.h"
#include "XPScriptSiteObj.h"
#include "XPServerObj.h"
#include "XP_RUNSCRIPTProcs.h"



//////////////////////////// Implementation /////////////////////////////////

CXPRunScriptSite::CXPRunScriptSite() : m_pESP(NULL)
{
}

STDMETHODIMP CXPRunScriptSite::GetLCID(LCID* /*plcid*/)
{
  return E_NOTIMPL;
}

STDMETHODIMP CXPRunScriptSite::GetItemInfo(LPCOLESTR pstrName, DWORD dwReturnMask, IUnknown** ppunkItem, ITypeInfo** ppTypeInfo)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (dwReturnMask & SCRIPTINFO_ITYPEINFO)
	{
		if (ppTypeInfo == NULL)
			return E_INVALIDARG;

		*ppTypeInfo = NULL;
	}

	if (dwReturnMask & SCRIPTINFO_IUNKNOWN)
	{
		if (!ppunkItem)
			return E_INVALIDARG;

		*ppunkItem = NULL;
	}

	// Global object
	if (!_wcsicmp(L"SQL", pstrName)) //Is it the "SQL" object we are being asked for
	{
		if (dwReturnMask & SCRIPTINFO_ITYPEINFO)
		{
			HRESULT hr = m_pESP->m_serverObj.GetTypeInfo(0, NULL, ppTypeInfo);
      if (SUCCEEDED(hr))
			  (*ppTypeInfo)->AddRef();
		}
		
		if (dwReturnMask & SCRIPTINFO_IUNKNOWN)
			return m_pESP->m_serverObj.QueryInterface(IID_IUnknown, (void**) ppunkItem);

		return S_OK;
	}

	return E_INVALIDARG;
}

STDMETHODIMP CXPRunScriptSite::GetDocVersionString(BSTR* /*pbstrVersionString*/)
{
  return E_NOTIMPL;
}

STDMETHODIMP CXPRunScriptSite::OnScriptTerminate(const VARIANT* /*pvarResult*/, const EXCEPINFO* /*pexcepinfo*/)
{
  return S_OK;
}

STDMETHODIMP CXPRunScriptSite::OnStateChange(SCRIPTSTATE /*ssScriptState*/)
{
  return S_OK;
}

STDMETHODIMP CXPRunScriptSite::OnScriptError(IActiveScriptError* pase)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);

  //Get the error position
  DWORD dwSourceContext = 0;
  ULONG ulLineNumber = (ULONG) -1;
  LONG nCharPosition = -1;
  pase->GetSourcePosition(&dwSourceContext, &ulLineNumber, &nCharPosition);

  //Get the error information
  EXCEPINFO ei;
  ZeroMemory(&ei, sizeof(ei));
  if (SUCCEEDED(pase->GetExceptionInfo(&ei)))
  {
    //Do we need to call the deferred fillin function
    if (ei.pfnDeferredFillIn)
      ei.pfnDeferredFillIn(&ei);

    //Don't bother reporting the error if the error is E_ABORT
    if (ei.scode != E_ABORT)
    {
      char* pszError = ei.bstrDescription ? W2A(ei.bstrDescription) : "";
      char* pszSource = ei.bstrSource ? W2A(ei.bstrSource) : "";
      if (ei.bstrSource)
        SysFreeString(ei.bstrSource);
      if (ei.bstrDescription)
        SysFreeString(ei.bstrDescription);
      if (ei.bstrHelpFile)
        SysFreeString(ei.bstrHelpFile);

      //Report the script error back to SQL Server
      char szErrorMsg[2048];
      sprintf(szErrorMsg, "%s: An error occuring running the script, Line: %d, Char: %d, Error: %s, Code: %X, Source: %s", m_pESP->m_szFunctionName, ulLineNumber, nCharPosition, pszError, ei.scode, pszSource);
      m_pESP->SendErrorMsg(szErrorMsg);
    }
  }
  else
  {
    //Report the script error back to SQL Server
    char szErrorMsg[2048];
    sprintf(szErrorMsg, "%s, An error occuring running the script, Line: %d, Char: %d, Code: %X", m_pESP->m_szFunctionName, ulLineNumber, nCharPosition, ei.scode);
    m_pESP->SendErrorMsg(szErrorMsg);
  }

  return S_OK;
}

STDMETHODIMP CXPRunScriptSite::OnEnterScript()
{
  return S_OK;
}

STDMETHODIMP CXPRunScriptSite::OnLeaveScript()
{
  return S_OK;
}

STDMETHODIMP CXPRunScriptSite::QueryContinue()
{
  ATLASSERT(m_pESP);

  return m_pESP->m_bTerminateScript ? S_FALSE : S_OK;
}