////////////////////////// Includes /////////////////////////////////////////

#include "stdafx.h"
#include "XP_RUNSCRIPT.h"
#include "XPServerObj.h"
#include "XPParametersObj.h"
#include "XP_RUNSCRIPTProcs.h"



//////////////////////////// Implementation /////////////////////////////////

CXPServerObj::CXPServerObj() : m_pESP(NULL)
{
}

CXPServerObj::~CXPServerObj()
{
  //Tidy up the heap memory we have been using
  int i;
  for (i=0; i<m_DescribeDefaultData.GetSize(); i++)
    delete [] m_DescribeDefaultData[i];

  for (i=0; i<m_DescribeRowData.GetSize(); i++)
    delete [] m_DescribeRowData[i];
}

STDMETHODIMP CXPServerObj::get_Parameters(IXPParameters** ppParameters)
{
  //Validate our parameters
  ATLASSERT(m_pESP);
	if (ppParameters == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  //Initialize out value to NULL
  *ppParameters = NULL;

  return m_pESP->m_parametersObj.QueryInterface(IID_IXPParameters, (void**) ppParameters);
}

STDMETHODIMP CXPServerObj::get_BindToken(BSTR* pVal)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  char szBindToken[255];
  if (m_pESP->GetBindToken(szBindToken) == SUCCEED)
    *pVal = SysAllocString(A2OLE(szBindToken));
  else
    return Error(IDS_FAIL_CALL_GETBINDTOKEN, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::MessageHandler(long ErrorNum, long Severity, long State, BSTR ErrorText)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);
	if (ErrorText == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  char* pszErrorText = W2A(ErrorText);
  if (m_pESP->MessageHandler(ErrorNum, (BYTE) Severity, (BYTE) State, 0, pszErrorText, strlen(pszErrorText), NULL, 0) != SUCCEED)
    return Error(IDS_FAIL_CALL_MESSAGEHANDLER, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::get_PField(XP_PFIELD Field, BSTR* pVal)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);
		
  int nLen;
  DBCHAR* pField = m_pESP->PField(Field, &nLen);
  if (pField)
    *pVal = SysAllocString(A2W(pField));
  return S_OK;
}

STDMETHODIMP CXPServerObj::get_SessionMessageLCID(long* pVal)
{
  //Validate our parameters
  ATLASSERT(m_pESP);
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);
		
  *pVal = m_pESP->SessionMessageLCID();
  return S_OK;
}

STDMETHODIMP CXPServerObj::get_InstanceName(BSTR* pVal)
{
  //Validate our parameters
  ATLASSERT(m_pESP);
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);
		
  WCHAR* wszInstanceName = m_pESP->InstanceName();
  if (wszInstanceName)
    *pVal = SysAllocString(wszInstanceName);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendDone()
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SendDone(0, 0, 0) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDONE, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendDoneWithError()
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SendDone(SRV_DONE_ERROR, 0, 0) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDONE, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendDoneWithCount(long Count)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SendDone(SRV_DONE_COUNT, 0, Count) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDONE, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendDoneWithCountAndError(long Count)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SendDone(SRV_DONE_COUNT | SRV_DONE_ERROR, 0, Count) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDONE, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendRow()
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SendRow() != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDROW, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SetUserDefinedType(long Column, long UserType)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  if (m_pESP->SetUserDefinedType(Column, UserType) != SUCCEED)
    return Error(IDS_FAIL_CALL_SETUSERDEFINEDTYPE, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendInfoMsg(BSTR Message, long MsgNum, long Class, long State, long LineNum)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);
	if (Message == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  if (m_pESP->SendInfoMsg(W2A(Message), SRV_NULLTERM, (DBINT) MsgNum, (DBTINYINT) Class, (DBTINYINT) State, (DBSMALLINT) LineNum) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDMSG, IID_IXPServer, E_FAIL); 
  return S_OK;
}

STDMETHODIMP CXPServerObj::SendErrorMsg(BSTR Message, long MsgNum, long Class, long State, long LineNum)
{
  USES_CONVERSION;

  //Validate our parameters
  ATLASSERT(m_pESP);
	if (Message == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  if (m_pESP->SendErrorMsg(W2A(Message), SRV_NULLTERM, (DBINT) MsgNum, (DBTINYINT) Class, (DBTINYINT) State, (DBSMALLINT) LineNum) != SUCCEED)
    return Error(IDS_FAIL_CALL_SENDMSG, IID_IXPServer, E_FAIL);
  return S_OK;
}

STDMETHODIMP CXPServerObj::Quit(VARIANT_BOOL bSuccess)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  //Set the flags in the stored procedure instance
  m_pESP->m_bTerminateScript = TRUE;
  m_pESP->m_bReturnCode = (bSuccess == VARIANT_TRUE) ? TRUE : FALSE;

  //Return an error code which will cause the script to terminate
  return E_ABORT; 
}

STDMETHODIMP CXPServerObj::put_ReturnValue(VARIANT_BOOL bSuccess)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  //Set the flag in the stored procedure instance
  m_pESP->m_bReturnCode = (bSuccess == VARIANT_TRUE) ? TRUE : FALSE;

  return S_OK; 
}

STDMETHODIMP CXPServerObj::get_ReturnValue(VARIANT_BOOL* pVal)
{
  //Validate our parameters
  ATLASSERT(m_pESP);
	if (pVal == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  //return the value in the output parameter
  *pVal = m_pESP->m_bReturnCode ? VARIANT_TRUE : VARIANT_FALSE;

  return S_OK; 
}

STDMETHODIMP CXPServerObj::Describe(BSTR ColumnName, XP_DATATYPE DestType, long DestLen, VARIANT Data)
{
  USES_CONVERSION;

  //Validate our parameters
	if (ColumnName == NULL)
		return Error(IDS_FAIL_NULL_POINTER, IID_IXPServer, E_POINTER);

  //Convert the incoming VARIANT into an ODS data type
  BYTE* pData;
  ULONG cbLen;
  HRESULT hr = CXPParameterObj::VariantToXPData(Data, DestType, DestLen, IID_IXPServer, pData, cbLen);
  if (FAILED(hr))
    return hr;

  //call the underlying function
  int nColumn = m_pESP->Describe(0, W2A(ColumnName), SRV_NULLTERM, DestType, DestLen, DestType, cbLen, pData);
  if (nColumn == 0)
  {
    //tidy up the heap memory before we return
    delete [] pData;

    return Error(IDS_FAIL_CALL_DESCRIBE, IID_IXPServer, E_FAIL);
  }

  //If successful, add to the cached data so that the data stays around for the lifetime of this XP
  long nDestType = DestType;
  m_DescribeDataTypes.Add(nDestType);
  ATLASSERT(nColumn == m_DescribeDataTypes.GetSize()); 

  long nDestLen = DestLen;
  m_DescribeDataLen.Add(nDestLen);
  ATLASSERT(nColumn == m_DescribeDataLen.GetSize()); 

  m_DescribeDefaultData.Add(pData);
  ATLASSERT(nColumn == m_DescribeDefaultData.GetSize());

  BYTE* pRowData = NULL;
  m_DescribeRowData.Add(pRowData);
  ATLASSERT(nColumn == m_DescribeRowData.GetSize());

  return S_OK;
}

STDMETHODIMP CXPServerObj::SetColumnData(long Column, VARIANT Data)
{
  //Validate our parameters
  ATLASSERT(m_pESP);

  //Column number must be valid
  if (Column < 1 || Column > m_DescribeDataTypes.GetSize())
    return Error(IDS_INVALID_COLUMN_FOR_SETCOLUMNDATA, IID_IXPServer, E_FAIL);

  //Convert the incoming VARIANT into an ODS data type
  BYTE* pData;
  ULONG cbLen;
  HRESULT hr = CXPParameterObj::VariantToXPData(Data, m_DescribeDataTypes[Column-1], m_DescribeDataLen[Column-1], IID_IXPServer, pData, cbLen);
  if (FAILED(hr))
    return hr;

  //Call the underlying functions
  if (m_pESP->SetColumnData(Column, pData) != SUCCEED)
  {
    //tidy up the heap memory before we return
    delete [] pData;

    return Error(IDS_FAIL_CALL_SETCOLUMNDATA, IID_IXPServer, E_FAIL);
  }
  if (m_pESP->SetColumnLength(Column, cbLen) != SUCCEED)
  {
    //tidy up the heap memory before we return
    delete [] pData;

    return Error(IDS_FAIL_CALL_SETCOLUMNLEN, IID_IXPServer, E_FAIL);
  }

  //cache the pointer in the appropiate location in the m_DescribeRowData array
  if (m_DescribeRowData[Column-1] != NULL)
    delete [] m_DescribeRowData[Column-1];
  m_DescribeRowData[Column-1] = pData;

	return S_OK;
}
