/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_RUNSCRIPT.h"
#include "XP_RUNScriptProcs.h"
#include "XPScriptSiteObj.h"
#include "XPServerObj.h"



/////////////////////////////// Implementation ////////////////////////////////

//A simple helper class which intializes COM for us in its constructor and
//automatically uninitializes it in the destructor
class CCOMInitialize
{
public:
//Constructors / Destructors
  CCOMInitialize()
  {
    m_hResult = CoInitialize(NULL);
  }

  ~CCOMInitialize()
  {
    CoUninitialize();
  }

//Member variables
  HRESULT m_hResult; //The result of CoInitialize since a constructor 
                     //by definition cannot have a return value
};


CRunScriptStoredProcedure::CRunScriptStoredProcedure() : m_bTerminateScript(FALSE), m_bReturnCode(TRUE), m_pParametersCOMArray(NULL), m_nParameters(0)
{
  m_parametersObj.m_pESP = this; //pointer back to this object
  m_serverObj.m_pESP     = this; //pointer back to this object
  m_scriptSiteObj.m_pESP = this; //pointer back to this object
}

CRunScriptStoredProcedure::~CRunScriptStoredProcedure()
{
  //Tidy up the heap memory we have been using
  if (m_pParametersCOMArray)
    delete [] m_pParametersCOMArray;
}

BOOL CRunScriptStoredProcedure::RunScript(const CComBSTR& bstrScript, LPCTSTR pszLanguage)
{
  USES_CONVERSION;

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Create the array of COM parameter objects
  m_nParameters = ParamCount();
  m_pParametersCOMArray = new CComObjectStackWithQI<CXPParameterObj>[m_nParameters];
  if (m_pParametersCOMArray == NULL)
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to allocate memory to hold the parameter data, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);

    return FALSE;    
  }
  for (int i=0; i<m_nParameters; i++)
  {
    //Initialize each parameter
    m_pParametersCOMArray[i].m_nParameterIndex = i+1;
    m_pParametersCOMArray[i].m_pESP = this;
  }

  //Get the CLSID from the ProgID of the language to use
  CLSID clsidEngine;
  HRESULT hr = CLSIDFromProgID(T2OLE((LPTSTR) pszLanguage), &clsidEngine);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to find the CLSID for the specified scripting Language, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Create the IActiveScript COM object
  CComPtr<IActiveScript> scriptEngine;
  hr = CoCreateInstance(clsidEngine, NULL, CLSCTX_INPROC_SERVER, IID_IActiveScript, (void**) &scriptEngine);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to create Active Scripting engine, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get ap pointer to the IActiveScriptParse which is what we need
  CComPtr<IActiveScriptParse> scriptParser;
  hr = scriptEngine->QueryInterface(&scriptParser);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to obtain the Active Scripting parser interface, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Inform the engine with our site pointer
  scriptEngine->SetScriptSite(&m_scriptSiteObj);

  //Initialize the script parser
  scriptParser->InitNew();

  //Inject our object into the engine's namespace and set the state to started
  scriptEngine->AddNamedItem(OLESTR("SQL"), SCRIPTITEM_ISVISIBLE|SCRIPTITEM_ISSOURCE);

  //Add our script to the script engine
  EXCEPINFO ei;
  hr = scriptParser->ParseScriptText(bstrScript, NULL, NULL, NULL, 0, 0, 0, NULL, &ei);
  if (hr != S_OK)
  {
    if (hr == DISP_E_EXCEPTION)
    {
      //Do we need to call the deferred fillin function
      if (ei.pfnDeferredFillIn)
        ei.pfnDeferredFillIn(&ei);

      char* pszError = ei.bstrDescription ? W2A(ei.bstrDescription) : "";
      char* pszSource = ei.bstrSource ? W2A(ei.bstrSource) : "";

      //Report the script error back to SQL Server
      char szErrorMsg[2048];
      sprintf(szErrorMsg, "%s: An error occurred parsing the script, Error: %s, Code: %X, Source: %s", m_szFunctionName, pszError, ei.wCode, pszSource);
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
    else
    {
      //Report the error to the client
      sprintf(szErrorMsg, "%s: Failed to parse the script file, Error:%X", m_szFunctionName, hr);
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }

    //Free up the details in the EXCEPINFO structure
    if (ei.bstrSource)
      SysFreeString(ei.bstrSource);
    if (ei.bstrDescription)
      SysFreeString(ei.bstrDescription);
    if (ei.bstrHelpFile)
      SysFreeString(ei.bstrHelpFile);

    return FALSE;
  }

  //Start executing the script
  scriptEngine->SetScriptState(SCRIPTSTATE_STARTED);

  //Now will not return until the script has finished executing, or has been prematurely terminated via SQL.Quit
  
  return m_bReturnCode;
}




#define RUNSCRIPT_DISK_LANGUAGE_INDEX 0
#define RUNSCRIPT_DISK_FILENAME_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CRunScriptDiskStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CRunScriptDiskStoredProcedure)

void CRunScriptDiskStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RUNSCRIPT_DISK @ScriptLanguage @ScriptFilename [... You can include other parameters here which you want to pass to the script to run]");
}

BOOL CRunScriptDiskStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Initialize COM
  CCOMInitialize coInitialize;
  if (FAILED(coInitialize.m_hResult))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to initialize COM, Error:%x", m_szFunctionName, coInitialize.m_hResult);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get the parameters
  TCHAR* pszLanguage = XP_TTEXT(m_pParameterData[RUNSCRIPT_DISK_LANGUAGE_INDEX].m_Data);
  XP_ASSERT(pszLanguage);
  TCHAR* pszScriptFilename = XP_TTEXT(m_pParameterData[RUNSCRIPT_DISK_FILENAME_INDEX].m_Data);
  XP_ASSERT(pszScriptFilename);

  //Load up the text from the script file into a BSTR so that we can call ParseScriptText with it

  //Open the specified file
  HANDLE hFile = CreateFile(pszScriptFilename, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  char* pszText = NULL;
  if (hFile != INVALID_HANDLE_VALUE)
  {
    //Get the size of the file
    DWORD dwSize = GetFileSize(hFile, NULL);
    if (dwSize == 0)
    {
      //Don't forget to close the file before we return
      CloseHandle(hFile);

      //Report the error to the client
      sprintf(szErrorMsg, "%s: Specified file does not contain any data", m_szFunctionName);
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }

    //Allocate some data to contain the data
    pszText = new char[dwSize + 1];
    if (pszText == NULL)
    {
      //Don't forget to close the file before we return
      CloseHandle(hFile);

      //Report the error to the client
      sprintf(szErrorMsg, "%s: Failed to allocate memory to read in the script contents", m_szFunctionName);
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }

    //And read the data from disk
    DWORD dwBytesRead = 0;
    ReadFile(hFile, pszText, dwSize, &dwBytesRead, NULL);
    pszText[dwBytesRead] = '\0';

    //Don't forget to close the file before we return
    CloseHandle(hFile);
  }
  else
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to open the specified script, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Convert the ascii text to a BSTR
  CComBSTR bstrScript(pszText);
  delete [] pszText;

  //Call the helper function which does most of the work
  return RunScript(bstrScript, pszLanguage);
}



#define RUNSCRIPT_PARAMETER_LANGUAGE_INDEX 0
#define RUNSCRIPT_PARAMETER_FILENAME_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CRunScriptParameterStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CRunScriptParameterStoredProcedure)

void CRunScriptParameterStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RUNSCRIPT_PARAMETER @ScriptLanguage @ScriptText [... You can include other parameters here which you want to pass to the script to run]");
}

BOOL CRunScriptParameterStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Initialize COM
  CCOMInitialize coInitialize;
  if (FAILED(coInitialize.m_hResult))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to initialize COM, Error:%x", m_szFunctionName, coInitialize.m_hResult);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get the parameters
  TCHAR* pszLanguage = XP_TTEXT(m_pParameterData[RUNSCRIPT_PARAMETER_LANGUAGE_INDEX].m_Data);
  XP_ASSERT(pszLanguage);
  TCHAR* pszScript = XP_TTEXT(m_pParameterData[RUNSCRIPT_PARAMETER_FILENAME_INDEX].m_Data);
  XP_ASSERT(pszScript);

  //Convert the text to a BSTR
  CComBSTR bstrScript(pszScript);

  //Call the helper function which does most of the work
  return RunScript(bstrScript, pszLanguage);
}








#define RUN_ACTIVEX_PROGID_INDEX 0
#define RUN_ACTIVEX_FUNCTIONNAME_INDEX 1

BEGIN_IMPLEMENT_XP_PARAMETERS(CRunActiveXStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CRunActiveXStoredProcedure)

void CRunActiveXStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RUN_ACTIVEX @ProgID @FunctionName [... You can include other parameters here which you want to pass to the script to run]");
}

BOOL CRunActiveXStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Initialize COM
  CCOMInitialize coInitialize;
  if (FAILED(coInitialize.m_hResult))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to initialize COM, Error:%x", m_szFunctionName, coInitialize.m_hResult);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get the parameters
  TCHAR* pszProgID = XP_TTEXT(m_pParameterData[RUN_ACTIVEX_PROGID_INDEX].m_Data);
  XP_ASSERT(pszProgID);
  TCHAR* pszFunctionName = XP_TTEXT(m_pParameterData[RUN_ACTIVEX_FUNCTIONNAME_INDEX].m_Data);
  XP_ASSERT(pszFunctionName);

  //Create the array of COM parameter objects
  m_nParameters = ParamCount();
  m_pParametersCOMArray = new CComObjectStackWithQI<CXPParameterObj>[m_nParameters];
  if (m_pParametersCOMArray == NULL)
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to allocate memory to hold the parameter data, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);

    return FALSE;    
  }
  for (int i=0; i<m_nParameters; i++)
  {
    //Initialize each parameter
    m_pParametersCOMArray[i].m_nParameterIndex = i+1;
    m_pParametersCOMArray[i].m_pESP = this;
  }

  //Setup the one and only parameter to the function which is the "IXPServer" COM object
  CComVariant varSQL;
  HRESULT hr = m_serverObj.QueryInterface(IID_IDispatch, (void**) &varSQL.pdispVal);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to initialize the SQL server object prior to calling the ActiveX object, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }
  varSQL.vt = VT_DISPATCH;

  //Get the CLSID from the ProgID specified
  CLSID clsidCOM;
  hr = CLSIDFromProgID(T2OLE((LPTSTR) pszProgID), &clsidCOM);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to find the CLSID for the specified ProgID, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Create the specified ActiveX object
  CComDispatchDriver activeXObject;
  hr = CoCreateInstance(clsidCOM, NULL, CLSCTX_INPROC_SERVER, IID_IDispatch, (void**) &activeXObject);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to create the specified ActiveX object, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Get the disp id of the specified function
  DISPID dispid;
  hr = activeXObject.GetIDOfName(T2OLE((LPTSTR) pszFunctionName), &dispid);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to find the specified function in the ActiveX object, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Now call the specified function
  CComVariant varRetVal;
  hr = activeXObject.Invoke1(dispid, &varSQL, &varRetVal);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed during call to the ActiveX object, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Convert the return value to a boolean
  hr = varRetVal.ChangeType(VT_BOOL);
  if (FAILED(hr))
  {
    //Report the error to the client
    sprintf(szErrorMsg, "%s: Failed to convert the return value to a boolean, Error:%X", m_szFunctionName, hr);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //return the success indicator
  return (varRetVal.boolVal == VARIANT_TRUE) ? TRUE : FALSE;
}


