#ifndef __XP_RUNSCRIPTPROCS_H__
#define __XP_RUNSCRIPTPROCS_H__

///////////////////// Includes ////////////////////////////////////////////////

#include "..\xp++\xp_pp.h"
#include "XPParameterObj.h"
#include "XPParametersObj.h"
#include "XPServerObj.h"
#include "XPScriptSiteObj.h"



//////////////////// Classes /////////////////////////////////////////

//A cross between CComObjectStack and CComObject which is stack based but does
//allow QI
template <class Base>
class CComObjectStackWithQI : public Base
{
public:
	typedef Base _BaseClass;
	CComObjectStackWithQI(void* = NULL)
  {
    m_hResFinalConstruct = FinalConstruct();
  }
	~CComObjectStackWithQI()
	{
		FinalRelease();
#ifdef _ATL_DEBUG_INTERFACES
		_Module.DeleteNonAddRefThunk(_GetRawUnknown());
#endif
	}

	STDMETHOD_(ULONG, AddRef)() 
  {
    return 0;
  }
	STDMETHOD_(ULONG, Release)()
  {
    return 0;
  }
	STDMETHOD(QueryInterface)(REFIID iid, void** ppvObject)
	{
    return _InternalQueryInterface(iid, ppvObject);
  }
	template <class Q>
	HRESULT STDMETHODCALLTYPE QueryInterface(Q** pp)
	{
		return QueryInterface(__uuidof(Q), (void**)pp);
	}
	HRESULT m_hResFinalConstruct;
};


///////////////// forward reference ////////////////////////////

class CXPParameterObj;



class CRunScriptStoredProcedure : public CExtendedStoredProcedure
{
public:
//Constructors / Destructors
  CRunScriptStoredProcedure();
  virtual ~CRunScriptStoredProcedure();

//Methods
  BOOL RunScript(const CComBSTR& bstrScript, LPCTSTR pszLanguage);

//Member variables
  BOOL                                     m_bTerminateScript;    //Should we terminate running the script early
  BOOL                                     m_bReturnCode;         //What will be the return value from Run() above
  int                                      m_nParameters;         //The size of the array m_ppParameters
  CComObjectStackWithQI<CXPServerObj>      m_serverObj;           //The IXPServer COM object
  CComObjectStackWithQI<CXPRunScriptSite>  m_scriptSiteObj;       //The IActiveScriptSite COM object
  CComObjectStackWithQI<CXPParameterObj>*  m_pParametersCOMArray; //The array of parameters each represented by a COM IXPParameter interface
  CComObjectStackWithQI<CXPParametersObj>  m_parametersObj;       //The IXPParameters COM object

  friend class CXPRunScriptSite;
};


//The Extended stored procedure class to run a script from a specified file
class CRunScriptDiskStoredProcedure : public CRunScriptStoredProcedure
{
public:
//Methods
  virtual void DisplayUsage(); //Override used to display help about this XP
  virtual BOOL Run();          //Override to do the real work of this XP

  DECLARE_XP_PARAMETERS()
};


//The Extended stored procedure class to run a script from a parameter
class CRunScriptParameterStoredProcedure : public CRunScriptStoredProcedure
{
public:
//Methods
  virtual void DisplayUsage(); //Override used to display help about this XP
  virtual BOOL Run();          //Override to do the real work of this XP

  DECLARE_XP_PARAMETERS()
};

//The Extended stored procedure class to run a COM / ActiveX object
class CRunActiveXStoredProcedure : public CRunScriptStoredProcedure
{
public:
//Methods
  virtual void DisplayUsage(); //Override used to display help about this XP
  virtual BOOL Run();          //Override to do the real work of this XP

  DECLARE_XP_PARAMETERS()
};

#endif //__XP_RUNSCRIPTPROCS_H__
