#include "stdafx.h"
#include "PingMachine.h"

//This function forms the Win32 DLL implementation
BOOL IcmpServerReachable(LPCTSTR pszServer)
{
  USES_CONVERSION;

  //Create the ICMP handle
  HANDLE hICMP = IcmpCreateFile();
  if (hICMP == INVALID_HANDLE_VALUE)
    return FALSE;

  //Do a DNS lookup of the server name if not already in dotted notation
	LPSTR lpszAsciiServer = T2A((LPTSTR) pszServer);
	unsigned long	ipAddress = inet_addr(lpszAsciiServer);
	if (ipAddress == INADDR_NONE)
	{
		//Not a dotted address, then do a lookup of the name
		hostent* hp = gethostbyname(lpszAsciiServer);
		if (hp)
			memcpy(&ipAddress, hp->h_addr, hp->h_length);
		else
			return FALSE;
	}


  //assume the worst
  BOOL bSuccess = FALSE;


  //do the actual ping, note we hard code a number of values to reasonable default values
  
  BYTE byDataToSend[32];
  WORD wDataSize = sizeof(byDataToSend);
  memset(byDataToSend, 'E', wDataSize); //Use 32 "E"'s which is what the command line ping utility uses by default

  //Set the options to use for the ping
  IP_OPTION_INFORMATION ipOptions;
  memset(&ipOptions, 0, sizeof(ipOptions));
  ipOptions.Ttl = 10;

  //Setup a stack buffer to receive the reply
  BYTE byReplyBuffer[60];  //60 is sizeof(ICMP_ECHO_REPLY) + 32 which is the data size we are sending
  DWORD nReplies = IcmpSendEcho(hICMP, ipAddress, byDataToSend, wDataSize, &ipOptions, byReplyBuffer, sizeof(byReplyBuffer), 5000); //Use a default timeout of 5 seconds

  //Check the reply if one was received
  if (nReplies == 1)
  {
    //Check what we got back is what we sent
    bSuccess = TRUE;
    ICMP_ECHO_REPLY* pEchoReply = (ICMP_ECHO_REPLY*) byReplyBuffer;
    char* pReplyData = (char*) pEchoReply->Data;
    for (int i=0; i<pEchoReply->DataSize && bSuccess; i++)
      bSuccess = pReplyData[i] == 'E';
  }

  //Free the ICMP handle we have been using
  IcmpCloseHandle(hICMP);

  //Return the success indicator
  return bSuccess;
}
