#include "stdafx.h"
#include "XPServerReachable.h"
#include "PingMachine.h"

IMPLEMENT_VERSION_XP()                                            //Support the optional __GetXpVersion function
IMPLEMENT_XP(XP_SERVERREACHABLE, CServerReachableStoredProcedure) //Implement the XP_SERVERREACHABLE Extended Stored Procedure

#define SERVER_INDEX          0
#define REACHABLE_INDEX       1

BEGIN_IMPLEMENT_XP_PARAMETERS(CServerReachableStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_OUTPUT_PARAMETER(2, NULL, XP_PT_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                      TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CServerReachableStoredProcedure)

void CServerReachableStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SERVERREACHABLE @Server, @Reachable OUTPUT");
}

BOOL CServerReachableStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //By default assume the worst
  BOOL bSuccess = FALSE;

  //Call the Win32 DLL function implementation
  TCHAR* pszServer = XP_TTEXT(m_pParameterData[SERVER_INDEX].m_Data);
  BOOL bReachable = IcmpServerReachable(pszServer);

  //convert the bReachable value into the output parameter data type
  int nConvertedDataSize = Convert(SRVINT4, &bReachable, sizeof(bReachable), m_pParameterData[REACHABLE_INDEX].m_bType, m_pParameterData[REACHABLE_INDEX].m_pData, m_pParameterData[REACHABLE_INDEX].m_cbMaxLen);
  if ((nConvertedDataSize > 0))
  {
    //Return the data in the output parameter
    bSuccess = (ParamSetOutput(REACHABLE_INDEX+1, m_pParameterData[REACHABLE_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to convert the verified value to the required output type, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}
