/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_SMTPSENDMAILProc.h"



/////////////////////////////// Implementation ////////////////////////////////

#define TO_INDEX              0
#define FROM_INDEX            1
#define SUBJECT_INDEX         2
#define BODY_INDEX            3
#define HTML_INDEX            4
#define SENDUSINGPICKUP_INDEX 5
#define SERVER_INDEX          6 
#define PORT_INDEX            7
#define TIMEOUT_INDEX         8
#define AUTHENTICATION_INDEX  9
#define USER_INDEX            10
#define PASSWORD_INDEX        11
#define CC_INDEX              12
#define BCC_INDEX             13
#define ATTACHMENTS_INDEX     14

BEGIN_IMPLEMENT_XP_PARAMETERS(CSMTPSendMailStoredProcedure)
  XP_INPUT_PARAMETER(0, "@To",              XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@From",            XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@Subject",         XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@Body",            XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@HTML",            XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(0, "@SendUsingPickup", XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(0, "@Server",          XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@Port",            XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(0, "@Timeout",         XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(0, "@Authentication",  XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4,                                                                  FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(0, "@User",            XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@Password",        XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@CC",              XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@BCC",             XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
  XP_INPUT_PARAMETER(0, "@Attachments",     XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, TRUE,  XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CSMTPSendMailStoredProcedure)

void CSMTPSendMailStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_SENDMAIL @To=Recipient(s), @From=Sender, @Subject=Something, @Body=Something, [@HTML=0|1], [@SendUsingPickup=0,1], [@Server=?], [@Port=?], [@Timeout=?], [@Authentication=?], [@User=?], [@Password=?], [@CC=?]. [@BCC=?], [@Attachments=?]");
}

BOOL CSMTPSendMailStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterDefinitions);
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  try
  {
    //create the CDO Message object
    CDO::IMessagePtr msg;
    HRESULT hr = msg.CreateInstance(__uuidof(CDO::Message));
    if (FAILED(hr))
      _com_issue_errorex(hr, msg, __uuidof(msg));

    //Handle the "From" field
    TCHAR* pszFrom = XP_TTEXT(m_pParameterData[FROM_INDEX].m_Data);
    XP_ASSERT(pszFrom);
    msg->From = pszFrom;

    //Handle the "To" field
    TCHAR* pszTo = XP_TTEXT(m_pParameterData[TO_INDEX].m_Data);
    XP_ASSERT(pszTo);
    msg->To = pszTo;

    //Handle the "Subject" field
    TCHAR* pszSubject = XP_TTEXT(m_pParameterData[SUBJECT_INDEX].m_Data);
    XP_ASSERT(pszSubject);
    msg->Subject = pszSubject;

    //Handle the "Body" field
    TCHAR* pszBody = XP_TTEXT(m_pParameterData[BODY_INDEX].m_Data);
    XP_ASSERT(pszBody);

    //Handle the optional "HTML" parameter
    if (m_pParameterData[HTML_INDEX].m_bType != SRVNULL)
    {
      long nHTML = XP_INT4(m_pParameterData[HTML_INDEX].m_Data);
      if (nHTML)
        msg->HTMLBody = pszBody;
      else
        msg->TextBody = pszBody;
    }
    else
      msg->TextBody = pszBody;

    //Handle the optional "CC" parameter
    if (m_pParameterData[CC_INDEX].m_bType != SRVNULL)
    {
      TCHAR* pszCC = XP_TTEXT(m_pParameterData[CC_INDEX].m_Data);
      XP_ASSERT(pszCC);
      msg->CC = pszCC;
    }

    //Handle the optional "BCC" parameter
    if (m_pParameterData[BCC_INDEX].m_bType != SRVNULL)
    {
      TCHAR* pszBCC = XP_TTEXT(m_pParameterData[BCC_INDEX].m_Data);
      XP_ASSERT(pszBCC);
      msg->BCC = pszBCC;
    }

    //Handle the optional "attachments" parameter
    if (m_pParameterData[ATTACHMENTS_INDEX].m_bType != SRVNULL)
    {
      TCHAR* pszAttachments = XP_TTEXT(m_pParameterData[ATTACHMENTS_INDEX].m_Data);
      int nAttachmentsLength = (int) _tcslen(pszAttachments);
      XP_ASSERT(pszAttachments);
	    for (int pos=0, start=0; pos<=nAttachmentsLength; pos++)
	    {
		    //Valid separators between attachments are ',' or ';'
		    if ((pszAttachments[pos] == ',') || (pszAttachments[pos] == ';') || (pszAttachments[pos] == '\0'))
		    {
			    pszAttachments[pos] = 0;	//Redundant when at the end of string, but who cares.

          //Add the attachment to the message  
          msg->AddAttachment(_bstr_t(&pszAttachments[start]), _bstr_t(""), _bstr_t(""));

          //Move on to the next position
			    start = pos + 1;
		    }
	    }
    }

    BOOL bRequiresFieldUpdate = FALSE;

    //Handle the optional "SendUsingPickup" parameter
    if (m_pParameterData[SENDUSINGPICKUP_INDEX].m_bType != SRVNULL)
    {
      long nSendUsingPickup = XP_INT4(m_pParameterData[SENDUSINGPICKUP_INDEX].m_Data);
      if (nSendUsingPickup)
      {
        msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/sendusing")]->Value = _variant_t((long)CDO::cdoSendUsingPickup);
        bRequiresFieldUpdate = TRUE;
      }  
      else
      {
        msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/sendusing")]->Value = _variant_t((long)CDO::cdoSendUsingPort);
        bRequiresFieldUpdate = TRUE;

        //handle the now required "Server" field
        if (m_pParameterData[SERVER_INDEX].m_bType == SRVNULL)
        {
          sprintf(szErrorMsg, "%s: \"@Server\" parameter is required if using sending using a port", m_szFunctionName);
          SendErrorMsg(szErrorMsg);
          return FALSE;
        }
        TCHAR* pszServer = XP_TTEXT(m_pParameterData[SERVER_INDEX].m_Data);
        XP_ASSERT(pszServer);
        msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpserver")]->Value = _variant_t(pszServer);

        //Handle the optional "Port" parameter
        if (m_pParameterData[PORT_INDEX].m_bType != SRVNULL)
        {
          long nPort = XP_INT4(m_pParameterData[PORT_INDEX].m_Data);
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpserverport")]->Value = _variant_t(nPort);
        }

        //Handle the optional "Timeout" parameter
        if (m_pParameterData[TIMEOUT_INDEX].m_bType != SRVNULL)
        {
          long nTimeout = XP_INT4(m_pParameterData[TIMEOUT_INDEX].m_Data);
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpconnectiontimeout")]->Value = _variant_t(nTimeout);
        }
      }
    }

    //Handle the optional "Authentication" parameter
    if (m_pParameterData[AUTHENTICATION_INDEX].m_bType != SRVNULL)
    {
      long nAuthentication = XP_INT4(m_pParameterData[AUTHENTICATION_INDEX].m_Data);
      switch (nAuthentication)
      {
        case CDO::cdoBasic:
        {
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpauthenticate")]->Value = _variant_t(nAuthentication);
          bRequiresFieldUpdate = TRUE;

          //handle the now required "User" field
          if (m_pParameterData[USER_INDEX].m_bType == SRVNULL)
          {
            sprintf(szErrorMsg, "%s: \"@User\" parameter is required if using Basic Authentication", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }
          TCHAR* pszUser = XP_TTEXT(m_pParameterData[USER_INDEX].m_Data);
          XP_ASSERT(pszUser);
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/sendusername")]->Value = _variant_t(pszUser);

          //Handle the now required "Password" field
          if (m_pParameterData[PASSWORD_INDEX].m_bType == SRVNULL)
          {
            sprintf(szErrorMsg, "%s: \"@Password\" parameter is required if using Basic Authentication", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
            return FALSE;
          }
          TCHAR* pszPassword = XP_TTEXT(m_pParameterData[PASSWORD_INDEX].m_Data);
          XP_ASSERT(pszPassword);
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/sendpassword")]->Value = _variant_t(pszPassword);

          break;
        }
        case CDO::cdoNTLM:
        {
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpauthenticate")]->Value = _variant_t(nAuthentication);
          bRequiresFieldUpdate = TRUE;
          break;
        }
        case CDO::cdoAnonymous:
        {
          msg->Configuration->Fields->Item[_bstr_t("http://schemas.microsoft.com/cdo/configuration/smtpauthenticate")]->Value = _variant_t(nAuthentication);
          bRequiresFieldUpdate = TRUE;
          break;
        }
        default:
        {
          sprintf(szErrorMsg, "%s: \"@Authentication\" parameter has an undefined value of %d", m_szFunctionName, nAuthentication);
          SendErrorMsg(szErrorMsg);
          return FALSE;

          break;
        }
      }
    }

    //Update the fields if necessary
    if (bRequiresFieldUpdate)
      msg->Configuration->Fields->Update();
    
    //Finally, send the message
    msg->Send();
  }
  catch(_com_error e)
  {
    //Report the error
    sprintf(szErrorMsg, "%s: An unexpected CDO error has occurred, HRESULT:%x, Description: %s", m_szFunctionName, e.Error(), e.Description().operator const char *());
    SendErrorMsg(szErrorMsg);
    return FALSE;
  }

  //We got here, so the running of the XP was successful
  return TRUE;
}
