#include "stdafx.h"

#define XP_NOERROR	   0 //return code - no error
#define XP_ERROR	     1 //return code - error
#define MAX_LEN_INPUT	20 //maximum length of input data

#ifdef __cplusplus
extern "C" {
#endif
RETCODE __declspec(dllexport) xp_buffov(SRV_PROC* srvproc);
#ifdef __cplusplus
}
#endif

//Automatically link to the ODS lib file
#pragma comment(lib, "Opends60.lib")

//return ODS version
__declspec(dllexport) ULONG __GetXpVersion()
{
  return ODS_VERSION;
}

//accept only char data (also NULL)
inline BOOL ValidInputDataTypes(BYTE bType)
{
  return (bType == SRVNULL || bType == SRVBIGCHAR || bType == SRVBIGVARCHAR || bType == SRVCHAR || bType == SRVVARCHAR);
}

//send message
void XPprint(char* Msg, SRV_PROC* srvproc)
{
  srv_sendmsg(srvproc, SRV_MSG_INFO, 0, (DBTINYINT)0, (DBTINYINT)0, NULL, 0, 0, Msg, SRV_NULLTERM);
}

//send error message+usage
void ShowErrorAndUsageMsg(char* ErrorMsg, SRV_PROC* srvproc)
{
  XPprint(ErrorMsg, srvproc);//send error message
  XPprint("Usage: EXEC xp_buffov data. data is ASCII, up to 20 characters", srvproc); //send usage
}

void test(char* str, SRV_PROC* srvproc) 
{
  char buff1[10];    //array to be overflowed
  strcpy(buff1,str); //this is when it happens
  XPprint(buff1, srvproc);
}

void neverHappens(char* str, SRV_PROC* srvproc) 
{
  str;      //Avoids unreferrenced level 4 warning
  srvproc;  //Avoids unreferrenced level 4 warning

  //this function is never called
  XPprint("This code should never run!", srvproc);
}

RETCODE __declspec(dllexport) xp_buffov(SRV_PROC* srvproc)
{
  int iParamCount = srv_rpcparams(srvproc);  
    
  //error if !=1 parameters entered
  if (iParamCount != 1)
  {
    ShowErrorAndUsageMsg("Error!, Wrong number of parameters.",srvproc);
    return XP_ERROR;
  }

  //validate parameter I/O type
  if ((srv_paramstatus(srvproc, 1) & SRV_PARAMRETURN) == 1)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 should not be an output parameter.", srvproc);
    return XP_ERROR;
  }

  //read parameter 1's attributes
  BYTE  bType;       //data type of parameter
  ULONG cbMaxLen;    //maximum length of parameter
  ULONG cbActualLen; //actual length of parameter
  BOOL  fNull;       //true if parameter is NULL
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, NULL, &fNull) == FAIL)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 must be ASCII data.",srvproc);
    return XP_ERROR;
  }

  //accept only char data for parameter 1
  if (!ValidInputDataTypes(bType))
  {
    ShowErrorAndUsageMsg("Error!, Wrong type of parameter 1.",srvproc);
    return XP_ERROR;
  }  

  //avoid buffer overflow in parameter 1 (at least until the deliberate buffer overflow in test()!!!)
  if (cbActualLen > MAX_LEN_INPUT)
  {
    ShowErrorAndUsageMsg("Error!, The size of parameter 1 is too big.",srvproc);
    return XP_ERROR;
  }

  //Reject the input parameter if it is NULL
  if (fNull)
  {
    ShowErrorAndUsageMsg("Error!, Parameter 1 cannot be null.", srvproc);
    return XP_ERROR;
  }

  //Allocate some heap memory to hold the parameter's data
  char* pszMessage = NULL;
  try
  {
    pszMessage = new char[cbActualLen+1];
  }
#ifdef _AFX
  catch(CMemoryException* pEx)
  {
    pEx->Delete();
  }
#endif
  catch(std::bad_alloc&)
  {
  }
  if (pszMessage == NULL)
  {
    ShowErrorAndUsageMsg("Error!, Failed to allocate memory to contain parameter 1 data.",srvproc);
    return XP_ERROR;
  }

  //Get the input parameter data
  if (srv_paraminfo(srvproc, 1, &bType, &cbMaxLen, &cbActualLen, (unsigned char *)pszMessage, &fNull) == FAIL)
  {
    //Tidy up the heap memory before we exit
    delete [] pszMessage;

    ShowErrorAndUsageMsg("Error!, Failed to obtain information for parameter 1.",srvproc);
    return XP_ERROR;
  }
  pszMessage[cbActualLen] = '\0'; //NULL terminate the data now that we have received its data

  //Call the test function which demonstrates the buffer overflow problem
  test(pszMessage, srvproc);

  //Tidy up the heap memory before we exit
  delete [] pszMessage;

  //We got this far so return success
  return XP_NOERROR; 
}

