/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPReadODBCTable.h"



/////////////////////////////// Macros / Defines //////////////////////////////

TCHAR* g_pszDSN = _T("local"); //The ODBC DSN to connect to, change to suit your setup



/////////////////////////////// Implementation ////////////////////////////////

#define TABLENAME_PARAMETER_INDEX   0

BEGIN_IMPLEMENT_XP_PARAMETERS(CReadODBCTableStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
END_IMPLEMENT_XP_PARAMETERS(CReadODBCTableStoredProcedure)

void CReadODBCTableStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: xp_odbc @TableName");
}

BOOL CReadODBCTableStoredProcedure::Run()
{
  USES_CONVERSION;

  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //Get the name of the Table we are to read from (already setup for us in CExtendedStoredProcedure::VerifyAndInitializeParameters)
  TCHAR* pszTableName = XP_TTEXT(m_pParameterData[TABLENAME_PARAMETER_INDEX].m_Data);
  XP_ASSERT(pszTableName);

  //Assume the worst
  BOOL bSuccess = FALSE;
  int nRows = 0;

  //Create an ODBC environment handle
  SQLHENV hEnv = SQL_NULL_HENV;
  SQLRETURN sret = SQLAllocHandle(SQL_HANDLE_ENV, NULL, &hEnv);
  if (sret == SQL_SUCCESS)
  {
    //We want ODBC 3.x behaviour
    SQLSetEnvAttr(hEnv, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER);

    //Create an ODBC connection handle
    BOOL bImpersonated = FALSE;
    TCHAR pszUserName[MAXNAME];
    TCHAR pszPassword[MAXNAME];
    HDBC hDBConnection = SQL_NULL_HDBC;
    sret = SQLAllocHandle(SQL_HANDLE_DBC, hEnv, &hDBConnection);
    if (sret == SQL_SUCCESS)
    {
      //Setup the security we use for the ODBC connection
      if (strcmp(PField(SRV_LSECURE), "TRUE") == 0)
      {
        //client connection is using integrated security
        //bImpersonated = this->ImpersonateClient();
        bImpersonated = (ImpersonateClient() == SUCCEED);
        SQLSetConnectAttr(hDBConnection, SQL_INTEGRATED_SECURITY, (SQLPOINTER) SQL_IS_ON, SQL_IS_INTEGER);
        _tcscpy(pszUserName, _T(""));
        _tcscpy(pszPassword, _T(""));
      }
      else
      {
        //client connection is using standard user name / password security
        _tcscpy(pszUserName, A2T(PField(SRV_USER)));
        _tcscpy(pszPassword, A2T(PField(SRV_PWD)));
      }

      //Now do the ODBC connection
      sret = SQLConnect(hDBConnection, (SQLCHAR*) g_pszDSN, SQL_NTS, (SQLCHAR*) pszUserName, SQL_NTS, (SQLCHAR*) pszPassword, SQL_NTS);
      if (sret == SQL_SUCCESS || sret == SQL_SUCCESS_WITH_INFO)
      {
        //Create the statement handle
        HSTMT hstmt = SQL_NULL_HSTMT;
        sret = SQLAllocHandle(SQL_HANDLE_STMT, hDBConnection, &hstmt);
        if (sret == SQL_SUCCESS)
        {
          //Get the bind token sent into the XP
          char pszBindToken[256];
          if (GetBindToken(pszBindToken) == SUCCEED)
          {
            //Set the just retrieved bind token as an input parameter to sp_binsession
            if (SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_VARCHAR, 255, 0, pszBindToken, 256, NULL) == SQL_SUCCESS)
            {
              //Now use sp_bindsession to tie our fresh ODBC session to the client's 
              //session so that we can share transaction space
              TCHAR* pszCallBindSession = _T("{call sp_bindsession(?)}");
              sret = SQLExecDirect(hstmt, (SQLCHAR*) pszCallBindSession, SQL_NTS);
              if (sret == SQL_SUCCESS || sret == SQL_SUCCESS_WITH_INFO)
              {
                //Reset the statement handle since we are about to reuse it
                SQLFreeStmt(hstmt, SQL_RESET_PARAMS);

                //Now finally we can select all records from the specified table
                TCHAR szExec[128 + (MAXNAME * 3)];
                _tcscpy(szExec, _T("SELECT * FROM "));
                _tcscat(szExec, pszTableName);
                sret = SQLExecDirect(hstmt, (SQLCHAR*) szExec, SQL_NTS);
                if (sret == SQL_SUCCESS)
                {
                  //Get the number of columns in the results set
                  SQLSMALLINT nCols = 0;
                  SQLNumResultCols(hstmt, &nCols);

                  //Allocate memory to hold the column data from the stack
                  HANDLE hHeap = GetProcessHeap();
                  PBYTE* ppColData = (PBYTE*) _alloca(nCols * sizeof(PBYTE));
                  memset(ppColData, 0, nCols * sizeof(PBYTE));
                  SQLINTEGER* pColTypes = (SQLINTEGER*) _alloca(nCols * sizeof(SQLINTEGER));

                  //Build the column descriptions for this result set
                  BOOL bMemoryAllocatedOK = TRUE;
                  int nCol;
                  for (nCol=0; nCol<nCols && bMemoryAllocatedOK; nCol++)
                  {
                    //Get the column name
                    TCHAR pszColumnName[MAXNAME];
                    SQLSMALLINT cbAttr;
                    SQLINTEGER iNumAttr;
                    SQLColAttribute(hstmt, (SQLSMALLINT) (nCol + 1), SQL_DESC_NAME, pszColumnName, MAXNAME, &cbAttr, &iNumAttr);

                    //Get the column length
                    SQLINTEGER cbColData = 0; 
                    SQLColAttribute(hstmt, (SQLSMALLINT) (nCol + 1), SQL_DESC_OCTET_LENGTH, NULL, 0, NULL, &cbColData);

                    //Get the column SQL Server data type
                    SQLSMALLINT eSQLType;
                    SQLColAttribute(hstmt, (SQLSMALLINT) (nCol + 1), SQL_CA_SS_COLUMN_SSTYPE, NULL, 0, NULL, &eSQLType);

                    //Over-write the column length returned by ODBC with the correct value to be used by ODS
		                switch (eSQLType)  
                    {
			                case SQLMONEYN: //Deliberate fallthrough
			                case SQLMONEY:
                      {
                        cbColData = sizeof(DBMONEY);
                        break;
                      }
			                case SQLDATETIMN: //Deliberate fallthrough
			                case SQLDATETIME:
                      {
				                cbColData = sizeof(DBDATETIME);
				                break;
                      }
			                case SQLNUMERIC: //Deliberate fallthrough
			                case SQLDECIMAL:
                      {
				                cbColData = sizeof(DBNUMERIC);
				                break;
                      }
			                case SQLMONEY4:
                      {
                        cbColData = sizeof(DBMONEY4);
                        break;
                      }
			                case SQLDATETIM4:
                      {
				                cbColData = sizeof(DBDATETIM4);
				                break;
		                  }
                      default:
                      {
                        break;
                      }
                    }
                    
                    //Allocate memory for the row data from the heap
                    ppColData[nCol] = (BYTE*) HeapAlloc(hHeap, HEAP_ZERO_MEMORY, cbColData);
                    bMemoryAllocatedOK = (ppColData[nCol] != NULL);
                    if (bMemoryAllocatedOK)
                    {
                      //Bind the column
                      SQLBindCol(hstmt, (SQLUSMALLINT) (nCol + 1), SQL_C_BINARY, ppColData[nCol], cbColData, &(pColTypes[nCol]));

                      //Prepare structure that will be returned back via ODS
                      Describe(nCol + 1, pszColumnName, SRV_NULLTERM, eSQLType, cbColData, eSQLType, cbColData, NULL);
                    }
                  }

                  if (bMemoryAllocatedOK)
                  {
                    //If we got here then we are successful
                    bSuccess = TRUE;

                    //now return each row of the results set
                    sret = SQLFetch(hstmt);
                    while ((sret == SQL_SUCCESS || sret == SQL_SUCCESS_WITH_INFO) && !GotAttention())
                    {
                      for (nCol=0; nCol<nCols; nCol++)
                      {
                        int nColSize = (pColTypes[nCol] == SQL_NULL_DATA) ? 0 : pColTypes[nCol];
                        SetColumnLength(nCol+1, nColSize);
                        SetColumnData(nCol+1, ppColData[nCol]);
                      }

                      //Send the completed row back to the client
                      if (SendRow() == SUCCEED)
                        nRows++;

                      //Prepare for the next time around
                      sret = SQLFetch(hstmt);
                    }
                  }
                  else
                  {
                    sprintf(szErrorMsg, "%s: Failed to allocate memory to contain results set column information", m_szFunctionName);
                    SendErrorMsg(szErrorMsg);
                  }

                  //Free up the column memory we have used
                  for (nCol=0; nCol<nCols; nCol++)
                  {
                    if (ppColData[nCol])
                      HeapFree(hHeap, 0, ppColData[nCol]);
                  }
                }
                else
                {
                  sprintf(szErrorMsg, "%s: Failed to retrieve results from table", m_szFunctionName);
                  HandleODBCError(szErrorMsg, hstmt, SQL_HANDLE_STMT);
                }
              }
              else
              {
                sprintf(szErrorMsg, "%s: Failed to execute sp_bindsession stored procedure", m_szFunctionName);
                HandleODBCError(szErrorMsg, hstmt, SQL_HANDLE_STMT);
              }
            }
            else
            {
              sprintf(szErrorMsg, "%s: Failed to set bind token", m_szFunctionName);
              HandleODBCError(szErrorMsg, hstmt, SQL_HANDLE_STMT);
            }
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to get bind token", m_szFunctionName);
            SendErrorMsg(szErrorMsg);
          }

          //Free up the statement handle
          SQLFreeHandle(SQL_HANDLE_STMT, hstmt);
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to allocate ODBC statement handle", m_szFunctionName);
          HandleODBCError(szErrorMsg, hDBConnection, SQL_HANDLE_DBC);
        }

        //Disconnect before we return
        SQLDisconnect(hDBConnection);
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to connect to database", m_szFunctionName);
        HandleODBCError(szErrorMsg, hDBConnection, SQL_HANDLE_DBC);
      }

      //Free the ODBC connection handle
      SQLFreeHandle(SQL_HANDLE_DBC, hDBConnection);

      //Revert the security if needed
      if (bImpersonated)
        RevertToSelf();
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to allocate ODBC connection handle", m_szFunctionName);
      HandleODBCError(szErrorMsg, hEnv, SQL_HANDLE_ENV);
    }

    //Free the ODBC environment handle
    SQLFreeEnv(hEnv);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to allocate ODBC environment handle", m_szFunctionName);
    HandleODBCError(szErrorMsg, hEnv, SQL_HANDLE_ENV);
  }
  
  //Finally tell ODS that we have completed
  if (bSuccess)
  {
    if (nRows > 0)
      SendDone(SRV_DONE_MORE | SRV_DONE_COUNT, 0, nRows);
    else
      SendDone(SRV_DONE_MORE, 0, 0);
  }
  else
    SendDone(SRV_DONE_ERROR | SRV_DONE_MORE, 0, 0);

  return bSuccess;
}

void CReadODBCTableStoredProcedure::HandleODBCError(DBCHAR* pszInitialMsg, SQLHANDLE herror, SQLSMALLINT htype)
{
  //Send the initial error message
  SendErrorMsg(pszInitialMsg);

  //Sent each of the ODBC error messages
  SQLSMALLINT cbErrorMsg;
  SQLCHAR szErrorMsg[SQL_MAX_MESSAGE_LENGTH + 1];
  SQLCHAR sqlState[6];
  SQLINTEGER NativeError;
  SQLSMALLINT nRec = 1;
  while (SQLGetDiagRec(htype, herror, nRec, sqlState, &NativeError, szErrorMsg, SQL_MAX_MESSAGE_LENGTH, &cbErrorMsg) == SQL_SUCCESS)
  {
    nRec++;
    SendErrorMsg((DBCHAR*) szErrorMsg);
  }
}

